"""Module for defining types used in neumann package."""

import numpy as np

boolnp = np.uint8          #np.bool
intnp = np.int
longnp = np.int64          #np.long
doublenp = np.float64      #np.double
complexnp = np.complex128
