/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": []
    }, 
    "module_name": "qmeq.approach.c_lindblad"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__qmeq__approach__c_lindblad
#define __PYX_HAVE_API__qmeq__approach__c_lindblad
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include <math.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "qmeq\\approach\\c_lindblad.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":725
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":726
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":727
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":728
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":732
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":733
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":734
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":735
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":739
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":740
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":749
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":750
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":751
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":753
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":754
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":755
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":757
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":758
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":760
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":761
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":762
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "specfuncc.pxd":4
 * cimport numpy as np
 * 
 * ctypedef np.uint8_t bool_t             # <<<<<<<<<<<<<<
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t
 */
typedef __pyx_t_5numpy_uint8_t __pyx_t_4qmeq_9specfuncc_bool_t;

/* "specfuncc.pxd":5
 * 
 * ctypedef np.uint8_t bool_t
 * ctypedef np.int_t int_t             # <<<<<<<<<<<<<<
 * ctypedef np.int64_t long_t
 * ctypedef np.float64_t double_t
 */
typedef __pyx_t_5numpy_int_t __pyx_t_4qmeq_9specfuncc_int_t;

/* "specfuncc.pxd":6
 * ctypedef np.uint8_t bool_t
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t             # <<<<<<<<<<<<<<
 * ctypedef np.float64_t double_t
 * ctypedef np.complex128_t complex_t
 */
typedef __pyx_t_5numpy_int64_t __pyx_t_4qmeq_9specfuncc_long_t;

/* "specfuncc.pxd":7
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t
 * ctypedef np.float64_t double_t             # <<<<<<<<<<<<<<
 * ctypedef np.complex128_t complex_t
 * 
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_4qmeq_9specfuncc_double_t;

/* "qmeq/approach/c_lindblad.pyx":19
 * cimport cython
 * 
 * ctypedef np.uint8_t bool_t             # <<<<<<<<<<<<<<
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t
 */
typedef __pyx_t_5numpy_uint8_t __pyx_t_4qmeq_8approach_10c_lindblad_bool_t;

/* "qmeq/approach/c_lindblad.pyx":20
 * 
 * ctypedef np.uint8_t bool_t
 * ctypedef np.int_t int_t             # <<<<<<<<<<<<<<
 * ctypedef np.int64_t long_t
 * ctypedef np.float64_t double_t
 */
typedef __pyx_t_5numpy_int_t __pyx_t_4qmeq_8approach_10c_lindblad_int_t;

/* "qmeq/approach/c_lindblad.pyx":21
 * ctypedef np.uint8_t bool_t
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t             # <<<<<<<<<<<<<<
 * ctypedef np.float64_t double_t
 * ctypedef np.complex128_t complex_t
 */
typedef __pyx_t_5numpy_int64_t __pyx_t_4qmeq_8approach_10c_lindblad_long_t;

/* "qmeq/approach/c_lindblad.pyx":22
 * ctypedef np.int_t int_t
 * ctypedef np.int64_t long_t
 * ctypedef np.float64_t double_t             # <<<<<<<<<<<<<<
 * ctypedef np.complex128_t complex_t
 * 
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_4qmeq_8approach_10c_lindblad_double_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< npy_float64 > __pyx_t_npy_float64_complex;
  #else
    typedef npy_float64 _Complex __pyx_t_npy_float64_complex;
  #endif
#else
    typedef struct { npy_float64 real, imag; } __pyx_t_npy_float64_complex;
#endif
static CYTHON_INLINE __pyx_t_npy_float64_complex __pyx_t_npy_float64_complex_from_parts(npy_float64, npy_float64);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);


/*--- Type declarations ---*/

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":764
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":765
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":766
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":768
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFormatCheck.proto */
static CYTHON_INLINE int  __Pyx_GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type); // PROTO

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
#define __Pyx_BufPtrStrided3d(type, buf, i0, s0, i1, s1, i2, s2) (type)((char*)buf + i0 * s0 + i1 * s1 + i2 * s2)
/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static CYTHON_INLINE int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* None.proto */
static Py_ssize_t __Pyx_zeros[] = {0, 0, 0, 0, 0, 0, 0, 0};
static Py_ssize_t __Pyx_minusones[] = {-1, -1, -1, -1, -1, -1, -1, -1};

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_long(npy_long value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value);

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_npy_float64(a, b)   ((a)==(b))
    #define __Pyx_c_sum_npy_float64(a, b)  ((a)+(b))
    #define __Pyx_c_diff_npy_float64(a, b) ((a)-(b))
    #define __Pyx_c_prod_npy_float64(a, b) ((a)*(b))
    #define __Pyx_c_quot_npy_float64(a, b) ((a)/(b))
    #define __Pyx_c_neg_npy_float64(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_npy_float64(z) ((z)==(npy_float64)0)
    #define __Pyx_c_conj_npy_float64(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_npy_float64(z)     (::std::abs(z))
        #define __Pyx_c_pow_npy_float64(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_npy_float64(z) ((z)==0)
    #define __Pyx_c_conj_npy_float64(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_npy_float64(z)     (cabs(z))
        #define __Pyx_c_pow_npy_float64(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_sum_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_diff_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_prod_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_quot_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_neg_npy_float64(__pyx_t_npy_float64_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_npy_float64(__pyx_t_npy_float64_complex);
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_conj_npy_float64(__pyx_t_npy_float64_complex);
    #if 1
        static CYTHON_INLINE npy_float64 __Pyx_c_abs_npy_float64(__pyx_t_npy_float64_complex);
        static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_pow_npy_float64(__pyx_t_npy_float64_complex, __pyx_t_npy_float64_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_long __Pyx_PyInt_As_npy_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'qmeq.specfuncc' */
static __pyx_t_4qmeq_9specfuncc_int_t (*__pyx_f_4qmeq_9specfuncc_func_pauli)(__pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_int_t, PyArrayObject *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.math' */

/* Module declarations from 'qmeq.approach.c_lindblad' */
static __Pyx_TypeInfo __Pyx_TypeInfo___pyx_t_double_complex = { "double complex", NULL, sizeof(__pyx_t_double_complex), { 0 }, 0, 'C', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t = { "double_t", NULL, sizeof(__pyx_t_4qmeq_8approach_10c_lindblad_double_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t = { "long_t", NULL, sizeof(__pyx_t_4qmeq_8approach_10c_lindblad_long_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_4qmeq_8approach_10c_lindblad_long_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_4qmeq_8approach_10c_lindblad_long_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t = { "bool_t", NULL, sizeof(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t), 0 };
#define __Pyx_MODULE_NAME "qmeq.approach.c_lindblad"
int __pyx_module_is_main_qmeq__approach__c_lindblad = 0;

/* Implementation of 'qmeq.approach.c_lindblad' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_E[] = "E";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_Ea[] = "Ea";
static const char __pyx_k_ap[] = "ap";
static const char __pyx_k_bb[] = "bb";
static const char __pyx_k_bp[] = "bp";
static const char __pyx_k_cp[] = "cp";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_qd[] = "qd";
static const char __pyx_k_si[] = "si";
static const char __pyx_k_Eba[] = "Eba";
static const char __pyx_k_Tba[] = "Tba";
static const char __pyx_k_aap[] = "aap";
static const char __pyx_k_bbp[] = "bbp";
static const char __pyx_k_bpp[] = "bpp";
static const char __pyx_k_ccp[] = "ccp";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_rez[] = "rez";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_aapi[] = "aapi";
static const char __pyx_k_bbpi[] = "bbpi";
static const char __pyx_k_bbpp[] = "bbpp";
static const char __pyx_k_bvec[] = "bvec";
static const char __pyx_k_ccpi[] = "ccpi";
static const char __pyx_k_dlst[] = "dlst";
static const char __pyx_k_fcta[] = "fcta";
static const char __pyx_k_fctc[] = "fctc";
static const char __pyx_k_imag[] = "imag";
static const char __pyx_k_kern[] = "kern";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_ndm0[] = "ndm0";
static const char __pyx_k_ndm1[] = "ndm1";
static const char __pyx_k_norm[] = "norm";
static const char __pyx_k_phi0[] = "phi0";
static const char __pyx_k_real[] = "real";
static const char __pyx_k_symq[] = "symq";
static const char __pyx_k_tLba[] = "tLba";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tlst[] = "tlst";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_bbppi[] = "bbppi";
static const char __pyx_k_bppbp[] = "bppbp";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_funcp[] = "funcp";
static const char __pyx_k_itype[] = "itype";
static const char __pyx_k_leads[] = "leads";
static const char __pyx_k_mulst[] = "mulst";
static const char __pyx_k_ndm0r[] = "ndm0r";
static const char __pyx_k_nmany[] = "nmany";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_phi0p[] = "phi0p";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_bppbpi[] = "bppbpi";
static const char __pyx_k_charge[] = "charge";
static const char __pyx_k_dictdm[] = "dictdm";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_lenlst[] = "lenlst";
static const char __pyx_k_mapdm0[] = "mapdm0";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_nleads[] = "nleads";
static const char __pyx_k_npauli[] = "npauli";
static const char __pyx_k_aap_sgn[] = "aap_sgn";
static const char __pyx_k_acharge[] = "acharge";
static const char __pyx_k_bcharge[] = "bcharge";
static const char __pyx_k_booldm0[] = "booldm0";
static const char __pyx_k_ccharge[] = "ccharge";
static const char __pyx_k_ccp_sgn[] = "ccp_sgn";
static const char __pyx_k_conjdm0[] = "conjdm0";
static const char __pyx_k_current[] = "current";
static const char __pyx_k_fct_aap[] = "fct_aap";
static const char __pyx_k_fct_ccp[] = "fct_ccp";
static const char __pyx_k_mytypes[] = "mytypes";
static const char __pyx_k_ncharge[] = "ncharge";
static const char __pyx_k_phi0aap[] = "phi0aap";
static const char __pyx_k_phi0bbp[] = "phi0bbp";
static const char __pyx_k_phi0ccp[] = "phi0ccp";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_product[] = "product";
static const char __pyx_k_Approach[] = "Approach";
static const char __pyx_k_Lindblad[] = "Lindblad";
static const char __pyx_k_aprclass[] = "aprclass";
static const char __pyx_k_bbp_bool[] = "bbp_bool";
static const char __pyx_k_bbp_conj[] = "bbp_conj";
static const char __pyx_k_bbpp_sgn[] = "bbpp_sgn";
static const char __pyx_k_doublenp[] = "doublenp";
static const char __pyx_k_fct_bbpp[] = "fct_bbpp";
static const char __pyx_k_kerntype[] = "kerntype";
static const char __pyx_k_last_row[] = "last_row";
static const char __pyx_k_norm_row[] = "norm_row";
static const char __pyx_k_phi0bbpp[] = "phi0bbpp";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_statesdm[] = "statesdm";
static const char __pyx_k_bbpi_bool[] = "bbpi_bool";
static const char __pyx_k_bppbp_sgn[] = "bppbp_sgn";
static const char __pyx_k_complexnp[] = "complexnp";
static const char __pyx_k_fct_bppbp[] = "fct_bppbp";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_norm_rowp[] = "norm_rowp";
static const char __pyx_k_phi0bppbp[] = "phi0bppbp";
static const char __pyx_k_shiftlst0[] = "shiftlst0";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_i_dphi0_dt[] = "i_dphi0_dt";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_concatenate[] = "concatenate";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_generate_fct[] = "generate_fct";
static const char __pyx_k_generate_vec[] = "generate_vec";
static const char __pyx_k_heat_current[] = "heat_current";
static const char __pyx_k_generate_kern[] = "generate_kern";
static const char __pyx_k_energy_current[] = "energy_current";
static const char __pyx_k_c_generate_tLba[] = "c_generate_tLba";
static const char __pyx_k_generate_current[] = "generate_current";
static const char __pyx_k_Approach_Lindblad[] = "Approach_Lindblad";
static const char __pyx_k_c_generate_vec_lindblad[] = "c_generate_vec_lindblad";
static const char __pyx_k_c_generate_kern_lindblad[] = "c_generate_kern_lindblad";
static const char __pyx_k_qmeq_approach_c_lindblad[] = "qmeq.approach.c_lindblad";
static const char __pyx_k_c_generate_current_lindblad[] = "c_generate_current_lindblad";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_combinations_with_replacement[] = "combinations_with_replacement";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_C_Users_Gediminas_Desktop_qmeq_1[] = "C:\\Users\\Gediminas\\Desktop\\qmeq-1.0-pypi\\qmeq\\approach\\c_lindblad.pyx";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Module_containing_cython_functio[] = "Module containing cython functions, which generate first order Lindblad kernel.\n   For docstrings see documentation of module lindblad.";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_n_s_Approach;
static PyObject *__pyx_n_s_Approach_Lindblad;
static PyObject *__pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1;
static PyObject *__pyx_n_s_E;
static PyObject *__pyx_n_s_Ea;
static PyObject *__pyx_n_s_Eba;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_Lindblad;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_Tba;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_aap;
static PyObject *__pyx_n_s_aap_sgn;
static PyObject *__pyx_n_s_aapi;
static PyObject *__pyx_n_s_acharge;
static PyObject *__pyx_n_s_ap;
static PyObject *__pyx_n_s_aprclass;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_bb;
static PyObject *__pyx_n_s_bbp;
static PyObject *__pyx_n_s_bbp_bool;
static PyObject *__pyx_n_s_bbp_conj;
static PyObject *__pyx_n_s_bbpi;
static PyObject *__pyx_n_s_bbpi_bool;
static PyObject *__pyx_n_s_bbpp;
static PyObject *__pyx_n_s_bbpp_sgn;
static PyObject *__pyx_n_s_bbppi;
static PyObject *__pyx_n_s_bcharge;
static PyObject *__pyx_n_s_booldm0;
static PyObject *__pyx_n_s_bp;
static PyObject *__pyx_n_s_bpp;
static PyObject *__pyx_n_s_bppbp;
static PyObject *__pyx_n_s_bppbp_sgn;
static PyObject *__pyx_n_s_bppbpi;
static PyObject *__pyx_n_s_bvec;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_s_c_generate_current_lindblad;
static PyObject *__pyx_n_s_c_generate_kern_lindblad;
static PyObject *__pyx_n_s_c_generate_tLba;
static PyObject *__pyx_n_s_c_generate_vec_lindblad;
static PyObject *__pyx_n_s_ccharge;
static PyObject *__pyx_n_s_ccp;
static PyObject *__pyx_n_s_ccp_sgn;
static PyObject *__pyx_n_s_ccpi;
static PyObject *__pyx_n_s_charge;
static PyObject *__pyx_n_s_combinations_with_replacement;
static PyObject *__pyx_n_s_complexnp;
static PyObject *__pyx_n_s_concatenate;
static PyObject *__pyx_n_s_conjdm0;
static PyObject *__pyx_n_s_cp;
static PyObject *__pyx_n_s_current;
static PyObject *__pyx_n_s_dictdm;
static PyObject *__pyx_n_s_dlst;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_doublenp;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_energy_current;
static PyObject *__pyx_n_s_fct_aap;
static PyObject *__pyx_n_s_fct_bbpp;
static PyObject *__pyx_n_s_fct_bppbp;
static PyObject *__pyx_n_s_fct_ccp;
static PyObject *__pyx_n_s_fcta;
static PyObject *__pyx_n_s_fctc;
static PyObject *__pyx_n_s_funcp;
static PyObject *__pyx_n_s_generate_current;
static PyObject *__pyx_n_s_generate_fct;
static PyObject *__pyx_n_s_generate_kern;
static PyObject *__pyx_n_s_generate_vec;
static PyObject *__pyx_n_s_heat_current;
static PyObject *__pyx_n_s_i_dphi0_dt;
static PyObject *__pyx_n_s_imag;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_itype;
static PyObject *__pyx_n_s_kern;
static PyObject *__pyx_n_s_kerntype;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_last_row;
static PyObject *__pyx_n_s_leads;
static PyObject *__pyx_n_s_lenlst;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mapdm0;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_mulst;
static PyObject *__pyx_n_s_mytypes;
static PyObject *__pyx_n_s_ncharge;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndm0;
static PyObject *__pyx_n_s_ndm0r;
static PyObject *__pyx_n_s_ndm1;
static PyObject *__pyx_n_s_nleads;
static PyObject *__pyx_n_s_nmany;
static PyObject *__pyx_n_s_norm;
static PyObject *__pyx_n_s_norm_row;
static PyObject *__pyx_n_s_norm_rowp;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_npauli;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_phi0;
static PyObject *__pyx_n_s_phi0aap;
static PyObject *__pyx_n_s_phi0bbp;
static PyObject *__pyx_n_s_phi0bbpp;
static PyObject *__pyx_n_s_phi0bppbp;
static PyObject *__pyx_n_s_phi0ccp;
static PyObject *__pyx_n_s_phi0p;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_product;
static PyObject *__pyx_n_s_qd;
static PyObject *__pyx_n_s_qmeq_approach_c_lindblad;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_real;
static PyObject *__pyx_n_s_rez;
static PyObject *__pyx_n_s_shiftlst0;
static PyObject *__pyx_n_s_si;
static PyObject *__pyx_n_s_statesdm;
static PyObject *__pyx_n_s_symq;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_tLba;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tlst;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_c_generate_tLba(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys); /* proto */
static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_2c_generate_kern_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys); /* proto */
static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_4c_generate_current_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys); /* proto */
static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_6c_generate_vec_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_phi0p, PyObject *__pyx_v_sys); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;

/* "qmeq/approach/c_lindblad.pyx":32
 * @cython.cdivision(True)
 * @cython.boundscheck(False)
 * def c_generate_tLba(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

/* Python wrapper */
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_1c_generate_tLba(PyObject *__pyx_self, PyObject *__pyx_v_sys); /*proto*/
static PyMethodDef __pyx_mdef_4qmeq_8approach_10c_lindblad_1c_generate_tLba = {"c_generate_tLba", (PyCFunction)__pyx_pw_4qmeq_8approach_10c_lindblad_1c_generate_tLba, METH_O, 0};
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_1c_generate_tLba(PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_generate_tLba (wrapper)", 0);
  __pyx_r = __pyx_pf_4qmeq_8approach_10c_lindblad_c_generate_tLba(__pyx_self, ((PyObject *)__pyx_v_sys));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_c_generate_tLba(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyArrayObject *__pyx_v_Tba = 0;
  PyArrayObject *__pyx_v_E = 0;
  PyObject *__pyx_v_si = NULL;
  PyArrayObject *__pyx_v_mulst = 0;
  PyArrayObject *__pyx_v_tlst = 0;
  PyArrayObject *__pyx_v_dlst = 0;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_itype;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_b;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_a;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bcharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_acharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_charge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_l;
  __pyx_t_4qmeq_8approach_10c_lindblad_double_t __pyx_v_Eba;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_nleads;
  PyArrayObject *__pyx_v_rez = 0;
  PyArrayObject *__pyx_v_tLba = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_E;
  __Pyx_Buffer __pyx_pybuffer_E;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_Tba;
  __Pyx_Buffer __pyx_pybuffer_Tba;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dlst;
  __Pyx_Buffer __pyx_pybuffer_dlst;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mulst;
  __Pyx_Buffer __pyx_pybuffer_mulst;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_rez;
  __Pyx_Buffer __pyx_pybuffer_rez;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tLba;
  __Pyx_Buffer __pyx_pybuffer_tLba;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tlst;
  __Pyx_Buffer __pyx_pybuffer_tlst;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyArrayObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  long __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  Py_ssize_t __pyx_t_17;
  PyObject *(*__pyx_t_18)(PyObject *);
  PyObject *(*__pyx_t_19)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_20;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_21;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_22;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_32;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_35;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_39;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_42;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_43;
  __Pyx_RefNannySetupContext("c_generate_tLba", 0);
  __pyx_pybuffer_Tba.pybuffer.buf = NULL;
  __pyx_pybuffer_Tba.refcount = 0;
  __pyx_pybuffernd_Tba.data = NULL;
  __pyx_pybuffernd_Tba.rcbuffer = &__pyx_pybuffer_Tba;
  __pyx_pybuffer_E.pybuffer.buf = NULL;
  __pyx_pybuffer_E.refcount = 0;
  __pyx_pybuffernd_E.data = NULL;
  __pyx_pybuffernd_E.rcbuffer = &__pyx_pybuffer_E;
  __pyx_pybuffer_mulst.pybuffer.buf = NULL;
  __pyx_pybuffer_mulst.refcount = 0;
  __pyx_pybuffernd_mulst.data = NULL;
  __pyx_pybuffernd_mulst.rcbuffer = &__pyx_pybuffer_mulst;
  __pyx_pybuffer_tlst.pybuffer.buf = NULL;
  __pyx_pybuffer_tlst.refcount = 0;
  __pyx_pybuffernd_tlst.data = NULL;
  __pyx_pybuffernd_tlst.rcbuffer = &__pyx_pybuffer_tlst;
  __pyx_pybuffer_dlst.pybuffer.buf = NULL;
  __pyx_pybuffer_dlst.refcount = 0;
  __pyx_pybuffernd_dlst.data = NULL;
  __pyx_pybuffernd_dlst.rcbuffer = &__pyx_pybuffer_dlst;
  __pyx_pybuffer_rez.pybuffer.buf = NULL;
  __pyx_pybuffer_rez.refcount = 0;
  __pyx_pybuffernd_rez.data = NULL;
  __pyx_pybuffernd_rez.rcbuffer = &__pyx_pybuffer_rez;
  __pyx_pybuffer_tLba.pybuffer.buf = NULL;
  __pyx_pybuffer_tLba.refcount = 0;
  __pyx_pybuffernd_tLba.data = NULL;
  __pyx_pybuffernd_tLba.rcbuffer = &__pyx_pybuffer_tLba;

  /* "qmeq/approach/c_lindblad.pyx":33
 * @cython.boundscheck(False)
 * def c_generate_tLba(sys):
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     si = sys.si
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_leads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Tba); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 33, __pyx_L1_error)
  __pyx_t_3 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_Tba.rcbuffer->pybuffer, (PyObject*)__pyx_t_3, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_Tba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_Tba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 33, __pyx_L1_error)
    } else {__pyx_pybuffernd_Tba.diminfo[0].strides = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_Tba.diminfo[0].shape = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_Tba.diminfo[1].strides = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_Tba.diminfo[1].shape = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_Tba.diminfo[2].strides = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_Tba.diminfo[2].shape = __pyx_pybuffernd_Tba.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_3 = 0;
  __pyx_v_Tba = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":34
 * def c_generate_tLba(sys):
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea             # <<<<<<<<<<<<<<
 *     si = sys.si
 *     cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_qd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_Ea); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_E.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_E = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_E.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 34, __pyx_L1_error)
    } else {__pyx_pybuffernd_E.diminfo[0].strides = __pyx_pybuffernd_E.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_E.diminfo[0].shape = __pyx_pybuffernd_E.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_E = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":35
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     si = sys.si             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
 *     cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_si); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_si = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":36
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     si = sys.si
 *     cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst
 *     cdef np.ndarray[double_t, ndim=2] dlst = sys.leads.dlst
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_leads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mulst); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mulst.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mulst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mulst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 36, __pyx_L1_error)
    } else {__pyx_pybuffernd_mulst.diminfo[0].strides = __pyx_pybuffernd_mulst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mulst.diminfo[0].shape = __pyx_pybuffernd_mulst.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_mulst = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":37
 *     si = sys.si
 *     cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
 *     cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=2] dlst = sys.leads.dlst
 *     #
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_leads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_tlst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 37, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tlst.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_tlst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tlst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 37, __pyx_L1_error)
    } else {__pyx_pybuffernd_tlst.diminfo[0].strides = __pyx_pybuffernd_tlst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tlst.diminfo[0].shape = __pyx_pybuffernd_tlst.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_tlst = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":38
 *     cdef np.ndarray[double_t, ndim=1] mulst = sys.leads.mulst
 *     cdef np.ndarray[double_t, ndim=1] tlst = sys.leads.tlst
 *     cdef np.ndarray[double_t, ndim=2] dlst = sys.leads.dlst             # <<<<<<<<<<<<<<
 *     #
 *     #mtype = sys.leads.mtype
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_leads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dlst); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dlst.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_dlst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_dlst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 38, __pyx_L1_error)
    } else {__pyx_pybuffernd_dlst.diminfo[0].strides = __pyx_pybuffernd_dlst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dlst.diminfo[0].shape = __pyx_pybuffernd_dlst.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_dlst.diminfo[1].strides = __pyx_pybuffernd_dlst.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_dlst.diminfo[1].shape = __pyx_pybuffernd_dlst.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_dlst = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":41
 *     #
 *     #mtype = sys.leads.mtype
 *     cdef int_t itype = sys.funcp.itype             # <<<<<<<<<<<<<<
 *     #
 *     cdef long_t b, a
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_funcp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_itype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_npy_long(__pyx_t_1); if (unlikely((__pyx_t_8 == ((npy_long)-1)) && PyErr_Occurred())) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_itype = __pyx_t_8;

  /* "qmeq/approach/c_lindblad.pyx":46
 *     cdef int_t bcharge, acharge, charge, l
 *     cdef double_t Eba
 *     cdef int_t nleads = si.nleads             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[double_t, ndim=1] rez = np.zeros(2, dtype=doublenp)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nleads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyInt_As_npy_long(__pyx_t_1); if (unlikely((__pyx_t_8 == ((npy_long)-1)) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nleads = __pyx_t_8;

  /* "qmeq/approach/c_lindblad.pyx":48
 *     cdef int_t nleads = si.nleads
 *     #
 *     cdef np.ndarray[double_t, ndim=1] rez = np.zeros(2, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[complex_t, ndim=3] tLba = np.zeros((nleads, si.nmany, si.nmany), dtype=complexnp)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple_, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_9);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_rez.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_rez = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_rez.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 48, __pyx_L1_error)
    } else {__pyx_pybuffernd_rez.diminfo[0].strides = __pyx_pybuffernd_rez.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_rez.diminfo[0].shape = __pyx_pybuffernd_rez.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_rez = ((PyArrayObject *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "qmeq/approach/c_lindblad.pyx":50
 *     cdef np.ndarray[double_t, ndim=1] rez = np.zeros(2, dtype=doublenp)
 *     #
 *     cdef np.ndarray[complex_t, ndim=3] tLba = np.zeros((nleads, si.nmany, si.nmany), dtype=complexnp)             # <<<<<<<<<<<<<<
 *     for charge in range(si.ncharge-1):
 *         bcharge = charge+1
 */
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyInt_From_npy_long(__pyx_v_nleads); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nmany); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nmany); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_11);
  __pyx_t_9 = 0;
  __pyx_t_2 = 0;
  __pyx_t_11 = 0;
  __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_12);
  __pyx_t_12 = 0;
  __pyx_t_12 = PyDict_New(); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_12, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_tLba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 50, __pyx_L1_error)
    } else {__pyx_pybuffernd_tLba.diminfo[0].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tLba.diminfo[0].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tLba.diminfo[1].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tLba.diminfo[1].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_tLba.diminfo[2].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_tLba.diminfo[2].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_13 = 0;
  __pyx_v_tLba = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":51
 *     #
 *     cdef np.ndarray[complex_t, ndim=3] tLba = np.zeros((nleads, si.nmany, si.nmany), dtype=complexnp)
 *     for charge in range(si.ncharge-1):             # <<<<<<<<<<<<<<
 *         bcharge = charge+1
 *         acharge = charge
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ncharge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_12 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_14 = __Pyx_PyInt_As_long(__pyx_t_12); if (unlikely((__pyx_t_14 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_14; __pyx_t_8+=1) {
    __pyx_v_charge = __pyx_t_8;

    /* "qmeq/approach/c_lindblad.pyx":52
 *     cdef np.ndarray[complex_t, ndim=3] tLba = np.zeros((nleads, si.nmany, si.nmany), dtype=complexnp)
 *     for charge in range(si.ncharge-1):
 *         bcharge = charge+1             # <<<<<<<<<<<<<<
 *         acharge = charge
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
 */
    __pyx_v_bcharge = (__pyx_v_charge + 1);

    /* "qmeq/approach/c_lindblad.pyx":53
 *     for charge in range(si.ncharge-1):
 *         bcharge = charge+1
 *         acharge = charge             # <<<<<<<<<<<<<<
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
 *             Eba = E[b]-E[a]
 */
    __pyx_v_acharge = __pyx_v_charge;

    /* "qmeq/approach/c_lindblad.pyx":54
 *         bcharge = charge+1
 *         acharge = charge
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):             # <<<<<<<<<<<<<<
 *             Eba = E[b]-E[a]
 *             for l in range(nleads):
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_product); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_bcharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 54, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_15 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_15 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_t_9};
      __pyx_t_12 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_t_9};
      __pyx_t_12 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else
    #endif
    {
      __pyx_t_16 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      if (__pyx_t_2) {
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_2); __pyx_t_2 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_15, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_15, __pyx_t_9);
      __pyx_t_1 = 0;
      __pyx_t_9 = 0;
      __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_16, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
      __pyx_t_11 = __pyx_t_12; __Pyx_INCREF(__pyx_t_11); __pyx_t_17 = 0;
      __pyx_t_18 = NULL;
    } else {
      __pyx_t_17 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_18 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 54, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    for (;;) {
      if (likely(!__pyx_t_18)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          if (__pyx_t_17 >= PyList_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_17); __Pyx_INCREF(__pyx_t_12); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 54, __pyx_L1_error)
          #else
          __pyx_t_12 = PySequence_ITEM(__pyx_t_11, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 54, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        } else {
          if (__pyx_t_17 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_17); __Pyx_INCREF(__pyx_t_12); __pyx_t_17++; if (unlikely(0 < 0)) __PYX_ERR(0, 54, __pyx_L1_error)
          #else
          __pyx_t_12 = PySequence_ITEM(__pyx_t_11, __pyx_t_17); __pyx_t_17++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 54, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        }
      } else {
        __pyx_t_12 = __pyx_t_18(__pyx_t_11);
        if (unlikely(!__pyx_t_12)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 54, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_12);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_12))) || (PyList_CheckExact(__pyx_t_12))) {
        PyObject* sequence = __pyx_t_12;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 54, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_16 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_9);
        #else
        __pyx_t_16 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 54, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 54, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_19 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_16 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_16)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_16);
        index = 1; __pyx_t_9 = __pyx_t_19(__pyx_t_1); if (unlikely(!__pyx_t_9)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_9);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_1), 2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
        __pyx_t_19 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_19 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 54, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_t_20 = __Pyx_PyInt_As_npy_int64(__pyx_t_16); if (unlikely((__pyx_t_20 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_21 = __Pyx_PyInt_As_npy_int64(__pyx_t_9); if (unlikely((__pyx_t_21 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_b = __pyx_t_20;
      __pyx_v_a = __pyx_t_21;

      /* "qmeq/approach/c_lindblad.pyx":55
 *         acharge = charge
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
 *             Eba = E[b]-E[a]             # <<<<<<<<<<<<<<
 *             for l in range(nleads):
 *                 #fct1 = fermi_func((E[b]-E[a]-mulst[l])/tlst[l])
 */
      __pyx_t_21 = __pyx_v_b;
      if (__pyx_t_21 < 0) __pyx_t_21 += __pyx_pybuffernd_E.diminfo[0].shape;
      __pyx_t_20 = __pyx_v_a;
      if (__pyx_t_20 < 0) __pyx_t_20 += __pyx_pybuffernd_E.diminfo[0].shape;
      __pyx_v_Eba = ((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_21, __pyx_pybuffernd_E.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_20, __pyx_pybuffernd_E.diminfo[0].strides)));

      /* "qmeq/approach/c_lindblad.pyx":56
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):
 *             Eba = E[b]-E[a]
 *             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                 #fct1 = fermi_func((E[b]-E[a]-mulst[l])/tlst[l])
 *                 #fct2 = 1-fct1
 */
      __pyx_t_22 = __pyx_v_nleads;
      for (__pyx_t_23 = 0; __pyx_t_23 < __pyx_t_22; __pyx_t_23+=1) {
        __pyx_v_l = __pyx_t_23;

        /* "qmeq/approach/c_lindblad.pyx":59
 *                 #fct1 = fermi_func((E[b]-E[a]-mulst[l])/tlst[l])
 *                 #fct2 = 1-fct1
 *                 func_pauli(Eba, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype, rez)             # <<<<<<<<<<<<<<
 *                 tLba[l, b, a] = sqrt(rez[0])*Tba[l, b, a]
 *                 tLba[l, a, b] = sqrt(rez[1])*Tba[l, a, b]
 */
        __pyx_t_24 = __pyx_v_l;
        if (__pyx_t_24 < 0) __pyx_t_24 += __pyx_pybuffernd_mulst.diminfo[0].shape;
        __pyx_t_25 = __pyx_v_l;
        if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_tlst.diminfo[0].shape;
        __pyx_t_26 = __pyx_v_l;
        __pyx_t_27 = 0;
        if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_dlst.diminfo[0].shape;
        if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_dlst.diminfo[1].shape;
        __pyx_t_28 = __pyx_v_l;
        __pyx_t_29 = 1;
        if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_dlst.diminfo[0].shape;
        if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_dlst.diminfo[1].shape;
        __pyx_f_4qmeq_9specfuncc_func_pauli(__pyx_v_Eba, (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_mulst.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_mulst.diminfo[0].strides)), (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_tlst.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_tlst.diminfo[0].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_dlst.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_dlst.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_dlst.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_dlst.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_dlst.diminfo[0].strides, __pyx_t_29, __pyx_pybuffernd_dlst.diminfo[1].strides)), __pyx_v_itype, ((PyArrayObject *)__pyx_v_rez));

        /* "qmeq/approach/c_lindblad.pyx":60
 *                 #fct2 = 1-fct1
 *                 func_pauli(Eba, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype, rez)
 *                 tLba[l, b, a] = sqrt(rez[0])*Tba[l, b, a]             # <<<<<<<<<<<<<<
 *                 tLba[l, a, b] = sqrt(rez[1])*Tba[l, a, b]
 *     sys.tLba = tLba
 */
        __pyx_t_30 = 0;
        if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_rez.diminfo[0].shape;
        __pyx_t_31 = __pyx_v_l;
        __pyx_t_32 = __pyx_v_b;
        __pyx_t_33 = __pyx_v_a;
        if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_Tba.diminfo[0].shape;
        if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_Tba.diminfo[1].shape;
        if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_Tba.diminfo[2].shape;
        __pyx_t_34 = __pyx_v_l;
        __pyx_t_35 = __pyx_v_b;
        __pyx_t_36 = __pyx_v_a;
        if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_tLba.diminfo[0].shape;
        if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_tLba.diminfo[1].shape;
        if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_tLba.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_35, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_36, __pyx_pybuffernd_tLba.diminfo[2].strides) = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(sqrt((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_rez.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_rez.diminfo[0].strides))), 0), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_Tba.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_Tba.diminfo[0].strides, __pyx_t_32, __pyx_pybuffernd_Tba.diminfo[1].strides, __pyx_t_33, __pyx_pybuffernd_Tba.diminfo[2].strides)));

        /* "qmeq/approach/c_lindblad.pyx":61
 *                 func_pauli(Eba, mulst[l], tlst[l], dlst[l,0], dlst[l,1], itype, rez)
 *                 tLba[l, b, a] = sqrt(rez[0])*Tba[l, b, a]
 *                 tLba[l, a, b] = sqrt(rez[1])*Tba[l, a, b]             # <<<<<<<<<<<<<<
 *     sys.tLba = tLba
 *     return 0
 */
        __pyx_t_37 = 1;
        if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_rez.diminfo[0].shape;
        __pyx_t_38 = __pyx_v_l;
        __pyx_t_39 = __pyx_v_a;
        __pyx_t_40 = __pyx_v_b;
        if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_Tba.diminfo[0].shape;
        if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_Tba.diminfo[1].shape;
        if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_Tba.diminfo[2].shape;
        __pyx_t_41 = __pyx_v_l;
        __pyx_t_42 = __pyx_v_a;
        __pyx_t_43 = __pyx_v_b;
        if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_tLba.diminfo[0].shape;
        if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_tLba.diminfo[1].shape;
        if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_tLba.diminfo[2].shape;
        *__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_43, __pyx_pybuffernd_tLba.diminfo[2].strides) = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(sqrt((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_rez.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_rez.diminfo[0].strides))), 0), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_Tba.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_Tba.diminfo[0].strides, __pyx_t_39, __pyx_pybuffernd_Tba.diminfo[1].strides, __pyx_t_40, __pyx_pybuffernd_Tba.diminfo[2].strides)));
      }

      /* "qmeq/approach/c_lindblad.pyx":54
 *         bcharge = charge+1
 *         acharge = charge
 *         for b, a in itertools.product(si.statesdm[bcharge], si.statesdm[acharge]):             # <<<<<<<<<<<<<<
 *             Eba = E[b]-E[a]
 *             for l in range(nleads):
 */
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }

  /* "qmeq/approach/c_lindblad.pyx":62
 *                 tLba[l, b, a] = sqrt(rez[0])*Tba[l, b, a]
 *                 tLba[l, a, b] = sqrt(rez[1])*Tba[l, a, b]
 *     sys.tLba = tLba             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_tLba, ((PyObject *)__pyx_v_tLba)) < 0) __PYX_ERR(0, 62, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":63
 *                 tLba[l, a, b] = sqrt(rez[1])*Tba[l, a, b]
 *     sys.tLba = tLba
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "qmeq/approach/c_lindblad.pyx":32
 * @cython.cdivision(True)
 * @cython.boundscheck(False)
 * def c_generate_tLba(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_16);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tba.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dlst.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mulst.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rez.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tlst.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("qmeq.approach.c_lindblad.c_generate_tLba", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_Tba.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dlst.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mulst.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_rez.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tlst.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_Tba);
  __Pyx_XDECREF((PyObject *)__pyx_v_E);
  __Pyx_XDECREF(__pyx_v_si);
  __Pyx_XDECREF((PyObject *)__pyx_v_mulst);
  __Pyx_XDECREF((PyObject *)__pyx_v_tlst);
  __Pyx_XDECREF((PyObject *)__pyx_v_dlst);
  __Pyx_XDECREF((PyObject *)__pyx_v_rez);
  __Pyx_XDECREF((PyObject *)__pyx_v_tLba);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "qmeq/approach/c_lindblad.pyx":66
 * 
 * @cython.boundscheck(False)
 * def c_generate_kern_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 */

/* Python wrapper */
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_3c_generate_kern_lindblad(PyObject *__pyx_self, PyObject *__pyx_v_sys); /*proto*/
static PyMethodDef __pyx_mdef_4qmeq_8approach_10c_lindblad_3c_generate_kern_lindblad = {"c_generate_kern_lindblad", (PyCFunction)__pyx_pw_4qmeq_8approach_10c_lindblad_3c_generate_kern_lindblad, METH_O, 0};
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_3c_generate_kern_lindblad(PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_generate_kern_lindblad (wrapper)", 0);
  __pyx_r = __pyx_pf_4qmeq_8approach_10c_lindblad_2c_generate_kern_lindblad(__pyx_self, ((PyObject *)__pyx_v_sys));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_2c_generate_kern_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyArrayObject *__pyx_v_E = 0;
  PyArrayObject *__pyx_v_tLba = 0;
  PyObject *__pyx_v_si = NULL;
  int __pyx_v_symq;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_norm_rowp;
  __pyx_t_4qmeq_8approach_10c_lindblad_bool_t __pyx_v_bbp_bool;
  __pyx_t_4qmeq_8approach_10c_lindblad_bool_t __pyx_v_bbpi_bool;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_charge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_acharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bcharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_ccharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_l;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_nleads;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_aap_sgn;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bppbp_sgn;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bbpp_sgn;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_ccp_sgn;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_b;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbpi;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bb;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_a;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ap;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_aap;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_aapi;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bpp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bppbp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bppbpi;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbpp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbppi;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_c;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_cp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ccp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ccpi;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_norm_row;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_last_row;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ndm0;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_npauli;
  __pyx_t_double_complex __pyx_v_fct_aap;
  __pyx_t_double_complex __pyx_v_fct_bppbp;
  __pyx_t_double_complex __pyx_v_fct_bbpp;
  __pyx_t_double_complex __pyx_v_fct_ccp;
  PyArrayObject *__pyx_v_lenlst = 0;
  PyArrayObject *__pyx_v_dictdm = 0;
  PyArrayObject *__pyx_v_shiftlst0 = 0;
  PyArrayObject *__pyx_v_mapdm0 = 0;
  PyArrayObject *__pyx_v_booldm0 = 0;
  PyArrayObject *__pyx_v_conjdm0 = 0;
  PyArrayObject *__pyx_v_kern = 0;
  PyArrayObject *__pyx_v_bvec = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_E;
  __Pyx_Buffer __pyx_pybuffer_E;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_booldm0;
  __Pyx_Buffer __pyx_pybuffer_booldm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_bvec;
  __Pyx_Buffer __pyx_pybuffer_bvec;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_conjdm0;
  __Pyx_Buffer __pyx_pybuffer_conjdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dictdm;
  __Pyx_Buffer __pyx_pybuffer_dictdm;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_kern;
  __Pyx_Buffer __pyx_pybuffer_kern;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_lenlst;
  __Pyx_Buffer __pyx_pybuffer_lenlst;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mapdm0;
  __Pyx_Buffer __pyx_pybuffer_mapdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_shiftlst0;
  __Pyx_Buffer __pyx_pybuffer_shiftlst0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tLba;
  __Pyx_Buffer __pyx_pybuffer_tLba;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyArrayObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_6;
  PyArrayObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_13;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  PyArrayObject *__pyx_t_18 = NULL;
  long __pyx_t_19;
  int __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  Py_ssize_t __pyx_t_22;
  PyObject *(*__pyx_t_23)(PyObject *);
  PyObject *(*__pyx_t_24)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_30;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_33;
  int __pyx_t_34;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_35;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_36;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_37;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_38;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_39;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_40;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_41;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_42;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_43;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_44;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_45;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_46;
  PyObject *__pyx_t_47 = NULL;
  PyObject *__pyx_t_48 = NULL;
  Py_ssize_t __pyx_t_49;
  PyObject *(*__pyx_t_50)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_51;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  Py_ssize_t __pyx_t_54;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_55;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_56;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_57;
  Py_ssize_t __pyx_t_58;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_59;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_60;
  Py_ssize_t __pyx_t_61;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_62;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_63;
  Py_ssize_t __pyx_t_64;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_65;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_66;
  Py_ssize_t __pyx_t_67;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_68;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_69;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_70;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_71;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_72;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_73;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_74;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_75;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_76;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_77;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_78;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_79;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_80;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_81;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_82;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_83;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_84;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_85;
  Py_ssize_t __pyx_t_86;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_87;
  Py_ssize_t __pyx_t_88;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_89;
  Py_ssize_t __pyx_t_90;
  PyObject *(*__pyx_t_91)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_92;
  Py_ssize_t __pyx_t_93;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_94;
  Py_ssize_t __pyx_t_95;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_96;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_97;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_98;
  Py_ssize_t __pyx_t_99;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_100;
  Py_ssize_t __pyx_t_101;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_102;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_103;
  Py_ssize_t __pyx_t_104;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_105;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_106;
  Py_ssize_t __pyx_t_107;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_108;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_109;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_110;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_111;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_112;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_113;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_114;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_115;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_116;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_117;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_118;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_119;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_120;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_121;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_122;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_123;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_124;
  Py_ssize_t __pyx_t_125;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_126;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_127;
  Py_ssize_t __pyx_t_128;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_129;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_130;
  Py_ssize_t __pyx_t_131;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_132;
  Py_ssize_t __pyx_t_133;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_134;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_135;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_136;
  Py_ssize_t __pyx_t_137;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_138;
  Py_ssize_t __pyx_t_139;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_140;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_141;
  Py_ssize_t __pyx_t_142;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_143;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_144;
  Py_ssize_t __pyx_t_145;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_146;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_147;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_148;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_149;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_150;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_151;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_152;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_153;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_154;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_155;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_156;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_157;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_158;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_159;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_160;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_161;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_162;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_163;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_164;
  Py_ssize_t __pyx_t_165;
  Py_ssize_t __pyx_t_166;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_167;
  Py_ssize_t __pyx_t_168;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_169;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_170;
  Py_ssize_t __pyx_t_171;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_172;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_173;
  Py_ssize_t __pyx_t_174;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_175;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_176;
  Py_ssize_t __pyx_t_177;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_178;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_179;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_180;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_181;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_182;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_183;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_184;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_185;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_186;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_187;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_188;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_189;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_190;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_191;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_192;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_193;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_194;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_195;
  Py_ssize_t __pyx_t_196;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_197;
  Py_ssize_t __pyx_t_198;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_199;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_200;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_201;
  __Pyx_RefNannySetupContext("c_generate_kern_lindblad", 0);
  __pyx_pybuffer_E.pybuffer.buf = NULL;
  __pyx_pybuffer_E.refcount = 0;
  __pyx_pybuffernd_E.data = NULL;
  __pyx_pybuffernd_E.rcbuffer = &__pyx_pybuffer_E;
  __pyx_pybuffer_tLba.pybuffer.buf = NULL;
  __pyx_pybuffer_tLba.refcount = 0;
  __pyx_pybuffernd_tLba.data = NULL;
  __pyx_pybuffernd_tLba.rcbuffer = &__pyx_pybuffer_tLba;
  __pyx_pybuffer_lenlst.pybuffer.buf = NULL;
  __pyx_pybuffer_lenlst.refcount = 0;
  __pyx_pybuffernd_lenlst.data = NULL;
  __pyx_pybuffernd_lenlst.rcbuffer = &__pyx_pybuffer_lenlst;
  __pyx_pybuffer_dictdm.pybuffer.buf = NULL;
  __pyx_pybuffer_dictdm.refcount = 0;
  __pyx_pybuffernd_dictdm.data = NULL;
  __pyx_pybuffernd_dictdm.rcbuffer = &__pyx_pybuffer_dictdm;
  __pyx_pybuffer_shiftlst0.pybuffer.buf = NULL;
  __pyx_pybuffer_shiftlst0.refcount = 0;
  __pyx_pybuffernd_shiftlst0.data = NULL;
  __pyx_pybuffernd_shiftlst0.rcbuffer = &__pyx_pybuffer_shiftlst0;
  __pyx_pybuffer_mapdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_mapdm0.refcount = 0;
  __pyx_pybuffernd_mapdm0.data = NULL;
  __pyx_pybuffernd_mapdm0.rcbuffer = &__pyx_pybuffer_mapdm0;
  __pyx_pybuffer_booldm0.pybuffer.buf = NULL;
  __pyx_pybuffer_booldm0.refcount = 0;
  __pyx_pybuffernd_booldm0.data = NULL;
  __pyx_pybuffernd_booldm0.rcbuffer = &__pyx_pybuffer_booldm0;
  __pyx_pybuffer_conjdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_conjdm0.refcount = 0;
  __pyx_pybuffernd_conjdm0.data = NULL;
  __pyx_pybuffernd_conjdm0.rcbuffer = &__pyx_pybuffer_conjdm0;
  __pyx_pybuffer_kern.pybuffer.buf = NULL;
  __pyx_pybuffer_kern.refcount = 0;
  __pyx_pybuffernd_kern.data = NULL;
  __pyx_pybuffernd_kern.rcbuffer = &__pyx_pybuffer_kern;
  __pyx_pybuffer_bvec.pybuffer.buf = NULL;
  __pyx_pybuffer_bvec.refcount = 0;
  __pyx_pybuffernd_bvec.data = NULL;
  __pyx_pybuffernd_bvec.rcbuffer = &__pyx_pybuffer_bvec;

  /* "qmeq/approach/c_lindblad.pyx":67
 * @cython.boundscheck(False)
 * def c_generate_kern_lindblad(sys):
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_qd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Ea); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_t_3 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_E.rcbuffer->pybuffer, (PyObject*)__pyx_t_3, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_E = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_E.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 67, __pyx_L1_error)
    } else {__pyx_pybuffernd_E.diminfo[0].strides = __pyx_pybuffernd_E.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_E.diminfo[0].shape = __pyx_pybuffernd_E.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_3 = 0;
  __pyx_v_E = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":68
 * def c_generate_kern_lindblad(sys):
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba             # <<<<<<<<<<<<<<
 *     si = sys.si
 *     cdef bint symq = sys.funcp.symq
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_tLba); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_tLba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 68, __pyx_L1_error)
    } else {__pyx_pybuffernd_tLba.diminfo[0].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tLba.diminfo[0].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tLba.diminfo[1].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tLba.diminfo[1].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_tLba.diminfo[2].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_tLba.diminfo[2].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_tLba = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":69
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si             # <<<<<<<<<<<<<<
 *     cdef bint symq = sys.funcp.symq
 *     cdef long_t norm_rowp = sys.funcp.norm_row
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_si); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_si = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":70
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si
 *     cdef bint symq = sys.funcp.symq             # <<<<<<<<<<<<<<
 *     cdef long_t norm_rowp = sys.funcp.norm_row
 *     #
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_funcp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_symq); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_symq = __pyx_t_5;

  /* "qmeq/approach/c_lindblad.pyx":71
 *     si = sys.si
 *     cdef bint symq = sys.funcp.symq
 *     cdef long_t norm_rowp = sys.funcp.norm_row             # <<<<<<<<<<<<<<
 *     #
 *     cdef bool_t bbp_bool, bbpi_bool
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_funcp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_norm_row); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_2); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_norm_rowp = __pyx_t_6;

  /* "qmeq/approach/c_lindblad.pyx":83
 *     cdef complex_t fct_aap, fct_bppbp, fct_bbpp, fct_ccp
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_lenlst); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 83, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_lenlst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 83, __pyx_L1_error)
    } else {__pyx_pybuffernd_lenlst.diminfo[0].strides = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_lenlst.diminfo[0].shape = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_lenlst = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":84
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_dictdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 84, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_dictdm = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 84, __pyx_L1_error)
    } else {__pyx_pybuffernd_dictdm.diminfo[0].strides = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dictdm.diminfo[0].shape = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_dictdm = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":85
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_shiftlst0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_shiftlst0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 85, __pyx_L1_error)
    } else {__pyx_pybuffernd_shiftlst0.diminfo[0].strides = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_shiftlst0.diminfo[0].shape = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_shiftlst0 = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":86
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_mapdm0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mapdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 86, __pyx_L1_error)
    } else {__pyx_pybuffernd_mapdm0.diminfo[0].strides = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mapdm0.diminfo[0].shape = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_mapdm0 = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":87
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_booldm0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 87, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_booldm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 87, __pyx_L1_error)
    } else {__pyx_pybuffernd_booldm0.diminfo[0].strides = __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_booldm0.diminfo[0].shape = __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_11 = 0;
  __pyx_v_booldm0 = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":88
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0             # <<<<<<<<<<<<<<
 *     #
 *     norm_row = norm_rowp if symq else si.ndm0r
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_conjdm0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 88, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_conjdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 88, __pyx_L1_error)
    } else {__pyx_pybuffernd_conjdm0.diminfo[0].strides = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_conjdm0.diminfo[0].shape = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_12 = 0;
  __pyx_v_conjdm0 = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":90
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 *     norm_row = norm_rowp if symq else si.ndm0r             # <<<<<<<<<<<<<<
 *     last_row = si.ndm0r-1 if symq else si.ndm0r
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
 */
  if ((__pyx_v_symq != 0)) {
    __pyx_t_6 = __pyx_v_norm_rowp;
  } else {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0r); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_13 = __Pyx_PyInt_As_npy_int64(__pyx_t_2); if (unlikely((__pyx_t_13 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_13;
  }
  __pyx_v_norm_row = __pyx_t_6;

  /* "qmeq/approach/c_lindblad.pyx":91
 *     #
 *     norm_row = norm_rowp if symq else si.ndm0r
 *     last_row = si.ndm0r-1 if symq else si.ndm0r             # <<<<<<<<<<<<<<
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
 *     #
 */
  if ((__pyx_v_symq != 0)) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0r); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_2, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_13 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_13;
  } else {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_13 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_13;
  }
  __pyx_v_last_row = __pyx_t_6;

  /* "qmeq/approach/c_lindblad.pyx":92
 *     norm_row = norm_rowp if symq else si.ndm0r
 *     last_row = si.ndm0r-1 if symq else si.ndm0r
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_npauli); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_13 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nleads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyInt_As_npy_long(__pyx_t_1); if (unlikely((__pyx_t_14 == ((npy_long)-1)) && PyErr_Occurred())) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ndm0 = __pyx_t_6;
  __pyx_v_npauli = __pyx_t_13;
  __pyx_v_nleads = __pyx_t_14;

  /* "qmeq/approach/c_lindblad.pyx":94
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
 *     #
 *     cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)
 *     bvec[norm_row] = 1
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64((__pyx_v_last_row + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0r); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_15);
  __pyx_t_1 = 0;
  __pyx_t_15 = 0;
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_kern.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 2, 0, __pyx_stack) == -1)) {
      __pyx_v_kern = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 94, __pyx_L1_error)
    } else {__pyx_pybuffernd_kern.diminfo[0].strides = __pyx_pybuffernd_kern.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_kern.diminfo[0].shape = __pyx_pybuffernd_kern.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_kern.diminfo[1].strides = __pyx_pybuffernd_kern.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_kern.diminfo[1].shape = __pyx_pybuffernd_kern.rcbuffer->pybuffer.shape[1];
    }
  }
  __pyx_t_17 = 0;
  __pyx_v_kern = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":95
 *     #
 *     cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
 *     cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     bvec[norm_row] = 1
 *     for charge in range(si.ncharge):
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64((__pyx_v_last_row + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_15, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_bvec.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_bvec = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_bvec.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 95, __pyx_L1_error)
    } else {__pyx_pybuffernd_bvec.diminfo[0].strides = __pyx_pybuffernd_bvec.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_bvec.diminfo[0].shape = __pyx_pybuffernd_bvec.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_18 = 0;
  __pyx_v_bvec = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":96
 *     cdef np.ndarray[double_t, ndim=2] kern = np.zeros((last_row+1, si.ndm0r), dtype=doublenp)
 *     cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)
 *     bvec[norm_row] = 1             # <<<<<<<<<<<<<<
 *     for charge in range(si.ncharge):
 *         acharge = charge-1
 */
  __pyx_t_13 = __pyx_v_norm_row;
  if (__pyx_t_13 < 0) __pyx_t_13 += __pyx_pybuffernd_bvec.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_bvec.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_bvec.diminfo[0].strides) = 1.0;

  /* "qmeq/approach/c_lindblad.pyx":97
 *     cdef np.ndarray[double_t, ndim=1] bvec = np.zeros(last_row+1, dtype=doublenp)
 *     bvec[norm_row] = 1
 *     for charge in range(si.ncharge):             # <<<<<<<<<<<<<<
 *         acharge = charge-1
 *         bcharge = charge
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ncharge); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = __Pyx_PyInt_As_long(__pyx_t_2); if (unlikely((__pyx_t_19 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_19; __pyx_t_14+=1) {
    __pyx_v_charge = __pyx_t_14;

    /* "qmeq/approach/c_lindblad.pyx":98
 *     bvec[norm_row] = 1
 *     for charge in range(si.ncharge):
 *         acharge = charge-1             # <<<<<<<<<<<<<<
 *         bcharge = charge
 *         ccharge = charge+1
 */
    __pyx_v_acharge = (__pyx_v_charge - 1);

    /* "qmeq/approach/c_lindblad.pyx":99
 *     for charge in range(si.ncharge):
 *         acharge = charge-1
 *         bcharge = charge             # <<<<<<<<<<<<<<
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
 */
    __pyx_v_bcharge = __pyx_v_charge;

    /* "qmeq/approach/c_lindblad.pyx":100
 *         acharge = charge-1
 *         bcharge = charge
 *         ccharge = charge+1             # <<<<<<<<<<<<<<
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
    __pyx_v_ccharge = (__pyx_v_charge + 1);

    /* "qmeq/approach/c_lindblad.pyx":101
 *         bcharge = charge
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_combinations_with_replacement); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_16 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_bcharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_20 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_15);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_15, function);
        __pyx_t_20 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_15)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_16, __pyx_int_2};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_16, __pyx_int_2};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    } else
    #endif
    {
      __pyx_t_21 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_21);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_16);
      PyTuple_SET_ITEM(__pyx_t_21, 0+__pyx_t_20, __pyx_t_16);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_21, 1+__pyx_t_20, __pyx_int_2);
      __pyx_t_16 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_15 = __pyx_t_2; __Pyx_INCREF(__pyx_t_15); __pyx_t_22 = 0;
      __pyx_t_23 = NULL;
    } else {
      __pyx_t_22 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_23 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 101, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_23)) {
        if (likely(PyList_CheckExact(__pyx_t_15))) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_22); __Pyx_INCREF(__pyx_t_2); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 101, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_15, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_22 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_22); __Pyx_INCREF(__pyx_t_2); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 101, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_15, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_23(__pyx_t_15);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 101, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
        PyObject* sequence = __pyx_t_2;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 101, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_21 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_21 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_21);
        __Pyx_INCREF(__pyx_t_16);
        #else
        __pyx_t_21 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 101, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 101, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        #endif
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_24 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_21 = __pyx_t_24(__pyx_t_1); if (unlikely(!__pyx_t_21)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_21);
        index = 1; __pyx_t_16 = __pyx_t_24(__pyx_t_1); if (unlikely(!__pyx_t_16)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_16);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_24(__pyx_t_1), 2) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
        __pyx_t_24 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_24 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 101, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_21); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_t_25 = __Pyx_PyInt_As_npy_int64(__pyx_t_16); if (unlikely((__pyx_t_25 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_v_b = __pyx_t_6;
      __pyx_v_bp = __pyx_t_25;

      /* "qmeq/approach/c_lindblad.pyx":102
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1 and bbp_bool:
 */
      __pyx_t_26 = __pyx_v_bcharge;
      if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
      __pyx_t_25 = __pyx_v_b;
      if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_6 = __pyx_v_bp;
      if (__pyx_t_6 < 0) __pyx_t_6 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_27 = __pyx_v_bcharge;
      if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
      __pyx_t_28 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
      if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
      __pyx_v_bbp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

      /* "qmeq/approach/c_lindblad.pyx":103
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *             if bbp != -1 and bbp_bool:
 *                 bbpi = ndm0 + bbp - npauli
 */
      __pyx_t_29 = __pyx_v_bcharge;
      if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
      __pyx_t_30 = __pyx_v_b;
      if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_31 = __pyx_v_bp;
      if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_32 = __pyx_v_bcharge;
      if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
      __pyx_t_33 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
      if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_booldm0.diminfo[0].shape;
      __pyx_v_bbp_bool = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_booldm0.diminfo[0].strides));

      /* "qmeq/approach/c_lindblad.pyx":104
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1 and bbp_bool:             # <<<<<<<<<<<<<<
 *                 bbpi = ndm0 + bbp - npauli
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 */
      __pyx_t_34 = ((__pyx_v_bbp != -1LL) != 0);
      if (__pyx_t_34) {
      } else {
        __pyx_t_5 = __pyx_t_34;
        goto __pyx_L10_bool_binop_done;
      }
      __pyx_t_34 = (__pyx_v_bbp_bool != 0);
      __pyx_t_5 = __pyx_t_34;
      __pyx_L10_bool_binop_done:;
      if (__pyx_t_5) {

        /* "qmeq/approach/c_lindblad.pyx":105
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1 and bbp_bool:
 *                 bbpi = ndm0 + bbp - npauli             # <<<<<<<<<<<<<<
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 *                 if bbpi_bool:
 */
        __pyx_v_bbpi = ((__pyx_v_ndm0 + __pyx_v_bbp) - __pyx_v_npauli);

        /* "qmeq/approach/c_lindblad.pyx":106
 *             if bbp != -1 and bbp_bool:
 *                 bbpi = ndm0 + bbp - npauli
 *                 bbpi_bool = True if bbpi >= ndm0 else False             # <<<<<<<<<<<<<<
 *                 if bbpi_bool:
 *                     kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]
 */
        if (((__pyx_v_bbpi >= __pyx_v_ndm0) != 0)) {
          __pyx_t_5 = 1;
        } else {
          __pyx_t_5 = 0;
        }
        __pyx_v_bbpi_bool = __pyx_t_5;

        /* "qmeq/approach/c_lindblad.pyx":107
 *                 bbpi = ndm0 + bbp - npauli
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 *                 if bbpi_bool:             # <<<<<<<<<<<<<<
 *                     kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
 */
        __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
        if (__pyx_t_5) {

          /* "qmeq/approach/c_lindblad.pyx":108
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 *                 if bbpi_bool:
 *                     kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]             # <<<<<<<<<<<<<<
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
 *                 #--------------------------------------------------
 */
          __pyx_t_35 = __pyx_v_bbp;
          __pyx_t_36 = __pyx_v_bbpi;
          if (__pyx_t_35 < 0) __pyx_t_35 += __pyx_pybuffernd_kern.diminfo[0].shape;
          if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_kern.diminfo[1].shape;
          __pyx_t_37 = __pyx_v_b;
          if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_38 = __pyx_v_bp;
          if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_39 = __pyx_v_bbp;
          __pyx_t_40 = __pyx_v_bbpi;
          if (__pyx_t_39 < 0) __pyx_t_39 += __pyx_pybuffernd_kern.diminfo[0].shape;
          if (__pyx_t_40 < 0) __pyx_t_40 += __pyx_pybuffernd_kern.diminfo[1].shape;
          *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_39, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_40, __pyx_pybuffernd_kern.diminfo[1].strides) = (((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_35, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_36, __pyx_pybuffernd_kern.diminfo[1].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_E.diminfo[0].strides))) - (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_E.diminfo[0].strides)));

          /* "qmeq/approach/c_lindblad.pyx":109
 *                 if bbpi_bool:
 *                     kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]             # <<<<<<<<<<<<<<
 *                 #--------------------------------------------------
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
 */
          __pyx_t_41 = __pyx_v_bbpi;
          __pyx_t_42 = __pyx_v_bbp;
          if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_kern.diminfo[0].shape;
          if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_kern.diminfo[1].shape;
          __pyx_t_43 = __pyx_v_bp;
          if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_44 = __pyx_v_b;
          if (__pyx_t_44 < 0) __pyx_t_44 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_45 = __pyx_v_bbpi;
          __pyx_t_46 = __pyx_v_bbp;
          if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_kern.diminfo[0].shape;
          if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_kern.diminfo[1].shape;
          *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_46, __pyx_pybuffernd_kern.diminfo[1].strides) = (((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_kern.diminfo[1].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_43, __pyx_pybuffernd_E.diminfo[0].strides))) - (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_44, __pyx_pybuffernd_E.diminfo[0].strides)));

          /* "qmeq/approach/c_lindblad.pyx":107
 *                 bbpi = ndm0 + bbp - npauli
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 *                 if bbpi_bool:             # <<<<<<<<<<<<<<
 *                     kern[bbp, bbpi] = kern[bbp, bbpi] + E[b]-E[bp]
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
 */
        }

        /* "qmeq/approach/c_lindblad.pyx":111
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
 *                 #--------------------------------------------------
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):             # <<<<<<<<<<<<<<
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                     if aap != -1:
 */
        __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_product); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_47 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_47);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = NULL;
        __pyx_t_20 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_21))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_21);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_21);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_21, function);
            __pyx_t_20 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_21)) {
          PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_1, __pyx_t_47};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_21)) {
          PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_1, __pyx_t_47};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_21, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
        } else
        #endif
        {
          __pyx_t_48 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_48);
          if (__pyx_t_16) {
            __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_48, 0, __pyx_t_16); __pyx_t_16 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_48, 0+__pyx_t_20, __pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_47);
          PyTuple_SET_ITEM(__pyx_t_48, 1+__pyx_t_20, __pyx_t_47);
          __pyx_t_1 = 0;
          __pyx_t_47 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_21, __pyx_t_48, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
        }
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_21 = __pyx_t_2; __Pyx_INCREF(__pyx_t_21); __pyx_t_49 = 0;
          __pyx_t_50 = NULL;
        } else {
          __pyx_t_49 = -1; __pyx_t_21 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_50 = Py_TYPE(__pyx_t_21)->tp_iternext; if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 111, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_50)) {
            if (likely(PyList_CheckExact(__pyx_t_21))) {
              if (__pyx_t_49 >= PyList_GET_SIZE(__pyx_t_21)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_21, __pyx_t_49); __Pyx_INCREF(__pyx_t_2); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 111, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_21, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_49 >= PyTuple_GET_SIZE(__pyx_t_21)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_21, __pyx_t_49); __Pyx_INCREF(__pyx_t_2); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 111, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_21, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_50(__pyx_t_21);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 111, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
            PyObject* sequence = __pyx_t_2;
            #if !CYTHON_COMPILING_IN_PYPY
            Py_ssize_t size = Py_SIZE(sequence);
            #else
            Py_ssize_t size = PySequence_Size(sequence);
            #endif
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 111, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_48 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_47 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_48 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_47 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_48);
            __Pyx_INCREF(__pyx_t_47);
            #else
            __pyx_t_48 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_48);
            __pyx_t_47 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_47);
            #endif
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_24 = Py_TYPE(__pyx_t_1)->tp_iternext;
            index = 0; __pyx_t_48 = __pyx_t_24(__pyx_t_1); if (unlikely(!__pyx_t_48)) goto __pyx_L15_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_48);
            index = 1; __pyx_t_47 = __pyx_t_24(__pyx_t_1); if (unlikely(!__pyx_t_47)) goto __pyx_L15_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_47);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_24(__pyx_t_1), 2) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
            __pyx_t_24 = NULL;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            goto __pyx_L16_unpacking_done;
            __pyx_L15_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_24 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 111, __pyx_L1_error)
            __pyx_L16_unpacking_done:;
          }
          __pyx_t_51 = __Pyx_PyInt_As_npy_int64(__pyx_t_48); if (unlikely((__pyx_t_51 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
          __pyx_t_52 = __Pyx_PyInt_As_npy_int64(__pyx_t_47); if (unlikely((__pyx_t_52 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
          __pyx_v_a = __pyx_t_51;
          __pyx_v_ap = __pyx_t_52;

          /* "qmeq/approach/c_lindblad.pyx":112
 *                 #--------------------------------------------------
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]             # <<<<<<<<<<<<<<
 *                     if aap != -1:
 *                         fct_aap = 0
 */
          __pyx_t_53 = __pyx_v_acharge;
          if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
          __pyx_t_52 = __pyx_v_a;
          if (__pyx_t_52 < 0) __pyx_t_52 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_51 = __pyx_v_ap;
          if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_54 = __pyx_v_acharge;
          if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
          __pyx_t_55 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_51, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_54, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
          if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
          __pyx_v_aap = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_55, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

          /* "qmeq/approach/c_lindblad.pyx":113
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                     if aap != -1:             # <<<<<<<<<<<<<<
 *                         fct_aap = 0
 *                         for l in range(nleads):
 */
          __pyx_t_5 = ((__pyx_v_aap != -1LL) != 0);
          if (__pyx_t_5) {

            /* "qmeq/approach/c_lindblad.pyx":114
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                     if aap != -1:
 *                         fct_aap = 0             # <<<<<<<<<<<<<<
 *                         for l in range(nleads):
 *                             fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 */
            __pyx_v_fct_aap = __pyx_t_double_complex_from_parts(0, 0);

            /* "qmeq/approach/c_lindblad.pyx":115
 *                     if aap != -1:
 *                         fct_aap = 0
 *                         for l in range(nleads):             # <<<<<<<<<<<<<<
 *                             fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                         aapi = ndm0 + aap - npauli
 */
            __pyx_t_56 = __pyx_v_nleads;
            for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
              __pyx_v_l = __pyx_t_57;

              /* "qmeq/approach/c_lindblad.pyx":116
 *                         fct_aap = 0
 *                         for l in range(nleads):
 *                             fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()             # <<<<<<<<<<<<<<
 *                         aapi = ndm0 + aap - npauli
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
 */
              __pyx_t_58 = __pyx_v_l;
              __pyx_t_59 = __pyx_v_b;
              __pyx_t_60 = __pyx_v_a;
              if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_tLba.diminfo[0].shape;
              if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_tLba.diminfo[1].shape;
              if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_tLba.diminfo[2].shape;
              __pyx_t_61 = __pyx_v_l;
              __pyx_t_62 = __pyx_v_bp;
              __pyx_t_63 = __pyx_v_ap;
              if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_tLba.diminfo[0].shape;
              if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_tLba.diminfo[1].shape;
              if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_tLba.diminfo[2].shape;
              __pyx_v_fct_aap = __Pyx_c_sum_double(__pyx_v_fct_aap, __Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_58, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_59, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_60, __pyx_pybuffernd_tLba.diminfo[2].strides)), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_61, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_62, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_63, __pyx_pybuffernd_tLba.diminfo[2].strides)))));
            }

            /* "qmeq/approach/c_lindblad.pyx":117
 *                         for l in range(nleads):
 *                             fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                         aapi = ndm0 + aap - npauli             # <<<<<<<<<<<<<<
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real
 */
            __pyx_v_aapi = ((__pyx_v_ndm0 + __pyx_v_aap) - __pyx_v_npauli);

            /* "qmeq/approach/c_lindblad.pyx":118
 *                             fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                         aapi = ndm0 + aap - npauli
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1             # <<<<<<<<<<<<<<
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real
 *                         if aapi >= ndm0:
 */
            __pyx_t_64 = __pyx_v_acharge;
            if (__pyx_t_64 < 0) __pyx_t_64 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_65 = __pyx_v_a;
            if (__pyx_t_65 < 0) __pyx_t_65 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_66 = __pyx_v_ap;
            if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_67 = __pyx_v_acharge;
            if (__pyx_t_67 < 0) __pyx_t_67 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_68 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_65, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_67, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
            if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
              __pyx_t_56 = 1;
            } else {
              __pyx_t_56 = -1L;
            }
            __pyx_v_aap_sgn = __pyx_t_56;

            /* "qmeq/approach/c_lindblad.pyx":119
 *                         aapi = ndm0 + aap - npauli
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real             # <<<<<<<<<<<<<<
 *                         if aapi >= ndm0:
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 */
            __pyx_t_69 = __pyx_v_bbp;
            __pyx_t_70 = __pyx_v_aap;
            if (__pyx_t_69 < 0) __pyx_t_69 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_70 < 0) __pyx_t_70 += __pyx_pybuffernd_kern.diminfo[1].shape;
            __pyx_t_71 = __pyx_v_bbp;
            __pyx_t_72 = __pyx_v_aap;
            if (__pyx_t_71 < 0) __pyx_t_71 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_72 < 0) __pyx_t_72 += __pyx_pybuffernd_kern.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_72, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_69, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_70, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CREAL(__pyx_v_fct_aap));

            /* "qmeq/approach/c_lindblad.pyx":120
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real
 *                         if aapi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 *                             if bbpi_bool:
 */
            __pyx_t_5 = ((__pyx_v_aapi >= __pyx_v_ndm0) != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":121
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real
 *                         if aapi >= ndm0:
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn             # <<<<<<<<<<<<<<
 *                             if bbpi_bool:
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 */
              __pyx_t_73 = __pyx_v_bbp;
              __pyx_t_74 = __pyx_v_aapi;
              if (__pyx_t_73 < 0) __pyx_t_73 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_74 < 0) __pyx_t_74 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_75 = __pyx_v_bbp;
              __pyx_t_76 = __pyx_v_aapi;
              if (__pyx_t_75 < 0) __pyx_t_75 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_76 < 0) __pyx_t_76 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_75, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_76, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_73, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_74, __pyx_pybuffernd_kern.diminfo[1].strides)) - (__Pyx_CIMAG(__pyx_v_fct_aap) * __pyx_v_aap_sgn));

              /* "qmeq/approach/c_lindblad.pyx":122
 *                         if aapi >= ndm0:
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 *                         if bbpi_bool:
 */
              __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
              if (__pyx_t_5) {

                /* "qmeq/approach/c_lindblad.pyx":123
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 *                             if bbpi_bool:
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn             # <<<<<<<<<<<<<<
 *                         if bbpi_bool:
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag
 */
                __pyx_t_77 = __pyx_v_bbpi;
                __pyx_t_78 = __pyx_v_aapi;
                if (__pyx_t_77 < 0) __pyx_t_77 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_78 < 0) __pyx_t_78 += __pyx_pybuffernd_kern.diminfo[1].shape;
                __pyx_t_79 = __pyx_v_bbpi;
                __pyx_t_80 = __pyx_v_aapi;
                if (__pyx_t_79 < 0) __pyx_t_79 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_80 < 0) __pyx_t_80 += __pyx_pybuffernd_kern.diminfo[1].shape;
                *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_79, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_80, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_77, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_78, __pyx_pybuffernd_kern.diminfo[1].strides)) + (__Pyx_CREAL(__pyx_v_fct_aap) * __pyx_v_aap_sgn));

                /* "qmeq/approach/c_lindblad.pyx":122
 *                         if aapi >= ndm0:
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 *                         if bbpi_bool:
 */
              }

              /* "qmeq/approach/c_lindblad.pyx":120
 *                         aap_sgn = +1 if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else -1
 *                         kern[bbp, aap] = kern[bbp, aap] + fct_aap.real                              # kern[bbp, aap]   += fct_aap.real
 *                         if aapi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, aapi] = kern[bbp, aapi] - fct_aap.imag*aap_sgn                # kern[bbp, aapi]  -= fct_aap.imag*aap_sgn
 *                             if bbpi_bool:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":124
 *                             if bbpi_bool:
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag
 *                 #--------------------------------------------------
 */
            __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":125
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 *                         if bbpi_bool:
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag             # <<<<<<<<<<<<<<
 *                 #--------------------------------------------------
 *                 for bpp in si.statesdm[bcharge]:
 */
              __pyx_t_81 = __pyx_v_bbpi;
              __pyx_t_82 = __pyx_v_aap;
              if (__pyx_t_81 < 0) __pyx_t_81 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_82 < 0) __pyx_t_82 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_83 = __pyx_v_bbpi;
              __pyx_t_84 = __pyx_v_aap;
              if (__pyx_t_83 < 0) __pyx_t_83 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_84 < 0) __pyx_t_84 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_83, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_84, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_81, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_82, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CIMAG(__pyx_v_fct_aap));

              /* "qmeq/approach/c_lindblad.pyx":124
 *                             if bbpi_bool:
 *                                 kern[bbpi, aapi] = kern[bbpi, aapi] + fct_aap.real*aap_sgn          # kern[bbpi, aapi] += fct_aap.real*aap_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag
 *                 #--------------------------------------------------
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":113
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                     if aap != -1:             # <<<<<<<<<<<<<<
 *                         fct_aap = 0
 *                         for l in range(nleads):
 */
          }

          /* "qmeq/approach/c_lindblad.pyx":111
 *                     kern[bbpi, bbp] = kern[bbpi, bbp] + E[bp]-E[b]
 *                 #--------------------------------------------------
 *                 for a, ap in itertools.product(si.statesdm[acharge], si.statesdm[acharge]):             # <<<<<<<<<<<<<<
 *                     aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                     if aap != -1:
 */
        }
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

        /* "qmeq/approach/c_lindblad.pyx":127
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag
 *                 #--------------------------------------------------
 *                 for bpp in si.statesdm[bcharge]:             # <<<<<<<<<<<<<<
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                     if bppbp != -1:
 */
        __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_21, __pyx_v_bcharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_21 = __pyx_t_2; __Pyx_INCREF(__pyx_t_21); __pyx_t_49 = 0;
          __pyx_t_50 = NULL;
        } else {
          __pyx_t_49 = -1; __pyx_t_21 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 127, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_50 = Py_TYPE(__pyx_t_21)->tp_iternext; if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 127, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_50)) {
            if (likely(PyList_CheckExact(__pyx_t_21))) {
              if (__pyx_t_49 >= PyList_GET_SIZE(__pyx_t_21)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_21, __pyx_t_49); __Pyx_INCREF(__pyx_t_2); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_21, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_49 >= PyTuple_GET_SIZE(__pyx_t_21)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_21, __pyx_t_49); __Pyx_INCREF(__pyx_t_2); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_21, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_50(__pyx_t_21);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 127, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __pyx_t_85 = __Pyx_PyInt_As_npy_int64(__pyx_t_2); if (unlikely((__pyx_t_85 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_v_bpp = __pyx_t_85;

          /* "qmeq/approach/c_lindblad.pyx":128
 *                 #--------------------------------------------------
 *                 for bpp in si.statesdm[bcharge]:
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                     if bppbp != -1:
 *                         fct_bppbp = 0
 */
          __pyx_t_86 = __pyx_v_bcharge;
          if (__pyx_t_86 < 0) __pyx_t_86 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
          __pyx_t_85 = __pyx_v_bpp;
          if (__pyx_t_85 < 0) __pyx_t_85 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_87 = __pyx_v_bp;
          if (__pyx_t_87 < 0) __pyx_t_87 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_88 = __pyx_v_bcharge;
          if (__pyx_t_88 < 0) __pyx_t_88 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
          __pyx_t_89 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_86, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_85, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_87, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
          if (__pyx_t_89 < 0) __pyx_t_89 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
          __pyx_v_bppbp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_89, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

          /* "qmeq/approach/c_lindblad.pyx":129
 *                 for bpp in si.statesdm[bcharge]:
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                     if bppbp != -1:             # <<<<<<<<<<<<<<
 *                         fct_bppbp = 0
 *                         for a in si.statesdm[acharge]:
 */
          __pyx_t_5 = ((__pyx_v_bppbp != -1LL) != 0);
          if (__pyx_t_5) {

            /* "qmeq/approach/c_lindblad.pyx":130
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                     if bppbp != -1:
 *                         fct_bppbp = 0             # <<<<<<<<<<<<<<
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):
 */
            __pyx_v_fct_bppbp = __pyx_t_double_complex_from_parts(0, 0);

            /* "qmeq/approach/c_lindblad.pyx":131
 *                     if bppbp != -1:
 *                         fct_bppbp = 0
 *                         for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_47 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 131, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_47);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (likely(PyList_CheckExact(__pyx_t_47)) || PyTuple_CheckExact(__pyx_t_47)) {
              __pyx_t_2 = __pyx_t_47; __Pyx_INCREF(__pyx_t_2); __pyx_t_90 = 0;
              __pyx_t_91 = NULL;
            } else {
              __pyx_t_90 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_47); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_91 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 131, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
            for (;;) {
              if (likely(!__pyx_t_91)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_90 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 131, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 131, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                } else {
                  if (__pyx_t_90 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 131, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 131, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                }
              } else {
                __pyx_t_47 = __pyx_t_91(__pyx_t_2);
                if (unlikely(!__pyx_t_47)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 131, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_47);
              }
              __pyx_t_92 = __Pyx_PyInt_As_npy_int64(__pyx_t_47); if (unlikely((__pyx_t_92 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
              __pyx_v_a = __pyx_t_92;

              /* "qmeq/approach/c_lindblad.pyx":132
 *                         fct_bppbp = 0
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                         for c in si.statesdm[ccharge]:
 */
              __pyx_t_56 = __pyx_v_nleads;
              for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
                __pyx_v_l = __pyx_t_57;

                /* "qmeq/approach/c_lindblad.pyx":133
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]             # <<<<<<<<<<<<<<
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):
 */
                __pyx_t_93 = __pyx_v_l;
                __pyx_t_92 = __pyx_v_a;
                __pyx_t_94 = __pyx_v_b;
                if (__pyx_t_93 < 0) __pyx_t_93 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_92 < 0) __pyx_t_92 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_94 < 0) __pyx_t_94 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_95 = __pyx_v_l;
                __pyx_t_96 = __pyx_v_a;
                __pyx_t_97 = __pyx_v_bpp;
                if (__pyx_t_95 < 0) __pyx_t_95 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_96 < 0) __pyx_t_96 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_97 < 0) __pyx_t_97 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_bppbp = __Pyx_c_sum_double(__pyx_v_fct_bppbp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_92, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_94, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_95, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_96, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_97, __pyx_pybuffernd_tLba.diminfo[2].strides))));
              }

              /* "qmeq/approach/c_lindblad.pyx":131
 *                     if bppbp != -1:
 *                         fct_bppbp = 0
 *                         for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "qmeq/approach/c_lindblad.pyx":134
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                         for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_47 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_ccharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 134, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_47);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (likely(PyList_CheckExact(__pyx_t_47)) || PyTuple_CheckExact(__pyx_t_47)) {
              __pyx_t_2 = __pyx_t_47; __Pyx_INCREF(__pyx_t_2); __pyx_t_90 = 0;
              __pyx_t_91 = NULL;
            } else {
              __pyx_t_90 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_47); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_91 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 134, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
            for (;;) {
              if (likely(!__pyx_t_91)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_90 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 134, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 134, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                } else {
                  if (__pyx_t_90 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 134, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 134, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                }
              } else {
                __pyx_t_47 = __pyx_t_91(__pyx_t_2);
                if (unlikely(!__pyx_t_47)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 134, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_47);
              }
              __pyx_t_98 = __Pyx_PyInt_As_npy_int64(__pyx_t_47); if (unlikely((__pyx_t_98 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 134, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
              __pyx_v_c = __pyx_t_98;

              /* "qmeq/approach/c_lindblad.pyx":135
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                         bppbpi = ndm0 + bppbp - npauli
 */
              __pyx_t_56 = __pyx_v_nleads;
              for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
                __pyx_v_l = __pyx_t_57;

                /* "qmeq/approach/c_lindblad.pyx":136
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]             # <<<<<<<<<<<<<<
 *                         bppbpi = ndm0 + bppbp - npauli
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
 */
                __pyx_t_99 = __pyx_v_l;
                __pyx_t_98 = __pyx_v_c;
                __pyx_t_100 = __pyx_v_b;
                if (__pyx_t_99 < 0) __pyx_t_99 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_98 < 0) __pyx_t_98 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_100 < 0) __pyx_t_100 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_101 = __pyx_v_l;
                __pyx_t_102 = __pyx_v_c;
                __pyx_t_103 = __pyx_v_bpp;
                if (__pyx_t_101 < 0) __pyx_t_101 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_102 < 0) __pyx_t_102 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_103 < 0) __pyx_t_103 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_bppbp = __Pyx_c_sum_double(__pyx_v_fct_bppbp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_99, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_98, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_100, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_101, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_102, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_103, __pyx_pybuffernd_tLba.diminfo[2].strides))));
              }

              /* "qmeq/approach/c_lindblad.pyx":134
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                         for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "qmeq/approach/c_lindblad.pyx":137
 *                             for l in range(nleads):
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                         bppbpi = ndm0 + bppbp - npauli             # <<<<<<<<<<<<<<
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real
 */
            __pyx_v_bppbpi = ((__pyx_v_ndm0 + __pyx_v_bppbp) - __pyx_v_npauli);

            /* "qmeq/approach/c_lindblad.pyx":138
 *                                 fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                         bppbpi = ndm0 + bppbp - npauli
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1             # <<<<<<<<<<<<<<
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real
 *                         if bppbpi >= ndm0:
 */
            __pyx_t_104 = __pyx_v_bcharge;
            if (__pyx_t_104 < 0) __pyx_t_104 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_105 = __pyx_v_bpp;
            if (__pyx_t_105 < 0) __pyx_t_105 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_106 = __pyx_v_bp;
            if (__pyx_t_106 < 0) __pyx_t_106 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_107 = __pyx_v_bcharge;
            if (__pyx_t_107 < 0) __pyx_t_107 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_108 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_104, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_105, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_106, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_107, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_108 < 0) __pyx_t_108 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
            if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_108, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
              __pyx_t_56 = 1;
            } else {
              __pyx_t_56 = -1L;
            }
            __pyx_v_bppbp_sgn = __pyx_t_56;

            /* "qmeq/approach/c_lindblad.pyx":139
 *                         bppbpi = ndm0 + bppbp - npauli
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real             # <<<<<<<<<<<<<<
 *                         if bppbpi >= ndm0:
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 */
            __pyx_t_109 = __pyx_v_bbp;
            __pyx_t_110 = __pyx_v_bppbp;
            if (__pyx_t_109 < 0) __pyx_t_109 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_110 < 0) __pyx_t_110 += __pyx_pybuffernd_kern.diminfo[1].shape;
            __pyx_t_111 = __pyx_v_bbp;
            __pyx_t_112 = __pyx_v_bppbp;
            if (__pyx_t_111 < 0) __pyx_t_111 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_112 < 0) __pyx_t_112 += __pyx_pybuffernd_kern.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_111, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_112, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_109, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_110, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CREAL(__pyx_v_fct_bppbp));

            /* "qmeq/approach/c_lindblad.pyx":140
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real
 *                         if bppbpi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 *                             if bbpi_bool:
 */
            __pyx_t_5 = ((__pyx_v_bppbpi >= __pyx_v_ndm0) != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":141
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real
 *                         if bppbpi >= ndm0:
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn             # <<<<<<<<<<<<<<
 *                             if bbpi_bool:
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 */
              __pyx_t_113 = __pyx_v_bbp;
              __pyx_t_114 = __pyx_v_bppbpi;
              if (__pyx_t_113 < 0) __pyx_t_113 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_114 < 0) __pyx_t_114 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_115 = __pyx_v_bbp;
              __pyx_t_116 = __pyx_v_bppbpi;
              if (__pyx_t_115 < 0) __pyx_t_115 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_116 < 0) __pyx_t_116 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_115, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_116, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_113, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_114, __pyx_pybuffernd_kern.diminfo[1].strides)) - (__Pyx_CIMAG(__pyx_v_fct_bppbp) * __pyx_v_bppbp_sgn));

              /* "qmeq/approach/c_lindblad.pyx":142
 *                         if bppbpi >= ndm0:
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 *                         if bbpi_bool:
 */
              __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
              if (__pyx_t_5) {

                /* "qmeq/approach/c_lindblad.pyx":143
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 *                             if bbpi_bool:
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn             # <<<<<<<<<<<<<<
 *                         if bbpi_bool:
 *                             kern[bbpi, bppbp] = kern[bbpi, bppbp] + fct_bppbp.imag                  # kern[bbpi, bppbp] += fct_bppbp.imag
 */
                __pyx_t_117 = __pyx_v_bbpi;
                __pyx_t_118 = __pyx_v_bppbpi;
                if (__pyx_t_117 < 0) __pyx_t_117 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_118 < 0) __pyx_t_118 += __pyx_pybuffernd_kern.diminfo[1].shape;
                __pyx_t_119 = __pyx_v_bbpi;
                __pyx_t_120 = __pyx_v_bppbpi;
                if (__pyx_t_119 < 0) __pyx_t_119 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_120 < 0) __pyx_t_120 += __pyx_pybuffernd_kern.diminfo[1].shape;
                *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_119, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_120, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_117, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_118, __pyx_pybuffernd_kern.diminfo[1].strides)) + (__Pyx_CREAL(__pyx_v_fct_bppbp) * __pyx_v_bppbp_sgn));

                /* "qmeq/approach/c_lindblad.pyx":142
 *                         if bppbpi >= ndm0:
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 *                         if bbpi_bool:
 */
              }

              /* "qmeq/approach/c_lindblad.pyx":140
 *                         bppbp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bppbp] = kern[bbp, bppbp] + fct_bppbp.real                        # kern[bbp, bppbp] += fct_bppbp.real
 *                         if bppbpi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, bppbpi] = kern[bbp, bppbpi] - fct_bppbp.imag*bppbp_sgn        # kern[bbp, bppbpi] -= fct_bppbp.imag*bppbp_sgn
 *                             if bbpi_bool:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":144
 *                             if bbpi_bool:
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, bppbp] = kern[bbpi, bppbp] + fct_bppbp.imag                  # kern[bbpi, bppbp] += fct_bppbp.imag
 *                     #--------------------------------------------------
 */
            __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":145
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 *                         if bbpi_bool:
 *                             kern[bbpi, bppbp] = kern[bbpi, bppbp] + fct_bppbp.imag                  # kern[bbpi, bppbp] += fct_bppbp.imag             # <<<<<<<<<<<<<<
 *                     #--------------------------------------------------
 *                     bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 */
              __pyx_t_121 = __pyx_v_bbpi;
              __pyx_t_122 = __pyx_v_bppbp;
              if (__pyx_t_121 < 0) __pyx_t_121 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_122 < 0) __pyx_t_122 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_123 = __pyx_v_bbpi;
              __pyx_t_124 = __pyx_v_bppbp;
              if (__pyx_t_123 < 0) __pyx_t_123 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_124 < 0) __pyx_t_124 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_123, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_124, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_121, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_122, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CIMAG(__pyx_v_fct_bppbp));

              /* "qmeq/approach/c_lindblad.pyx":144
 *                             if bbpi_bool:
 *                                 kern[bbpi, bppbpi] = kern[bbpi, bppbpi] + fct_bppbp.real*bppbp_sgn  # kern[bbpi, bppbpi] += fct_bppbp.real*bppbp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, bppbp] = kern[bbpi, bppbp] + fct_bppbp.imag                  # kern[bbpi, bppbp] += fct_bppbp.imag
 *                     #--------------------------------------------------
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":129
 *                 for bpp in si.statesdm[bcharge]:
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                     if bppbp != -1:             # <<<<<<<<<<<<<<
 *                         fct_bppbp = 0
 *                         for a in si.statesdm[acharge]:
 */
          }

          /* "qmeq/approach/c_lindblad.pyx":147
 *                             kern[bbpi, bppbp] = kern[bbpi, bppbp] + fct_bppbp.imag                  # kern[bbpi, bppbp] += fct_bppbp.imag
 *                     #--------------------------------------------------
 *                     bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                     if bbpp != -1:
 *                         fct_bbpp = 0
 */
          __pyx_t_125 = __pyx_v_bcharge;
          if (__pyx_t_125 < 0) __pyx_t_125 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
          __pyx_t_126 = __pyx_v_b;
          if (__pyx_t_126 < 0) __pyx_t_126 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_127 = __pyx_v_bpp;
          if (__pyx_t_127 < 0) __pyx_t_127 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_128 = __pyx_v_bcharge;
          if (__pyx_t_128 < 0) __pyx_t_128 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
          __pyx_t_129 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_125, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_126, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_127, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_128, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
          if (__pyx_t_129 < 0) __pyx_t_129 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
          __pyx_v_bbpp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_129, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

          /* "qmeq/approach/c_lindblad.pyx":148
 *                     #--------------------------------------------------
 *                     bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                     if bbpp != -1:             # <<<<<<<<<<<<<<
 *                         fct_bbpp = 0
 *                         for a in si.statesdm[acharge]:
 */
          __pyx_t_5 = ((__pyx_v_bbpp != -1LL) != 0);
          if (__pyx_t_5) {

            /* "qmeq/approach/c_lindblad.pyx":149
 *                     bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                     if bbpp != -1:
 *                         fct_bbpp = 0             # <<<<<<<<<<<<<<
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):
 */
            __pyx_v_fct_bbpp = __pyx_t_double_complex_from_parts(0, 0);

            /* "qmeq/approach/c_lindblad.pyx":150
 *                     if bbpp != -1:
 *                         fct_bbpp = 0
 *                         for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_47 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 150, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_47);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (likely(PyList_CheckExact(__pyx_t_47)) || PyTuple_CheckExact(__pyx_t_47)) {
              __pyx_t_2 = __pyx_t_47; __Pyx_INCREF(__pyx_t_2); __pyx_t_90 = 0;
              __pyx_t_91 = NULL;
            } else {
              __pyx_t_90 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_47); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_91 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 150, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
            for (;;) {
              if (likely(!__pyx_t_91)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_90 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 150, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 150, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                } else {
                  if (__pyx_t_90 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 150, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 150, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                }
              } else {
                __pyx_t_47 = __pyx_t_91(__pyx_t_2);
                if (unlikely(!__pyx_t_47)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 150, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_47);
              }
              __pyx_t_130 = __Pyx_PyInt_As_npy_int64(__pyx_t_47); if (unlikely((__pyx_t_130 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 150, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
              __pyx_v_a = __pyx_t_130;

              /* "qmeq/approach/c_lindblad.pyx":151
 *                         fct_bbpp = 0
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                         for c in si.statesdm[ccharge]:
 */
              __pyx_t_56 = __pyx_v_nleads;
              for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
                __pyx_v_l = __pyx_t_57;

                /* "qmeq/approach/c_lindblad.pyx":152
 *                         for a in si.statesdm[acharge]:
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]             # <<<<<<<<<<<<<<
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):
 */
                __pyx_t_131 = __pyx_v_l;
                __pyx_t_130 = __pyx_v_a;
                __pyx_t_132 = __pyx_v_bpp;
                if (__pyx_t_131 < 0) __pyx_t_131 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_130 < 0) __pyx_t_130 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_132 < 0) __pyx_t_132 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_133 = __pyx_v_l;
                __pyx_t_134 = __pyx_v_a;
                __pyx_t_135 = __pyx_v_bp;
                if (__pyx_t_133 < 0) __pyx_t_133 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_134 < 0) __pyx_t_134 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_135 < 0) __pyx_t_135 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_bbpp = __Pyx_c_sum_double(__pyx_v_fct_bbpp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_131, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_130, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_132, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_133, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_134, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_135, __pyx_pybuffernd_tLba.diminfo[2].strides))));
              }

              /* "qmeq/approach/c_lindblad.pyx":150
 *                     if bbpp != -1:
 *                         fct_bbpp = 0
 *                         for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "qmeq/approach/c_lindblad.pyx":153
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                         for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 */
            __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_47 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_ccharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 153, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_47);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (likely(PyList_CheckExact(__pyx_t_47)) || PyTuple_CheckExact(__pyx_t_47)) {
              __pyx_t_2 = __pyx_t_47; __Pyx_INCREF(__pyx_t_2); __pyx_t_90 = 0;
              __pyx_t_91 = NULL;
            } else {
              __pyx_t_90 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_47); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_91 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_91)) __PYX_ERR(0, 153, __pyx_L1_error)
            }
            __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
            for (;;) {
              if (likely(!__pyx_t_91)) {
                if (likely(PyList_CheckExact(__pyx_t_2))) {
                  if (__pyx_t_90 >= PyList_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 153, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 153, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                } else {
                  if (__pyx_t_90 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
                  #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                  __pyx_t_47 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_90); __Pyx_INCREF(__pyx_t_47); __pyx_t_90++; if (unlikely(0 < 0)) __PYX_ERR(0, 153, __pyx_L1_error)
                  #else
                  __pyx_t_47 = PySequence_ITEM(__pyx_t_2, __pyx_t_90); __pyx_t_90++; if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 153, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_47);
                  #endif
                }
              } else {
                __pyx_t_47 = __pyx_t_91(__pyx_t_2);
                if (unlikely(!__pyx_t_47)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 153, __pyx_L1_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_47);
              }
              __pyx_t_136 = __Pyx_PyInt_As_npy_int64(__pyx_t_47); if (unlikely((__pyx_t_136 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 153, __pyx_L1_error)
              __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
              __pyx_v_c = __pyx_t_136;

              /* "qmeq/approach/c_lindblad.pyx":154
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                         bbppi = ndm0 + bbpp - npauli
 */
              __pyx_t_56 = __pyx_v_nleads;
              for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
                __pyx_v_l = __pyx_t_57;

                /* "qmeq/approach/c_lindblad.pyx":155
 *                         for c in si.statesdm[ccharge]:
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]             # <<<<<<<<<<<<<<
 *                         bbppi = ndm0 + bbpp - npauli
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
 */
                __pyx_t_137 = __pyx_v_l;
                __pyx_t_136 = __pyx_v_c;
                __pyx_t_138 = __pyx_v_bpp;
                if (__pyx_t_137 < 0) __pyx_t_137 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_136 < 0) __pyx_t_136 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_138 < 0) __pyx_t_138 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_139 = __pyx_v_l;
                __pyx_t_140 = __pyx_v_c;
                __pyx_t_141 = __pyx_v_bp;
                if (__pyx_t_139 < 0) __pyx_t_139 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_140 < 0) __pyx_t_140 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_141 < 0) __pyx_t_141 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_bbpp = __Pyx_c_sum_double(__pyx_v_fct_bbpp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_137, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_136, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_138, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_139, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_140, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_141, __pyx_pybuffernd_tLba.diminfo[2].strides))));
              }

              /* "qmeq/approach/c_lindblad.pyx":153
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                         for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "qmeq/approach/c_lindblad.pyx":156
 *                             for l in range(nleads):
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                         bbppi = ndm0 + bbpp - npauli             # <<<<<<<<<<<<<<
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real
 */
            __pyx_v_bbppi = ((__pyx_v_ndm0 + __pyx_v_bbpp) - __pyx_v_npauli);

            /* "qmeq/approach/c_lindblad.pyx":157
 *                                 fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                         bbppi = ndm0 + bbpp - npauli
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1             # <<<<<<<<<<<<<<
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real
 *                         if bbppi >= ndm0:
 */
            __pyx_t_142 = __pyx_v_bcharge;
            if (__pyx_t_142 < 0) __pyx_t_142 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_143 = __pyx_v_b;
            if (__pyx_t_143 < 0) __pyx_t_143 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_144 = __pyx_v_bpp;
            if (__pyx_t_144 < 0) __pyx_t_144 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_145 = __pyx_v_bcharge;
            if (__pyx_t_145 < 0) __pyx_t_145 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_146 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_142, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_143, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_144, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_145, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_146 < 0) __pyx_t_146 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
            if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_146, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
              __pyx_t_56 = 1;
            } else {
              __pyx_t_56 = -1L;
            }
            __pyx_v_bbpp_sgn = __pyx_t_56;

            /* "qmeq/approach/c_lindblad.pyx":158
 *                         bbppi = ndm0 + bbpp - npauli
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real             # <<<<<<<<<<<<<<
 *                         if bbppi >= ndm0:
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 */
            __pyx_t_147 = __pyx_v_bbp;
            __pyx_t_148 = __pyx_v_bbpp;
            if (__pyx_t_147 < 0) __pyx_t_147 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_148 < 0) __pyx_t_148 += __pyx_pybuffernd_kern.diminfo[1].shape;
            __pyx_t_149 = __pyx_v_bbp;
            __pyx_t_150 = __pyx_v_bbpp;
            if (__pyx_t_149 < 0) __pyx_t_149 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_150 < 0) __pyx_t_150 += __pyx_pybuffernd_kern.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_149, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_150, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_147, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_148, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CREAL(__pyx_v_fct_bbpp));

            /* "qmeq/approach/c_lindblad.pyx":159
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real
 *                         if bbppi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 *                             if bbpi_bool:
 */
            __pyx_t_5 = ((__pyx_v_bbppi >= __pyx_v_ndm0) != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":160
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real
 *                         if bbppi >= ndm0:
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn             # <<<<<<<<<<<<<<
 *                             if bbpi_bool:
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 */
              __pyx_t_151 = __pyx_v_bbp;
              __pyx_t_152 = __pyx_v_bbppi;
              if (__pyx_t_151 < 0) __pyx_t_151 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_152 < 0) __pyx_t_152 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_153 = __pyx_v_bbp;
              __pyx_t_154 = __pyx_v_bbppi;
              if (__pyx_t_153 < 0) __pyx_t_153 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_154 < 0) __pyx_t_154 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_153, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_154, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_151, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_152, __pyx_pybuffernd_kern.diminfo[1].strides)) - (__Pyx_CIMAG(__pyx_v_fct_bbpp) * __pyx_v_bbpp_sgn));

              /* "qmeq/approach/c_lindblad.pyx":161
 *                         if bbppi >= ndm0:
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 *                         if bbpi_bool:
 */
              __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
              if (__pyx_t_5) {

                /* "qmeq/approach/c_lindblad.pyx":162
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 *                             if bbpi_bool:
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn             # <<<<<<<<<<<<<<
 *                         if bbpi_bool:
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag
 */
                __pyx_t_155 = __pyx_v_bbpi;
                __pyx_t_156 = __pyx_v_bbppi;
                if (__pyx_t_155 < 0) __pyx_t_155 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_156 < 0) __pyx_t_156 += __pyx_pybuffernd_kern.diminfo[1].shape;
                __pyx_t_157 = __pyx_v_bbpi;
                __pyx_t_158 = __pyx_v_bbppi;
                if (__pyx_t_157 < 0) __pyx_t_157 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_158 < 0) __pyx_t_158 += __pyx_pybuffernd_kern.diminfo[1].shape;
                *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_157, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_158, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_155, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_156, __pyx_pybuffernd_kern.diminfo[1].strides)) + (__Pyx_CREAL(__pyx_v_fct_bbpp) * __pyx_v_bbpp_sgn));

                /* "qmeq/approach/c_lindblad.pyx":161
 *                         if bbppi >= ndm0:
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 *                         if bbpi_bool:
 */
              }

              /* "qmeq/approach/c_lindblad.pyx":159
 *                         bbpp_sgn = +1 if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else -1
 *                         kern[bbp, bbpp] = kern[bbp, bbpp] + fct_bbpp.real                           # kern[bbp, bbpp] += fct_bbpp.real
 *                         if bbppi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, bbppi] = kern[bbp, bbppi] - fct_bbpp.imag*bbpp_sgn            # kern[bbp, bbppi] -= fct_bbpp.imag*bbpp_sgn
 *                             if bbpi_bool:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":163
 *                             if bbpi_bool:
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag
 *                 #--------------------------------------------------
 */
            __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":164
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 *                         if bbpi_bool:
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag             # <<<<<<<<<<<<<<
 *                 #--------------------------------------------------
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
 */
              __pyx_t_159 = __pyx_v_bbpi;
              __pyx_t_160 = __pyx_v_bbpp;
              if (__pyx_t_159 < 0) __pyx_t_159 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_160 < 0) __pyx_t_160 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_161 = __pyx_v_bbpi;
              __pyx_t_162 = __pyx_v_bbpp;
              if (__pyx_t_161 < 0) __pyx_t_161 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_162 < 0) __pyx_t_162 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_161, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_162, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_159, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_160, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CIMAG(__pyx_v_fct_bbpp));

              /* "qmeq/approach/c_lindblad.pyx":163
 *                             if bbpi_bool:
 *                                 kern[bbpi, bbppi] = kern[bbpi, bbppi] + fct_bbpp.real*bbpp_sgn      # kern[bbpi, bbppi] += fct_bbpp.real*bbpp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag
 *                 #--------------------------------------------------
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":148
 *                     #--------------------------------------------------
 *                     bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                     if bbpp != -1:             # <<<<<<<<<<<<<<
 *                         fct_bbpp = 0
 *                         for a in si.statesdm[acharge]:
 */
          }

          /* "qmeq/approach/c_lindblad.pyx":127
 *                             kern[bbpi, aap] = kern[bbpi, aap] + fct_aap.imag                        # kern[bbpi, aap]  += fct_aap.imag
 *                 #--------------------------------------------------
 *                 for bpp in si.statesdm[bcharge]:             # <<<<<<<<<<<<<<
 *                     bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                     if bppbp != -1:
 */
        }
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;

        /* "qmeq/approach/c_lindblad.pyx":166
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag
 *                 #--------------------------------------------------
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):             # <<<<<<<<<<<<<<
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                     if ccp != -1:
 */
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_47 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_product); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_47);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_48 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_ccharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_48);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_2, __pyx_v_ccharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        __pyx_t_20 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_47))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_47);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_47);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_47, function);
            __pyx_t_20 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_47)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_48, __pyx_t_1};
          __pyx_t_21 = __Pyx_PyFunction_FastCall(__pyx_t_47, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_47)) {
          PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_48, __pyx_t_1};
          __pyx_t_21 = __Pyx_PyCFunction_FastCall(__pyx_t_47, __pyx_temp+1-__pyx_t_20, 2+__pyx_t_20); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else
        #endif
        {
          __pyx_t_16 = PyTuple_New(2+__pyx_t_20); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (__pyx_t_2) {
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_2); __pyx_t_2 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_48);
          PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_20, __pyx_t_48);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_20, __pyx_t_1);
          __pyx_t_48 = 0;
          __pyx_t_1 = 0;
          __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_16, NULL); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        }
        __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
        if (likely(PyList_CheckExact(__pyx_t_21)) || PyTuple_CheckExact(__pyx_t_21)) {
          __pyx_t_47 = __pyx_t_21; __Pyx_INCREF(__pyx_t_47); __pyx_t_49 = 0;
          __pyx_t_50 = NULL;
        } else {
          __pyx_t_49 = -1; __pyx_t_47 = PyObject_GetIter(__pyx_t_21); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_47);
          __pyx_t_50 = Py_TYPE(__pyx_t_47)->tp_iternext; if (unlikely(!__pyx_t_50)) __PYX_ERR(0, 166, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
        for (;;) {
          if (likely(!__pyx_t_50)) {
            if (likely(PyList_CheckExact(__pyx_t_47))) {
              if (__pyx_t_49 >= PyList_GET_SIZE(__pyx_t_47)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_21 = PyList_GET_ITEM(__pyx_t_47, __pyx_t_49); __Pyx_INCREF(__pyx_t_21); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
              #else
              __pyx_t_21 = PySequence_ITEM(__pyx_t_47, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 166, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_21);
              #endif
            } else {
              if (__pyx_t_49 >= PyTuple_GET_SIZE(__pyx_t_47)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_21 = PyTuple_GET_ITEM(__pyx_t_47, __pyx_t_49); __Pyx_INCREF(__pyx_t_21); __pyx_t_49++; if (unlikely(0 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
              #else
              __pyx_t_21 = PySequence_ITEM(__pyx_t_47, __pyx_t_49); __pyx_t_49++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 166, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_21);
              #endif
            }
          } else {
            __pyx_t_21 = __pyx_t_50(__pyx_t_47);
            if (unlikely(!__pyx_t_21)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 166, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_21);
          }
          if ((likely(PyTuple_CheckExact(__pyx_t_21))) || (PyList_CheckExact(__pyx_t_21))) {
            PyObject* sequence = __pyx_t_21;
            #if !CYTHON_COMPILING_IN_PYPY
            Py_ssize_t size = Py_SIZE(sequence);
            #else
            Py_ssize_t size = PySequence_Size(sequence);
            #endif
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 166, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_16 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_16 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(__pyx_t_1);
            #else
            __pyx_t_16 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 166, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_48 = PyObject_GetIter(__pyx_t_21); if (unlikely(!__pyx_t_48)) __PYX_ERR(0, 166, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_48);
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_t_24 = Py_TYPE(__pyx_t_48)->tp_iternext;
            index = 0; __pyx_t_16 = __pyx_t_24(__pyx_t_48); if (unlikely(!__pyx_t_16)) goto __pyx_L51_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_16);
            index = 1; __pyx_t_1 = __pyx_t_24(__pyx_t_48); if (unlikely(!__pyx_t_1)) goto __pyx_L51_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_1);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_24(__pyx_t_48), 2) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
            __pyx_t_24 = NULL;
            __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
            goto __pyx_L52_unpacking_done;
            __pyx_L51_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_48); __pyx_t_48 = 0;
            __pyx_t_24 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 166, __pyx_L1_error)
            __pyx_L52_unpacking_done:;
          }
          __pyx_t_163 = __Pyx_PyInt_As_npy_int64(__pyx_t_16); if (unlikely((__pyx_t_163 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_164 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_164 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_c = __pyx_t_163;
          __pyx_v_cp = __pyx_t_164;

          /* "qmeq/approach/c_lindblad.pyx":167
 *                 #--------------------------------------------------
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]             # <<<<<<<<<<<<<<
 *                     if ccp != -1:
 *                         fct_ccp = 0
 */
          __pyx_t_165 = __pyx_v_ccharge;
          if (__pyx_t_165 < 0) __pyx_t_165 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
          __pyx_t_164 = __pyx_v_c;
          if (__pyx_t_164 < 0) __pyx_t_164 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_163 = __pyx_v_cp;
          if (__pyx_t_163 < 0) __pyx_t_163 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
          __pyx_t_166 = __pyx_v_ccharge;
          if (__pyx_t_166 < 0) __pyx_t_166 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
          __pyx_t_167 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_165, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_164, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_163, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_166, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
          if (__pyx_t_167 < 0) __pyx_t_167 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
          __pyx_v_ccp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_167, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

          /* "qmeq/approach/c_lindblad.pyx":168
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                     if ccp != -1:             # <<<<<<<<<<<<<<
 *                         fct_ccp = 0
 *                         for l in range(nleads):
 */
          __pyx_t_5 = ((__pyx_v_ccp != -1LL) != 0);
          if (__pyx_t_5) {

            /* "qmeq/approach/c_lindblad.pyx":169
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                     if ccp != -1:
 *                         fct_ccp = 0             # <<<<<<<<<<<<<<
 *                         for l in range(nleads):
 *                             fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 */
            __pyx_v_fct_ccp = __pyx_t_double_complex_from_parts(0, 0);

            /* "qmeq/approach/c_lindblad.pyx":170
 *                     if ccp != -1:
 *                         fct_ccp = 0
 *                         for l in range(nleads):             # <<<<<<<<<<<<<<
 *                             fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                         ccpi = ndm0 + ccp - npauli
 */
            __pyx_t_56 = __pyx_v_nleads;
            for (__pyx_t_57 = 0; __pyx_t_57 < __pyx_t_56; __pyx_t_57+=1) {
              __pyx_v_l = __pyx_t_57;

              /* "qmeq/approach/c_lindblad.pyx":171
 *                         fct_ccp = 0
 *                         for l in range(nleads):
 *                             fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()             # <<<<<<<<<<<<<<
 *                         ccpi = ndm0 + ccp - npauli
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
 */
              __pyx_t_168 = __pyx_v_l;
              __pyx_t_169 = __pyx_v_b;
              __pyx_t_170 = __pyx_v_c;
              if (__pyx_t_168 < 0) __pyx_t_168 += __pyx_pybuffernd_tLba.diminfo[0].shape;
              if (__pyx_t_169 < 0) __pyx_t_169 += __pyx_pybuffernd_tLba.diminfo[1].shape;
              if (__pyx_t_170 < 0) __pyx_t_170 += __pyx_pybuffernd_tLba.diminfo[2].shape;
              __pyx_t_171 = __pyx_v_l;
              __pyx_t_172 = __pyx_v_bp;
              __pyx_t_173 = __pyx_v_cp;
              if (__pyx_t_171 < 0) __pyx_t_171 += __pyx_pybuffernd_tLba.diminfo[0].shape;
              if (__pyx_t_172 < 0) __pyx_t_172 += __pyx_pybuffernd_tLba.diminfo[1].shape;
              if (__pyx_t_173 < 0) __pyx_t_173 += __pyx_pybuffernd_tLba.diminfo[2].shape;
              __pyx_v_fct_ccp = __Pyx_c_sum_double(__pyx_v_fct_ccp, __Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_168, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_169, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_170, __pyx_pybuffernd_tLba.diminfo[2].strides)), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_171, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_172, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_173, __pyx_pybuffernd_tLba.diminfo[2].strides)))));
            }

            /* "qmeq/approach/c_lindblad.pyx":172
 *                         for l in range(nleads):
 *                             fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                         ccpi = ndm0 + ccp - npauli             # <<<<<<<<<<<<<<
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real
 */
            __pyx_v_ccpi = ((__pyx_v_ndm0 + __pyx_v_ccp) - __pyx_v_npauli);

            /* "qmeq/approach/c_lindblad.pyx":173
 *                             fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                         ccpi = ndm0 + ccp - npauli
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1             # <<<<<<<<<<<<<<
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real
 *                         if ccpi >= ndm0:
 */
            __pyx_t_174 = __pyx_v_ccharge;
            if (__pyx_t_174 < 0) __pyx_t_174 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_175 = __pyx_v_c;
            if (__pyx_t_175 < 0) __pyx_t_175 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_176 = __pyx_v_cp;
            if (__pyx_t_176 < 0) __pyx_t_176 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_177 = __pyx_v_ccharge;
            if (__pyx_t_177 < 0) __pyx_t_177 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_178 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_174, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_175, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_176, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_177, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_178 < 0) __pyx_t_178 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
            if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_178, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
              __pyx_t_56 = 1;
            } else {
              __pyx_t_56 = -1L;
            }
            __pyx_v_ccp_sgn = __pyx_t_56;

            /* "qmeq/approach/c_lindblad.pyx":174
 *                         ccpi = ndm0 + ccp - npauli
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real             # <<<<<<<<<<<<<<
 *                         if ccpi >= ndm0:
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 */
            __pyx_t_179 = __pyx_v_bbp;
            __pyx_t_180 = __pyx_v_ccp;
            if (__pyx_t_179 < 0) __pyx_t_179 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_180 < 0) __pyx_t_180 += __pyx_pybuffernd_kern.diminfo[1].shape;
            __pyx_t_181 = __pyx_v_bbp;
            __pyx_t_182 = __pyx_v_ccp;
            if (__pyx_t_181 < 0) __pyx_t_181 += __pyx_pybuffernd_kern.diminfo[0].shape;
            if (__pyx_t_182 < 0) __pyx_t_182 += __pyx_pybuffernd_kern.diminfo[1].shape;
            *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_181, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_182, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_179, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_180, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CREAL(__pyx_v_fct_ccp));

            /* "qmeq/approach/c_lindblad.pyx":175
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real
 *                         if ccpi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 *                             if bbpi_bool:
 */
            __pyx_t_5 = ((__pyx_v_ccpi >= __pyx_v_ndm0) != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":176
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real
 *                         if ccpi >= ndm0:
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn             # <<<<<<<<<<<<<<
 *                             if bbpi_bool:
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 */
              __pyx_t_183 = __pyx_v_bbp;
              __pyx_t_184 = __pyx_v_ccpi;
              if (__pyx_t_183 < 0) __pyx_t_183 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_184 < 0) __pyx_t_184 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_185 = __pyx_v_bbp;
              __pyx_t_186 = __pyx_v_ccpi;
              if (__pyx_t_185 < 0) __pyx_t_185 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_186 < 0) __pyx_t_186 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_185, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_186, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_183, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_184, __pyx_pybuffernd_kern.diminfo[1].strides)) - (__Pyx_CIMAG(__pyx_v_fct_ccp) * __pyx_v_ccp_sgn));

              /* "qmeq/approach/c_lindblad.pyx":177
 *                         if ccpi >= ndm0:
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 *                         if bbpi_bool:
 */
              __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
              if (__pyx_t_5) {

                /* "qmeq/approach/c_lindblad.pyx":178
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 *                             if bbpi_bool:
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn             # <<<<<<<<<<<<<<
 *                         if bbpi_bool:
 *                             kern[bbpi, ccp] = kern[bbpi, ccp] + fct_ccp.imag                        # kern[bbpi, ccp] += fct_ccp.imag
 */
                __pyx_t_187 = __pyx_v_bbpi;
                __pyx_t_188 = __pyx_v_ccpi;
                if (__pyx_t_187 < 0) __pyx_t_187 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_188 < 0) __pyx_t_188 += __pyx_pybuffernd_kern.diminfo[1].shape;
                __pyx_t_189 = __pyx_v_bbpi;
                __pyx_t_190 = __pyx_v_ccpi;
                if (__pyx_t_189 < 0) __pyx_t_189 += __pyx_pybuffernd_kern.diminfo[0].shape;
                if (__pyx_t_190 < 0) __pyx_t_190 += __pyx_pybuffernd_kern.diminfo[1].shape;
                *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_189, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_190, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_187, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_188, __pyx_pybuffernd_kern.diminfo[1].strides)) + (__Pyx_CREAL(__pyx_v_fct_ccp) * __pyx_v_ccp_sgn));

                /* "qmeq/approach/c_lindblad.pyx":177
 *                         if ccpi >= ndm0:
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 *                             if bbpi_bool:             # <<<<<<<<<<<<<<
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 *                         if bbpi_bool:
 */
              }

              /* "qmeq/approach/c_lindblad.pyx":175
 *                         ccp_sgn = +1 if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else -1
 *                         kern[bbp, ccp] = kern[bbp, ccp] + fct_ccp.real                              # kern[bbp, ccp] += fct_ccp.real
 *                         if ccpi >= ndm0:             # <<<<<<<<<<<<<<
 *                             kern[bbp, ccpi] = kern[bbp, ccpi] - fct_ccp.imag*ccp_sgn                # kern[bbp, ccpi] -= fct_ccp.imag*ccp_sgn
 *                             if bbpi_bool:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":179
 *                             if bbpi_bool:
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, ccp] = kern[bbpi, ccp] + fct_ccp.imag                        # kern[bbpi, ccp] += fct_ccp.imag
 *                 #--------------------------------------------------
 */
            __pyx_t_5 = (__pyx_v_bbpi_bool != 0);
            if (__pyx_t_5) {

              /* "qmeq/approach/c_lindblad.pyx":180
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 *                         if bbpi_bool:
 *                             kern[bbpi, ccp] = kern[bbpi, ccp] + fct_ccp.imag                        # kern[bbpi, ccp] += fct_ccp.imag             # <<<<<<<<<<<<<<
 *                 #--------------------------------------------------
 *     # Normalisation condition
 */
              __pyx_t_191 = __pyx_v_bbpi;
              __pyx_t_192 = __pyx_v_ccp;
              if (__pyx_t_191 < 0) __pyx_t_191 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_192 < 0) __pyx_t_192 += __pyx_pybuffernd_kern.diminfo[1].shape;
              __pyx_t_193 = __pyx_v_bbpi;
              __pyx_t_194 = __pyx_v_ccp;
              if (__pyx_t_193 < 0) __pyx_t_193 += __pyx_pybuffernd_kern.diminfo[0].shape;
              if (__pyx_t_194 < 0) __pyx_t_194 += __pyx_pybuffernd_kern.diminfo[1].shape;
              *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_193, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_194, __pyx_pybuffernd_kern.diminfo[1].strides) = ((*__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_191, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_192, __pyx_pybuffernd_kern.diminfo[1].strides)) + __Pyx_CIMAG(__pyx_v_fct_ccp));

              /* "qmeq/approach/c_lindblad.pyx":179
 *                             if bbpi_bool:
 *                                 kern[bbpi, ccpi] = kern[bbpi, ccpi] + fct_ccp.real*ccp_sgn          # kern[bbpi, ccpi] += fct_ccp.real*ccp_sgn
 *                         if bbpi_bool:             # <<<<<<<<<<<<<<
 *                             kern[bbpi, ccp] = kern[bbpi, ccp] + fct_ccp.imag                        # kern[bbpi, ccp] += fct_ccp.imag
 *                 #--------------------------------------------------
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":168
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                     if ccp != -1:             # <<<<<<<<<<<<<<
 *                         fct_ccp = 0
 *                         for l in range(nleads):
 */
          }

          /* "qmeq/approach/c_lindblad.pyx":166
 *                             kern[bbpi, bbpp] = kern[bbpi, bbpp] + fct_bbpp.imag                     # kern[bbpi, bbpp] += fct_bbpp.imag
 *                 #--------------------------------------------------
 *                 for c, cp in itertools.product(si.statesdm[ccharge], si.statesdm[ccharge]):             # <<<<<<<<<<<<<<
 *                     ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                     if ccp != -1:
 */
        }
        __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;

        /* "qmeq/approach/c_lindblad.pyx":104
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1 and bbp_bool:             # <<<<<<<<<<<<<<
 *                 bbpi = ndm0 + bbp - npauli
 *                 bbpi_bool = True if bbpi >= ndm0 else False
 */
      }

      /* "qmeq/approach/c_lindblad.pyx":101
 *         bcharge = charge
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[bcharge], 2):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  }

  /* "qmeq/approach/c_lindblad.pyx":183
 *                 #--------------------------------------------------
 *     # Normalisation condition
 *     kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     for charge in range(si.ncharge):
 *         for b in si.statesdm[charge]:
 */
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_47 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_zeros); if (unlikely(!__pyx_t_47)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_47);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0r); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_47, __pyx_t_21, __pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_47); __pyx_t_47 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (unlikely(__Pyx_SetItemInt(((PyObject *)__pyx_v_kern), __pyx_v_norm_row, __pyx_t_1, __pyx_t_4qmeq_8approach_10c_lindblad_long_t, 1, __Pyx_PyInt_From_npy_int64, 0, 1, 0) < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":184
 *     # Normalisation condition
 *     kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
 *     for charge in range(si.ncharge):             # <<<<<<<<<<<<<<
 *         for b in si.statesdm[charge]:
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ncharge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_19 = __Pyx_PyInt_As_long(__pyx_t_1); if (unlikely((__pyx_t_19 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_19; __pyx_t_14+=1) {
    __pyx_v_charge = __pyx_t_14;

    /* "qmeq/approach/c_lindblad.pyx":185
 *     kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
 *     for charge in range(si.ncharge):
 *         for b in si.statesdm[charge]:             # <<<<<<<<<<<<<<
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
 *             kern[norm_row, bb] += 1
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_charge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
      __pyx_t_1 = __pyx_t_15; __Pyx_INCREF(__pyx_t_1); __pyx_t_22 = 0;
      __pyx_t_23 = NULL;
    } else {
      __pyx_t_22 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_23 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 185, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    for (;;) {
      if (likely(!__pyx_t_23)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_22 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_22); __Pyx_INCREF(__pyx_t_15); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_1, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        } else {
          if (__pyx_t_22 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_22); __Pyx_INCREF(__pyx_t_15); __pyx_t_22++; if (unlikely(0 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_1, __pyx_t_22); __pyx_t_22++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 185, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        }
      } else {
        __pyx_t_15 = __pyx_t_23(__pyx_t_1);
        if (unlikely(!__pyx_t_15)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 185, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_15);
      }
      __pyx_t_195 = __Pyx_PyInt_As_npy_int64(__pyx_t_15); if (unlikely((__pyx_t_195 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_v_b = __pyx_t_195;

      /* "qmeq/approach/c_lindblad.pyx":186
 *     for charge in range(si.ncharge):
 *         for b in si.statesdm[charge]:
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]             # <<<<<<<<<<<<<<
 *             kern[norm_row, bb] += 1
 *     sys.kern = kern
 */
      __pyx_t_196 = __pyx_v_charge;
      if (__pyx_t_196 < 0) __pyx_t_196 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
      __pyx_t_195 = __pyx_v_b;
      if (__pyx_t_195 < 0) __pyx_t_195 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_197 = __pyx_v_b;
      if (__pyx_t_197 < 0) __pyx_t_197 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_198 = __pyx_v_charge;
      if (__pyx_t_198 < 0) __pyx_t_198 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
      __pyx_t_199 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_196, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_195, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_197, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_198, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
      if (__pyx_t_199 < 0) __pyx_t_199 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
      __pyx_v_bb = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_199, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

      /* "qmeq/approach/c_lindblad.pyx":187
 *         for b in si.statesdm[charge]:
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
 *             kern[norm_row, bb] += 1             # <<<<<<<<<<<<<<
 *     sys.kern = kern
 *     sys.bvec = bvec
 */
      __pyx_t_200 = __pyx_v_norm_row;
      __pyx_t_201 = __pyx_v_bb;
      if (__pyx_t_200 < 0) __pyx_t_200 += __pyx_pybuffernd_kern.diminfo[0].shape;
      if (__pyx_t_201 < 0) __pyx_t_201 += __pyx_pybuffernd_kern.diminfo[1].shape;
      *__Pyx_BufPtrStrided2d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_kern.rcbuffer->pybuffer.buf, __pyx_t_200, __pyx_pybuffernd_kern.diminfo[0].strides, __pyx_t_201, __pyx_pybuffernd_kern.diminfo[1].strides) += 1.0;

      /* "qmeq/approach/c_lindblad.pyx":185
 *     kern[norm_row] = np.zeros(si.ndm0r, dtype=doublenp)
 *     for charge in range(si.ncharge):
 *         for b in si.statesdm[charge]:             # <<<<<<<<<<<<<<
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
 *             kern[norm_row, bb] += 1
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "qmeq/approach/c_lindblad.pyx":188
 *             bb = mapdm0[lenlst[charge]*dictdm[b] + dictdm[b] + shiftlst0[charge]]
 *             kern[norm_row, bb] += 1
 *     sys.kern = kern             # <<<<<<<<<<<<<<
 *     sys.bvec = bvec
 *     return 0
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_kern, ((PyObject *)__pyx_v_kern)) < 0) __PYX_ERR(0, 188, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":189
 *             kern[norm_row, bb] += 1
 *     sys.kern = kern
 *     sys.bvec = bvec             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_bvec, ((PyObject *)__pyx_v_bvec)) < 0) __PYX_ERR(0, 189, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":190
 *     sys.kern = kern
 *     sys.bvec = bvec
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "qmeq/approach/c_lindblad.pyx":66
 * 
 * @cython.boundscheck(False)
 * def c_generate_kern_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_47);
  __Pyx_XDECREF(__pyx_t_48);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bvec.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_kern.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("qmeq.approach.c_lindblad.c_generate_kern_lindblad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_bvec.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_kern.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_E);
  __Pyx_XDECREF((PyObject *)__pyx_v_tLba);
  __Pyx_XDECREF(__pyx_v_si);
  __Pyx_XDECREF((PyObject *)__pyx_v_lenlst);
  __Pyx_XDECREF((PyObject *)__pyx_v_dictdm);
  __Pyx_XDECREF((PyObject *)__pyx_v_shiftlst0);
  __Pyx_XDECREF((PyObject *)__pyx_v_mapdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_booldm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_conjdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_kern);
  __Pyx_XDECREF((PyObject *)__pyx_v_bvec);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "qmeq/approach/c_lindblad.pyx":193
 * 
 * @cython.boundscheck(False)
 * def c_generate_current_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

/* Python wrapper */
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_5c_generate_current_lindblad(PyObject *__pyx_self, PyObject *__pyx_v_sys); /*proto*/
static PyMethodDef __pyx_mdef_4qmeq_8approach_10c_lindblad_5c_generate_current_lindblad = {"c_generate_current_lindblad", (PyCFunction)__pyx_pw_4qmeq_8approach_10c_lindblad_5c_generate_current_lindblad, METH_O, 0};
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_5c_generate_current_lindblad(PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_generate_current_lindblad (wrapper)", 0);
  __pyx_r = __pyx_pf_4qmeq_8approach_10c_lindblad_4c_generate_current_lindblad(__pyx_self, ((PyObject *)__pyx_v_sys));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_4c_generate_current_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sys) {
  PyArrayObject *__pyx_v_phi0p = 0;
  PyArrayObject *__pyx_v_E = 0;
  PyArrayObject *__pyx_v_tLba = 0;
  PyObject *__pyx_v_si = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_bool_t __pyx_v_bbp_conj;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_acharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bcharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_ccharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_charge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_l;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_nleads;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_c;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_a;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_b;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ndm0;
  CYTHON_UNUSED __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ndm1;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_npauli;
  __pyx_t_double_complex __pyx_v_fcta;
  __pyx_t_double_complex __pyx_v_fctc;
  __pyx_t_double_complex __pyx_v_phi0bbp;
  PyArrayObject *__pyx_v_lenlst = 0;
  PyArrayObject *__pyx_v_dictdm = 0;
  PyArrayObject *__pyx_v_shiftlst0 = 0;
  PyArrayObject *__pyx_v_mapdm0 = 0;
  PyArrayObject *__pyx_v_conjdm0 = 0;
  PyArrayObject *__pyx_v_current = 0;
  PyArrayObject *__pyx_v_energy_current = 0;
  PyArrayObject *__pyx_v_phi0 = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_E;
  __Pyx_Buffer __pyx_pybuffer_E;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_conjdm0;
  __Pyx_Buffer __pyx_pybuffer_conjdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_current;
  __Pyx_Buffer __pyx_pybuffer_current;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dictdm;
  __Pyx_Buffer __pyx_pybuffer_dictdm;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_energy_current;
  __Pyx_Buffer __pyx_pybuffer_energy_current;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_lenlst;
  __Pyx_Buffer __pyx_pybuffer_lenlst;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mapdm0;
  __Pyx_Buffer __pyx_pybuffer_mapdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phi0;
  __Pyx_Buffer __pyx_pybuffer_phi0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phi0p;
  __Pyx_Buffer __pyx_pybuffer_phi0p;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_shiftlst0;
  __Pyx_Buffer __pyx_pybuffer_shiftlst0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tLba;
  __Pyx_Buffer __pyx_pybuffer_tLba;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyArrayObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_6;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_7;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_8;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_9;
  PyArrayObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  PyArrayObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  PyArrayObject *__pyx_t_18 = NULL;
  PyArrayObject *__pyx_t_19 = NULL;
  long __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  Py_ssize_t __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  PyObject *(*__pyx_t_26)(PyObject *);
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  int __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_31;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_34;
  __pyx_t_double_complex __pyx_t_35;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_36;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_37;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_38;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  PyObject *(*__pyx_t_41)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_42;
  Py_ssize_t __pyx_t_43;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_44;
  Py_ssize_t __pyx_t_45;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_46;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_47;
  Py_ssize_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_51;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_52;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_53;
  __pyx_t_npy_float64_complex __pyx_t_54;
  Py_ssize_t __pyx_t_55;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_56;
  Py_ssize_t __pyx_t_57;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_60;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  Py_ssize_t __pyx_t_63;
  Py_ssize_t __pyx_t_64;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_65;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_66;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_67;
  Py_ssize_t __pyx_t_68;
  __Pyx_RefNannySetupContext("c_generate_current_lindblad", 0);
  __pyx_pybuffer_phi0p.pybuffer.buf = NULL;
  __pyx_pybuffer_phi0p.refcount = 0;
  __pyx_pybuffernd_phi0p.data = NULL;
  __pyx_pybuffernd_phi0p.rcbuffer = &__pyx_pybuffer_phi0p;
  __pyx_pybuffer_E.pybuffer.buf = NULL;
  __pyx_pybuffer_E.refcount = 0;
  __pyx_pybuffernd_E.data = NULL;
  __pyx_pybuffernd_E.rcbuffer = &__pyx_pybuffer_E;
  __pyx_pybuffer_tLba.pybuffer.buf = NULL;
  __pyx_pybuffer_tLba.refcount = 0;
  __pyx_pybuffernd_tLba.data = NULL;
  __pyx_pybuffernd_tLba.rcbuffer = &__pyx_pybuffer_tLba;
  __pyx_pybuffer_lenlst.pybuffer.buf = NULL;
  __pyx_pybuffer_lenlst.refcount = 0;
  __pyx_pybuffernd_lenlst.data = NULL;
  __pyx_pybuffernd_lenlst.rcbuffer = &__pyx_pybuffer_lenlst;
  __pyx_pybuffer_dictdm.pybuffer.buf = NULL;
  __pyx_pybuffer_dictdm.refcount = 0;
  __pyx_pybuffernd_dictdm.data = NULL;
  __pyx_pybuffernd_dictdm.rcbuffer = &__pyx_pybuffer_dictdm;
  __pyx_pybuffer_shiftlst0.pybuffer.buf = NULL;
  __pyx_pybuffer_shiftlst0.refcount = 0;
  __pyx_pybuffernd_shiftlst0.data = NULL;
  __pyx_pybuffernd_shiftlst0.rcbuffer = &__pyx_pybuffer_shiftlst0;
  __pyx_pybuffer_mapdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_mapdm0.refcount = 0;
  __pyx_pybuffernd_mapdm0.data = NULL;
  __pyx_pybuffernd_mapdm0.rcbuffer = &__pyx_pybuffer_mapdm0;
  __pyx_pybuffer_conjdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_conjdm0.refcount = 0;
  __pyx_pybuffernd_conjdm0.data = NULL;
  __pyx_pybuffernd_conjdm0.rcbuffer = &__pyx_pybuffer_conjdm0;
  __pyx_pybuffer_current.pybuffer.buf = NULL;
  __pyx_pybuffer_current.refcount = 0;
  __pyx_pybuffernd_current.data = NULL;
  __pyx_pybuffernd_current.rcbuffer = &__pyx_pybuffer_current;
  __pyx_pybuffer_energy_current.pybuffer.buf = NULL;
  __pyx_pybuffer_energy_current.refcount = 0;
  __pyx_pybuffernd_energy_current.data = NULL;
  __pyx_pybuffernd_energy_current.rcbuffer = &__pyx_pybuffer_energy_current;
  __pyx_pybuffer_phi0.pybuffer.buf = NULL;
  __pyx_pybuffer_phi0.refcount = 0;
  __pyx_pybuffernd_phi0.data = NULL;
  __pyx_pybuffernd_phi0.rcbuffer = &__pyx_pybuffer_phi0;

  /* "qmeq/approach/c_lindblad.pyx":194
 * @cython.boundscheck(False)
 * def c_generate_current_lindblad(sys):
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_phi0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_t_2 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer, (PyObject*)__pyx_t_2, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_phi0p = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_phi0p.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 194, __pyx_L1_error)
    } else {__pyx_pybuffernd_phi0p.diminfo[0].strides = __pyx_pybuffernd_phi0p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phi0p.diminfo[0].shape = __pyx_pybuffernd_phi0p.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_2 = 0;
  __pyx_v_phi0p = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":195
 * def c_generate_current_lindblad(sys):
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_qd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Ea); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 195, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_E.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_E = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_E.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 195, __pyx_L1_error)
    } else {__pyx_pybuffernd_E.diminfo[0].strides = __pyx_pybuffernd_E.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_E.diminfo[0].shape = __pyx_pybuffernd_E.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_E = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":196
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba             # <<<<<<<<<<<<<<
 *     si = sys.si
 *     #
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_tLba); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 196, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_tLba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 196, __pyx_L1_error)
    } else {__pyx_pybuffernd_tLba.diminfo[0].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tLba.diminfo[0].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tLba.diminfo[1].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tLba.diminfo[1].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_tLba.diminfo[2].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_tLba.diminfo[2].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_tLba = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":197
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si             # <<<<<<<<<<<<<<
 *     #
 *     cdef bool_t bbp_conj
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_si); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_si = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":204
 *     cdef long_t ndm0, ndm1, npauli
 *     cdef complex_t fcta, fctc, phi0bbp
 *     ndm0, ndm1, npauli, nleads = si.ndm0, si.ndm1, si.npauli, si.nleads             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_7 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_npauli); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_8 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nleads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = __Pyx_PyInt_As_npy_long(__pyx_t_3); if (unlikely((__pyx_t_9 == ((npy_long)-1)) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ndm0 = __pyx_t_6;
  __pyx_v_ndm1 = __pyx_t_7;
  __pyx_v_npauli = __pyx_t_8;
  __pyx_v_nleads = __pyx_t_9;

  /* "qmeq/approach/c_lindblad.pyx":206
 *     ndm0, ndm1, npauli, nleads = si.ndm0, si.ndm1, si.npauli, si.nleads
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_lenlst); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 206, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_lenlst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 206, __pyx_L1_error)
    } else {__pyx_pybuffernd_lenlst.diminfo[0].strides = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_lenlst.diminfo[0].shape = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_lenlst = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":207
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_dictdm); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 207, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_dictdm = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 207, __pyx_L1_error)
    } else {__pyx_pybuffernd_dictdm.diminfo[0].strides = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dictdm.diminfo[0].shape = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_11 = 0;
  __pyx_v_dictdm = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":208
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_shiftlst0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_shiftlst0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 208, __pyx_L1_error)
    } else {__pyx_pybuffernd_shiftlst0.diminfo[0].strides = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_shiftlst0.diminfo[0].shape = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_12 = 0;
  __pyx_v_shiftlst0 = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":209
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_mapdm0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mapdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 209, __pyx_L1_error)
    } else {__pyx_pybuffernd_mapdm0.diminfo[0].strides = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mapdm0.diminfo[0].shape = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_13 = 0;
  __pyx_v_mapdm0 = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":210
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] current = np.zeros(nleads, dtype=complexnp)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_conjdm0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 210, __pyx_L1_error)
  __pyx_t_14 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_14, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_conjdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 210, __pyx_L1_error)
    } else {__pyx_pybuffernd_conjdm0.diminfo[0].strides = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_conjdm0.diminfo[0].shape = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_14 = 0;
  __pyx_v_conjdm0 = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":212
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] current = np.zeros(nleads, dtype=complexnp)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=1] energy_current = np.zeros(nleads, dtype=complexnp)
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_npy_long(__pyx_v_nleads); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_16) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_15, __pyx_t_3); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_16) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_16, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 212, __pyx_L1_error)
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_16);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_current.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_current = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_current.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 212, __pyx_L1_error)
    } else {__pyx_pybuffernd_current.diminfo[0].strides = __pyx_pybuffernd_current.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_current.diminfo[0].shape = __pyx_pybuffernd_current.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_17 = 0;
  __pyx_v_current = ((PyArrayObject *)__pyx_t_16);
  __pyx_t_16 = 0;

  /* "qmeq/approach/c_lindblad.pyx":213
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] current = np.zeros(nleads, dtype=complexnp)
 *     cdef np.ndarray[complex_t, ndim=1] energy_current = np.zeros(nleads, dtype=complexnp)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
 *     #
 */
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = __Pyx_PyInt_From_npy_long(__pyx_v_nleads); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16);
  __pyx_t_16 = 0;
  __pyx_t_16 = PyDict_New(); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 213, __pyx_L1_error)
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_energy_current.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_energy_current = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 213, __pyx_L1_error)
    } else {__pyx_pybuffernd_energy_current.diminfo[0].strides = __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_energy_current.diminfo[0].shape = __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_18 = 0;
  __pyx_v_energy_current = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":214
 *     cdef np.ndarray[complex_t, ndim=1] current = np.zeros(nleads, dtype=complexnp)
 *     cdef np.ndarray[complex_t, ndim=1] energy_current = np.zeros(nleads, dtype=complexnp)
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)             # <<<<<<<<<<<<<<
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_ndm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = PyTuple_New(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_3) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_15, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 214, __pyx_L1_error)
  __pyx_t_19 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer, (PyObject*)__pyx_t_19, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_phi0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 214, __pyx_L1_error)
    } else {__pyx_pybuffernd_phi0.diminfo[0].strides = __pyx_pybuffernd_phi0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phi0.diminfo[0].shape = __pyx_pybuffernd_phi0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_19 = 0;
  __pyx_v_phi0 = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":216
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]             # <<<<<<<<<<<<<<
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
 *     #
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), 0, __pyx_v_npauli, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_PyObject_SetSlice(((PyObject *)__pyx_v_phi0), __pyx_t_3, 0, __pyx_v_npauli, NULL, NULL, NULL, 1, 1, 1) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":217
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]             # <<<<<<<<<<<<<<
 *     #
 *     for charge in range(si.ncharge):
 */
  __pyx_t_3 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), __pyx_v_npauli, __pyx_v_ndm0, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyComplex_FromDoubles(0.0, 1.0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), __pyx_v_ndm0, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = PyNumber_Multiply(__pyx_t_1, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = PyNumber_Add(__pyx_t_3, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (__Pyx_PyObject_SetSlice(((PyObject *)__pyx_v_phi0), __pyx_t_15, __pyx_v_npauli, __pyx_v_ndm0, NULL, NULL, NULL, 1, 1, 1) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "qmeq/approach/c_lindblad.pyx":219
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
 *     #
 *     for charge in range(si.ncharge):             # <<<<<<<<<<<<<<
 *         ccharge = charge+1
 *         bcharge = charge
 */
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ncharge); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_20 = __Pyx_PyInt_As_long(__pyx_t_15); if (unlikely((__pyx_t_20 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_20; __pyx_t_9+=1) {
    __pyx_v_charge = __pyx_t_9;

    /* "qmeq/approach/c_lindblad.pyx":220
 *     #
 *     for charge in range(si.ncharge):
 *         ccharge = charge+1             # <<<<<<<<<<<<<<
 *         bcharge = charge
 *         acharge = charge-1
 */
    __pyx_v_ccharge = (__pyx_v_charge + 1);

    /* "qmeq/approach/c_lindblad.pyx":221
 *     for charge in range(si.ncharge):
 *         ccharge = charge+1
 *         bcharge = charge             # <<<<<<<<<<<<<<
 *         acharge = charge-1
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
 */
    __pyx_v_bcharge = __pyx_v_charge;

    /* "qmeq/approach/c_lindblad.pyx":222
 *         ccharge = charge+1
 *         bcharge = charge
 *         acharge = charge-1             # <<<<<<<<<<<<<<
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
    __pyx_v_acharge = (__pyx_v_charge - 1);

    /* "qmeq/approach/c_lindblad.pyx":223
 *         bcharge = charge
 *         acharge = charge-1
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 */
    __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_16, __pyx_n_s_product); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_bcharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_21 = __Pyx_GetItemInt(__pyx_t_16, __pyx_v_bcharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_21);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = NULL;
    __pyx_t_22 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_22 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_1, __pyx_t_21};
      __pyx_t_15 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_16, __pyx_t_1, __pyx_t_21};
      __pyx_t_15 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_22, 2+__pyx_t_22); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
    } else
    #endif
    {
      __pyx_t_23 = PyTuple_New(2+__pyx_t_22); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      if (__pyx_t_16) {
        __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_23, 0, __pyx_t_16); __pyx_t_16 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_23, 0+__pyx_t_22, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_21);
      PyTuple_SET_ITEM(__pyx_t_23, 1+__pyx_t_22, __pyx_t_21);
      __pyx_t_1 = 0;
      __pyx_t_21 = 0;
      __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_23, NULL); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_15)) || PyTuple_CheckExact(__pyx_t_15)) {
      __pyx_t_3 = __pyx_t_15; __Pyx_INCREF(__pyx_t_3); __pyx_t_24 = 0;
      __pyx_t_25 = NULL;
    } else {
      __pyx_t_24 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_25 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 223, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    for (;;) {
      if (likely(!__pyx_t_25)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_24); __Pyx_INCREF(__pyx_t_15); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 223, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_3, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 223, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        } else {
          if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_15 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_24); __Pyx_INCREF(__pyx_t_15); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 223, __pyx_L1_error)
          #else
          __pyx_t_15 = PySequence_ITEM(__pyx_t_3, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 223, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          #endif
        }
      } else {
        __pyx_t_15 = __pyx_t_25(__pyx_t_3);
        if (unlikely(!__pyx_t_15)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 223, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_15);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_15))) || (PyList_CheckExact(__pyx_t_15))) {
        PyObject* sequence = __pyx_t_15;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 223, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_23 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_21 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_23 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_21 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_23);
        __Pyx_INCREF(__pyx_t_21);
        #else
        __pyx_t_23 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 223, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_23);
        __pyx_t_21 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 223, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_21);
        #endif
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_26 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_23 = __pyx_t_26(__pyx_t_1); if (unlikely(!__pyx_t_23)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_23);
        index = 1; __pyx_t_21 = __pyx_t_26(__pyx_t_1); if (unlikely(!__pyx_t_21)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_21);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_26(__pyx_t_1), 2) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
        __pyx_t_26 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_26 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 223, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_t_8 = __Pyx_PyInt_As_npy_int64(__pyx_t_23); if (unlikely((__pyx_t_8 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_npy_int64(__pyx_t_21); if (unlikely((__pyx_t_7 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 223, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
      __pyx_v_b = __pyx_t_8;
      __pyx_v_bp = __pyx_t_7;

      /* "qmeq/approach/c_lindblad.pyx":224
 *         acharge = charge-1
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *             if bbp != -1:
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
      __pyx_t_27 = __pyx_v_bcharge;
      if (__pyx_t_27 < 0) __pyx_t_27 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
      __pyx_t_7 = __pyx_v_b;
      if (__pyx_t_7 < 0) __pyx_t_7 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_8 = __pyx_v_bp;
      if (__pyx_t_8 < 0) __pyx_t_8 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_28 = __pyx_v_bcharge;
      if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
      __pyx_t_6 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
      if (__pyx_t_6 < 0) __pyx_t_6 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
      __pyx_v_bbp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

      /* "qmeq/approach/c_lindblad.pyx":225
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:             # <<<<<<<<<<<<<<
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 */
      __pyx_t_29 = ((__pyx_v_bbp != -1LL) != 0);
      if (__pyx_t_29) {

        /* "qmeq/approach/c_lindblad.pyx":226
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 *                 for l in range(nleads):
 */
        __pyx_t_30 = __pyx_v_bcharge;
        if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
        __pyx_t_31 = __pyx_v_b;
        if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
        __pyx_t_32 = __pyx_v_bp;
        if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
        __pyx_t_33 = __pyx_v_bcharge;
        if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
        __pyx_t_34 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
        if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
        __pyx_v_bbp_conj = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_conjdm0.diminfo[0].strides));

        /* "qmeq/approach/c_lindblad.pyx":227
 *             if bbp != -1:
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()             # <<<<<<<<<<<<<<
 *                 for l in range(nleads):
 *                     for a in si.statesdm[acharge]:
 */
        if ((__pyx_v_bbp_conj != 0)) {
          __pyx_t_36 = __pyx_v_bbp;
          if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_phi0.diminfo[0].shape;
          __pyx_t_35 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_phi0.diminfo[0].strides));
        } else {
          __pyx_t_37 = __pyx_v_bbp;
          if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_phi0.diminfo[0].shape;
          __pyx_t_35 = __Pyx_c_conj_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_phi0.diminfo[0].strides)));
        }
        __pyx_v_phi0bbp = __pyx_t_35;

        /* "qmeq/approach/c_lindblad.pyx":228
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 *                 for l in range(nleads):             # <<<<<<<<<<<<<<
 *                     for a in si.statesdm[acharge]:
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
 */
        __pyx_t_38 = __pyx_v_nleads;
        for (__pyx_t_39 = 0; __pyx_t_39 < __pyx_t_38; __pyx_t_39+=1) {
          __pyx_v_l = __pyx_t_39;

          /* "qmeq/approach/c_lindblad.pyx":229
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 *                 for l in range(nleads):
 *                     for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
 *                         current[l] = current[l] - fcta
 */
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 229, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_21 = __Pyx_GetItemInt(__pyx_t_15, __pyx_v_acharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 229, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (likely(PyList_CheckExact(__pyx_t_21)) || PyTuple_CheckExact(__pyx_t_21)) {
            __pyx_t_15 = __pyx_t_21; __Pyx_INCREF(__pyx_t_15); __pyx_t_40 = 0;
            __pyx_t_41 = NULL;
          } else {
            __pyx_t_40 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_21); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 229, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_41 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 229, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          for (;;) {
            if (likely(!__pyx_t_41)) {
              if (likely(PyList_CheckExact(__pyx_t_15))) {
                if (__pyx_t_40 >= PyList_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_21 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_40); __Pyx_INCREF(__pyx_t_21); __pyx_t_40++; if (unlikely(0 < 0)) __PYX_ERR(0, 229, __pyx_L1_error)
                #else
                __pyx_t_21 = PySequence_ITEM(__pyx_t_15, __pyx_t_40); __pyx_t_40++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 229, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                #endif
              } else {
                if (__pyx_t_40 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_21 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_40); __Pyx_INCREF(__pyx_t_21); __pyx_t_40++; if (unlikely(0 < 0)) __PYX_ERR(0, 229, __pyx_L1_error)
                #else
                __pyx_t_21 = PySequence_ITEM(__pyx_t_15, __pyx_t_40); __pyx_t_40++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 229, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                #endif
              }
            } else {
              __pyx_t_21 = __pyx_t_41(__pyx_t_15);
              if (unlikely(!__pyx_t_21)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 229, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_21);
            }
            __pyx_t_42 = __Pyx_PyInt_As_npy_int64(__pyx_t_21); if (unlikely((__pyx_t_42 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 229, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_v_a = __pyx_t_42;

            /* "qmeq/approach/c_lindblad.pyx":230
 *                 for l in range(nleads):
 *                     for a in si.statesdm[acharge]:
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()             # <<<<<<<<<<<<<<
 *                         current[l] = current[l] - fcta
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta
 */
            __pyx_t_43 = __pyx_v_l;
            __pyx_t_42 = __pyx_v_a;
            __pyx_t_44 = __pyx_v_b;
            if (__pyx_t_43 < 0) __pyx_t_43 += __pyx_pybuffernd_tLba.diminfo[0].shape;
            if (__pyx_t_42 < 0) __pyx_t_42 += __pyx_pybuffernd_tLba.diminfo[1].shape;
            if (__pyx_t_44 < 0) __pyx_t_44 += __pyx_pybuffernd_tLba.diminfo[2].shape;
            __pyx_t_45 = __pyx_v_l;
            __pyx_t_46 = __pyx_v_a;
            __pyx_t_47 = __pyx_v_bp;
            if (__pyx_t_45 < 0) __pyx_t_45 += __pyx_pybuffernd_tLba.diminfo[0].shape;
            if (__pyx_t_46 < 0) __pyx_t_46 += __pyx_pybuffernd_tLba.diminfo[1].shape;
            if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_tLba.diminfo[2].shape;
            __pyx_v_fcta = __Pyx_c_prod_double(__Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_43, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_42, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_44, __pyx_pybuffernd_tLba.diminfo[2].strides)), __pyx_v_phi0bbp), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_45, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_46, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_47, __pyx_pybuffernd_tLba.diminfo[2].strides))));

            /* "qmeq/approach/c_lindblad.pyx":231
 *                     for a in si.statesdm[acharge]:
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
 *                         current[l] = current[l] - fcta             # <<<<<<<<<<<<<<
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta
 *                     for c in si.statesdm[ccharge]:
 */
            __pyx_t_48 = __pyx_v_l;
            if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_current.diminfo[0].shape;
            __pyx_t_49 = __pyx_v_l;
            if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_current.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_current.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_current.diminfo[0].strides) = __Pyx_c_diff_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_current.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_current.diminfo[0].strides)), __pyx_v_fcta);

            /* "qmeq/approach/c_lindblad.pyx":232
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
 *                         current[l] = current[l] - fcta
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta             # <<<<<<<<<<<<<<
 *                     for c in si.statesdm[ccharge]:
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
 */
            __pyx_t_50 = __pyx_v_l;
            if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_energy_current.diminfo[0].shape;
            __pyx_t_35 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_energy_current.diminfo[0].strides));
            __pyx_t_51 = __pyx_v_a;
            if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_52 = __pyx_v_b;
            if (__pyx_t_52 < 0) __pyx_t_52 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_53 = __pyx_v_bp;
            if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_54 = __Pyx_c_sum_npy_float64(__pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_t_35), __Pyx_CIMAG(__pyx_t_35)), __Pyx_c_prod_npy_float64(__pyx_t_npy_float64_complex_from_parts(((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_51, __pyx_pybuffernd_E.diminfo[0].strides)) - (0.5 * ((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_52, __pyx_pybuffernd_E.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_E.diminfo[0].strides))))), 0), __pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_v_fcta), __Pyx_CIMAG(__pyx_v_fcta))));
            __pyx_t_55 = __pyx_v_l;
            if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_energy_current.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.buf, __pyx_t_55, __pyx_pybuffernd_energy_current.diminfo[0].strides) = __pyx_t_double_complex_from_parts(__Pyx_CREAL(__pyx_t_54), __Pyx_CIMAG(__pyx_t_54));

            /* "qmeq/approach/c_lindblad.pyx":229
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 *                 for l in range(nleads):
 *                     for a in si.statesdm[acharge]:             # <<<<<<<<<<<<<<
 *                         fcta = tLba[l, a, b]*phi0bbp*tLba[l, a, bp].conjugate()
 *                         current[l] = current[l] - fcta
 */
          }
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "qmeq/approach/c_lindblad.pyx":233
 *                         current[l] = current[l] - fcta
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta
 *                     for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
 *                         current[l] = current[l] + fctc
 */
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 233, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_21 = __Pyx_GetItemInt(__pyx_t_15, __pyx_v_ccharge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 233, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (likely(PyList_CheckExact(__pyx_t_21)) || PyTuple_CheckExact(__pyx_t_21)) {
            __pyx_t_15 = __pyx_t_21; __Pyx_INCREF(__pyx_t_15); __pyx_t_40 = 0;
            __pyx_t_41 = NULL;
          } else {
            __pyx_t_40 = -1; __pyx_t_15 = PyObject_GetIter(__pyx_t_21); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 233, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_41 = Py_TYPE(__pyx_t_15)->tp_iternext; if (unlikely(!__pyx_t_41)) __PYX_ERR(0, 233, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          for (;;) {
            if (likely(!__pyx_t_41)) {
              if (likely(PyList_CheckExact(__pyx_t_15))) {
                if (__pyx_t_40 >= PyList_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_21 = PyList_GET_ITEM(__pyx_t_15, __pyx_t_40); __Pyx_INCREF(__pyx_t_21); __pyx_t_40++; if (unlikely(0 < 0)) __PYX_ERR(0, 233, __pyx_L1_error)
                #else
                __pyx_t_21 = PySequence_ITEM(__pyx_t_15, __pyx_t_40); __pyx_t_40++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 233, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                #endif
              } else {
                if (__pyx_t_40 >= PyTuple_GET_SIZE(__pyx_t_15)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_21 = PyTuple_GET_ITEM(__pyx_t_15, __pyx_t_40); __Pyx_INCREF(__pyx_t_21); __pyx_t_40++; if (unlikely(0 < 0)) __PYX_ERR(0, 233, __pyx_L1_error)
                #else
                __pyx_t_21 = PySequence_ITEM(__pyx_t_15, __pyx_t_40); __pyx_t_40++; if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 233, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_21);
                #endif
              }
            } else {
              __pyx_t_21 = __pyx_t_41(__pyx_t_15);
              if (unlikely(!__pyx_t_21)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 233, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_21);
            }
            __pyx_t_56 = __Pyx_PyInt_As_npy_int64(__pyx_t_21); if (unlikely((__pyx_t_56 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
            __pyx_v_c = __pyx_t_56;

            /* "qmeq/approach/c_lindblad.pyx":234
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta
 *                     for c in si.statesdm[ccharge]:
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()             # <<<<<<<<<<<<<<
 *                         current[l] = current[l] + fctc
 *                         energy_current[l] = energy_current[l] + (E[c]-0.5*(E[b]+E[bp]))*fctc
 */
            __pyx_t_57 = __pyx_v_l;
            __pyx_t_56 = __pyx_v_c;
            __pyx_t_58 = __pyx_v_b;
            if (__pyx_t_57 < 0) __pyx_t_57 += __pyx_pybuffernd_tLba.diminfo[0].shape;
            if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_tLba.diminfo[1].shape;
            if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_tLba.diminfo[2].shape;
            __pyx_t_59 = __pyx_v_l;
            __pyx_t_60 = __pyx_v_c;
            __pyx_t_61 = __pyx_v_bp;
            if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_tLba.diminfo[0].shape;
            if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_tLba.diminfo[1].shape;
            if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_tLba.diminfo[2].shape;
            __pyx_v_fctc = __Pyx_c_prod_double(__Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_57, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_56, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_58, __pyx_pybuffernd_tLba.diminfo[2].strides)), __pyx_v_phi0bbp), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_59, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_60, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_61, __pyx_pybuffernd_tLba.diminfo[2].strides))));

            /* "qmeq/approach/c_lindblad.pyx":235
 *                     for c in si.statesdm[ccharge]:
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
 *                         current[l] = current[l] + fctc             # <<<<<<<<<<<<<<
 *                         energy_current[l] = energy_current[l] + (E[c]-0.5*(E[b]+E[bp]))*fctc
 *     #
 */
            __pyx_t_62 = __pyx_v_l;
            if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_current.diminfo[0].shape;
            __pyx_t_63 = __pyx_v_l;
            if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_current.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_current.rcbuffer->pybuffer.buf, __pyx_t_63, __pyx_pybuffernd_current.diminfo[0].strides) = __Pyx_c_sum_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_current.rcbuffer->pybuffer.buf, __pyx_t_62, __pyx_pybuffernd_current.diminfo[0].strides)), __pyx_v_fctc);

            /* "qmeq/approach/c_lindblad.pyx":236
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
 *                         current[l] = current[l] + fctc
 *                         energy_current[l] = energy_current[l] + (E[c]-0.5*(E[b]+E[bp]))*fctc             # <<<<<<<<<<<<<<
 *     #
 *     sys.current = np.array(current.real, dtype=doublenp)
 */
            __pyx_t_64 = __pyx_v_l;
            if (__pyx_t_64 < 0) __pyx_t_64 += __pyx_pybuffernd_energy_current.diminfo[0].shape;
            __pyx_t_35 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_energy_current.diminfo[0].strides));
            __pyx_t_65 = __pyx_v_c;
            if (__pyx_t_65 < 0) __pyx_t_65 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_66 = __pyx_v_b;
            if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_67 = __pyx_v_bp;
            if (__pyx_t_67 < 0) __pyx_t_67 += __pyx_pybuffernd_E.diminfo[0].shape;
            __pyx_t_54 = __Pyx_c_sum_npy_float64(__pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_t_35), __Pyx_CIMAG(__pyx_t_35)), __Pyx_c_prod_npy_float64(__pyx_t_npy_float64_complex_from_parts(((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_65, __pyx_pybuffernd_E.diminfo[0].strides)) - (0.5 * ((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_E.diminfo[0].strides)) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_67, __pyx_pybuffernd_E.diminfo[0].strides))))), 0), __pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_v_fctc), __Pyx_CIMAG(__pyx_v_fctc))));
            __pyx_t_68 = __pyx_v_l;
            if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_energy_current.diminfo[0].shape;
            *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_energy_current.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_energy_current.diminfo[0].strides) = __pyx_t_double_complex_from_parts(__Pyx_CREAL(__pyx_t_54), __Pyx_CIMAG(__pyx_t_54));

            /* "qmeq/approach/c_lindblad.pyx":233
 *                         current[l] = current[l] - fcta
 *                         energy_current[l] = energy_current[l] + (E[a]-0.5*(E[b]+E[bp]))*fcta
 *                     for c in si.statesdm[ccharge]:             # <<<<<<<<<<<<<<
 *                         fctc = tLba[l, c, b]*phi0bbp*tLba[l, c, bp].conjugate()
 *                         current[l] = current[l] + fctc
 */
          }
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }

        /* "qmeq/approach/c_lindblad.pyx":225
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:             # <<<<<<<<<<<<<<
 *                 bbp_conj = conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 phi0bbp = phi0[bbp] if bbp_conj else phi0[bbp].conjugate()
 */
      }

      /* "qmeq/approach/c_lindblad.pyx":223
 *         bcharge = charge
 *         acharge = charge-1
 *         for b, bp in itertools.product(si.statesdm[bcharge], si.statesdm[bcharge]):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "qmeq/approach/c_lindblad.pyx":238
 *                         energy_current[l] = energy_current[l] + (E[c]-0.5*(E[b]+E[bp]))*fctc
 *     #
 *     sys.current = np.array(current.real, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     sys.energy_current = np.array(energy_current.real, dtype=doublenp)
 *     sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_current), __pyx_n_s_real); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_23) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_21, __pyx_t_3); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_current, __pyx_t_23) < 0) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "qmeq/approach/c_lindblad.pyx":239
 *     #
 *     sys.current = np.array(current.real, dtype=doublenp)
 *     sys.energy_current = np.array(energy_current.real, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
 *     return 0
 */
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_23, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_energy_current), __pyx_n_s_real); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_21 = PyTuple_New(1); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_23);
  PyTuple_SET_ITEM(__pyx_t_21, 0, __pyx_t_23);
  __pyx_t_23 = 0;
  __pyx_t_23 = PyDict_New(); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_doublenp); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_23, __pyx_n_s_dtype, __pyx_t_15) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_21, __pyx_t_23); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_energy_current, __pyx_t_15) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "qmeq/approach/c_lindblad.pyx":240
 *     sys.current = np.array(current.real, dtype=doublenp)
 *     sys.energy_current = np.array(energy_current.real, dtype=doublenp)
 *     sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_energy_current); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_current); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_21 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_leads); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_21, __pyx_n_s_mulst); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  __pyx_t_21 = PyNumber_Multiply(__pyx_t_23, __pyx_t_3); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_15, __pyx_t_21); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_sys, __pyx_n_s_heat_current, __pyx_t_3) < 0) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":241
 *     sys.energy_current = np.array(energy_current.real, dtype=doublenp)
 *     sys.heat_current = sys.energy_current - sys.current*sys.leads.mulst
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "qmeq/approach/c_lindblad.pyx":193
 * 
 * @cython.boundscheck(False)
 * def c_generate_current_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_23);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_current.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_energy_current.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("qmeq.approach.c_lindblad.c_generate_current_lindblad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_current.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_energy_current.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_phi0p);
  __Pyx_XDECREF((PyObject *)__pyx_v_E);
  __Pyx_XDECREF((PyObject *)__pyx_v_tLba);
  __Pyx_XDECREF(__pyx_v_si);
  __Pyx_XDECREF((PyObject *)__pyx_v_lenlst);
  __Pyx_XDECREF((PyObject *)__pyx_v_dictdm);
  __Pyx_XDECREF((PyObject *)__pyx_v_shiftlst0);
  __Pyx_XDECREF((PyObject *)__pyx_v_mapdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_conjdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_current);
  __Pyx_XDECREF((PyObject *)__pyx_v_energy_current);
  __Pyx_XDECREF((PyObject *)__pyx_v_phi0);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "qmeq/approach/c_lindblad.pyx":244
 * 
 * @cython.boundscheck(False)
 * def c_generate_vec_lindblad(np.ndarray[double_t, ndim=1] phi0p, sys):             # <<<<<<<<<<<<<<
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

/* Python wrapper */
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_7c_generate_vec_lindblad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_4qmeq_8approach_10c_lindblad_7c_generate_vec_lindblad = {"c_generate_vec_lindblad", (PyCFunction)__pyx_pw_4qmeq_8approach_10c_lindblad_7c_generate_vec_lindblad, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_4qmeq_8approach_10c_lindblad_7c_generate_vec_lindblad(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_phi0p = 0;
  PyObject *__pyx_v_sys = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_generate_vec_lindblad (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_phi0p,&__pyx_n_s_sys,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_phi0p)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_sys)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("c_generate_vec_lindblad", 1, 2, 2, 1); __PYX_ERR(0, 244, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "c_generate_vec_lindblad") < 0)) __PYX_ERR(0, 244, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_phi0p = ((PyArrayObject *)values[0]);
    __pyx_v_sys = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("c_generate_vec_lindblad", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 244, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("qmeq.approach.c_lindblad.c_generate_vec_lindblad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_phi0p), __pyx_ptype_5numpy_ndarray, 1, "phi0p", 0))) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_r = __pyx_pf_4qmeq_8approach_10c_lindblad_6c_generate_vec_lindblad(__pyx_self, __pyx_v_phi0p, __pyx_v_sys);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4qmeq_8approach_10c_lindblad_6c_generate_vec_lindblad(CYTHON_UNUSED PyObject *__pyx_self, PyArrayObject *__pyx_v_phi0p, PyObject *__pyx_v_sys) {
  PyArrayObject *__pyx_v_E = 0;
  PyArrayObject *__pyx_v_tLba = 0;
  PyObject *__pyx_v_si = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_norm_row;
  __pyx_t_4qmeq_8approach_10c_lindblad_bool_t __pyx_v_bbp_bool;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_charge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_acharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_bcharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_ccharge;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_l;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_v_nleads;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_b;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_a;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ap;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_aap;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bpp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bppbp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_bbpp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_c;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_cp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ccp;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_ndm0;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_v_npauli;
  __pyx_t_double_complex __pyx_v_fct_aap;
  __pyx_t_double_complex __pyx_v_fct_bppbp;
  __pyx_t_double_complex __pyx_v_fct_bbpp;
  __pyx_t_double_complex __pyx_v_fct_ccp;
  __pyx_t_double_complex __pyx_v_norm;
  __pyx_t_double_complex __pyx_v_phi0aap;
  __pyx_t_double_complex __pyx_v_phi0bppbp;
  __pyx_t_double_complex __pyx_v_phi0bbpp;
  __pyx_t_double_complex __pyx_v_phi0ccp;
  PyArrayObject *__pyx_v_lenlst = 0;
  PyArrayObject *__pyx_v_dictdm = 0;
  PyArrayObject *__pyx_v_shiftlst0 = 0;
  PyArrayObject *__pyx_v_mapdm0 = 0;
  PyArrayObject *__pyx_v_booldm0 = 0;
  PyArrayObject *__pyx_v_conjdm0 = 0;
  PyArrayObject *__pyx_v_phi0 = 0;
  PyArrayObject *__pyx_v_i_dphi0_dt = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_E;
  __Pyx_Buffer __pyx_pybuffer_E;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_booldm0;
  __Pyx_Buffer __pyx_pybuffer_booldm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_conjdm0;
  __Pyx_Buffer __pyx_pybuffer_conjdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_dictdm;
  __Pyx_Buffer __pyx_pybuffer_dictdm;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_i_dphi0_dt;
  __Pyx_Buffer __pyx_pybuffer_i_dphi0_dt;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_lenlst;
  __Pyx_Buffer __pyx_pybuffer_lenlst;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mapdm0;
  __Pyx_Buffer __pyx_pybuffer_mapdm0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phi0;
  __Pyx_Buffer __pyx_pybuffer_phi0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_phi0p;
  __Pyx_Buffer __pyx_pybuffer_phi0p;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_shiftlst0;
  __Pyx_Buffer __pyx_pybuffer_shiftlst0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_tLba;
  __Pyx_Buffer __pyx_pybuffer_tLba;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyArrayObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_5;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_6;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_7;
  PyArrayObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  PyArrayObject *__pyx_t_11 = NULL;
  PyArrayObject *__pyx_t_12 = NULL;
  PyArrayObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyArrayObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  long __pyx_t_18;
  int __pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  Py_ssize_t __pyx_t_21;
  PyObject *(*__pyx_t_22)(PyObject *);
  PyObject *(*__pyx_t_23)(PyObject *);
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_26;
  int __pyx_t_27;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_30;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_33;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_34;
  __pyx_t_double_complex __pyx_t_35;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_36;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_37;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_38;
  __pyx_t_double_complex __pyx_t_39;
  __pyx_t_npy_float64_complex __pyx_t_40;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_41;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_42;
  PyObject *__pyx_t_43 = NULL;
  PyObject *__pyx_t_44 = NULL;
  Py_ssize_t __pyx_t_45;
  PyObject *(*__pyx_t_46)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_47;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_48;
  Py_ssize_t __pyx_t_49;
  Py_ssize_t __pyx_t_50;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_51;
  __pyx_t_4qmeq_8approach_10c_lindblad_int_t __pyx_t_52;
  Py_ssize_t __pyx_t_53;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_54;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_55;
  Py_ssize_t __pyx_t_56;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_57;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_58;
  Py_ssize_t __pyx_t_59;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_60;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_61;
  Py_ssize_t __pyx_t_62;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_63;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_64;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_65;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_66;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_67;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_68;
  Py_ssize_t __pyx_t_69;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_70;
  Py_ssize_t __pyx_t_71;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_72;
  Py_ssize_t __pyx_t_73;
  PyObject *(*__pyx_t_74)(PyObject *);
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_75;
  Py_ssize_t __pyx_t_76;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_77;
  Py_ssize_t __pyx_t_78;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_79;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_80;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_81;
  Py_ssize_t __pyx_t_82;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_83;
  Py_ssize_t __pyx_t_84;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_85;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_86;
  Py_ssize_t __pyx_t_87;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_88;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_89;
  Py_ssize_t __pyx_t_90;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_91;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_92;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_93;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_94;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_95;
  Py_ssize_t __pyx_t_96;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_97;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_98;
  Py_ssize_t __pyx_t_99;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_100;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_101;
  Py_ssize_t __pyx_t_102;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_103;
  Py_ssize_t __pyx_t_104;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_105;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_106;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_107;
  Py_ssize_t __pyx_t_108;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_109;
  Py_ssize_t __pyx_t_110;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_111;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_112;
  Py_ssize_t __pyx_t_113;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_114;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_115;
  Py_ssize_t __pyx_t_116;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_117;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_118;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_119;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_120;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_121;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_122;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_123;
  Py_ssize_t __pyx_t_124;
  Py_ssize_t __pyx_t_125;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_126;
  Py_ssize_t __pyx_t_127;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_128;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_129;
  Py_ssize_t __pyx_t_130;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_131;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_132;
  Py_ssize_t __pyx_t_133;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_134;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_135;
  Py_ssize_t __pyx_t_136;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_137;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_138;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_139;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_140;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_141;
  __pyx_t_4qmeq_8approach_10c_lindblad_long_t __pyx_t_142;
  __Pyx_RefNannySetupContext("c_generate_vec_lindblad", 0);
  __pyx_pybuffer_E.pybuffer.buf = NULL;
  __pyx_pybuffer_E.refcount = 0;
  __pyx_pybuffernd_E.data = NULL;
  __pyx_pybuffernd_E.rcbuffer = &__pyx_pybuffer_E;
  __pyx_pybuffer_tLba.pybuffer.buf = NULL;
  __pyx_pybuffer_tLba.refcount = 0;
  __pyx_pybuffernd_tLba.data = NULL;
  __pyx_pybuffernd_tLba.rcbuffer = &__pyx_pybuffer_tLba;
  __pyx_pybuffer_lenlst.pybuffer.buf = NULL;
  __pyx_pybuffer_lenlst.refcount = 0;
  __pyx_pybuffernd_lenlst.data = NULL;
  __pyx_pybuffernd_lenlst.rcbuffer = &__pyx_pybuffer_lenlst;
  __pyx_pybuffer_dictdm.pybuffer.buf = NULL;
  __pyx_pybuffer_dictdm.refcount = 0;
  __pyx_pybuffernd_dictdm.data = NULL;
  __pyx_pybuffernd_dictdm.rcbuffer = &__pyx_pybuffer_dictdm;
  __pyx_pybuffer_shiftlst0.pybuffer.buf = NULL;
  __pyx_pybuffer_shiftlst0.refcount = 0;
  __pyx_pybuffernd_shiftlst0.data = NULL;
  __pyx_pybuffernd_shiftlst0.rcbuffer = &__pyx_pybuffer_shiftlst0;
  __pyx_pybuffer_mapdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_mapdm0.refcount = 0;
  __pyx_pybuffernd_mapdm0.data = NULL;
  __pyx_pybuffernd_mapdm0.rcbuffer = &__pyx_pybuffer_mapdm0;
  __pyx_pybuffer_booldm0.pybuffer.buf = NULL;
  __pyx_pybuffer_booldm0.refcount = 0;
  __pyx_pybuffernd_booldm0.data = NULL;
  __pyx_pybuffernd_booldm0.rcbuffer = &__pyx_pybuffer_booldm0;
  __pyx_pybuffer_conjdm0.pybuffer.buf = NULL;
  __pyx_pybuffer_conjdm0.refcount = 0;
  __pyx_pybuffernd_conjdm0.data = NULL;
  __pyx_pybuffernd_conjdm0.rcbuffer = &__pyx_pybuffer_conjdm0;
  __pyx_pybuffer_phi0.pybuffer.buf = NULL;
  __pyx_pybuffer_phi0.refcount = 0;
  __pyx_pybuffernd_phi0.data = NULL;
  __pyx_pybuffernd_phi0.rcbuffer = &__pyx_pybuffer_phi0;
  __pyx_pybuffer_i_dphi0_dt.pybuffer.buf = NULL;
  __pyx_pybuffer_i_dphi0_dt.refcount = 0;
  __pyx_pybuffernd_i_dphi0_dt.data = NULL;
  __pyx_pybuffernd_i_dphi0_dt.rcbuffer = &__pyx_pybuffer_i_dphi0_dt;
  __pyx_pybuffer_phi0p.pybuffer.buf = NULL;
  __pyx_pybuffer_phi0p.refcount = 0;
  __pyx_pybuffernd_phi0p.data = NULL;
  __pyx_pybuffernd_phi0p.rcbuffer = &__pyx_pybuffer_phi0p;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer, (PyObject*)__pyx_v_phi0p, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 244, __pyx_L1_error)
  }
  __pyx_pybuffernd_phi0p.diminfo[0].strides = __pyx_pybuffernd_phi0p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phi0p.diminfo[0].shape = __pyx_pybuffernd_phi0p.rcbuffer->pybuffer.shape[0];

  /* "qmeq/approach/c_lindblad.pyx":246
 * def c_generate_vec_lindblad(np.ndarray[double_t, ndim=1] phi0p, sys):
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_qd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Ea); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 246, __pyx_L1_error)
  __pyx_t_3 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_E.rcbuffer->pybuffer, (PyObject*)__pyx_t_3, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_double_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_E = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_E.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 246, __pyx_L1_error)
    } else {__pyx_pybuffernd_E.diminfo[0].strides = __pyx_pybuffernd_E.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_E.diminfo[0].shape = __pyx_pybuffernd_E.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_3 = 0;
  __pyx_v_E = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":247
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba             # <<<<<<<<<<<<<<
 *     si = sys.si
 *     cdef long_t norm_row = sys.funcp.norm_row
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_tLba); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 247, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 3, 0, __pyx_stack) == -1)) {
      __pyx_v_tLba = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 247, __pyx_L1_error)
    } else {__pyx_pybuffernd_tLba.diminfo[0].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_tLba.diminfo[0].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_tLba.diminfo[1].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_tLba.diminfo[1].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[1]; __pyx_pybuffernd_tLba.diminfo[2].strides = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.strides[2]; __pyx_pybuffernd_tLba.diminfo[2].shape = __pyx_pybuffernd_tLba.rcbuffer->pybuffer.shape[2];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_tLba = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":248
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si             # <<<<<<<<<<<<<<
 *     cdef long_t norm_row = sys.funcp.norm_row
 *     #
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_si); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_si = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":249
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 *     si = sys.si
 *     cdef long_t norm_row = sys.funcp.norm_row             # <<<<<<<<<<<<<<
 *     #
 *     cdef bool_t bbp_bool
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_funcp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_norm_row); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_norm_row = __pyx_t_5;

  /* "qmeq/approach/c_lindblad.pyx":261
 *     cdef complex_t fct_aap, fct_bppbp, fct_bbpp, fct_ccp, norm
 *     cdef complex_t phi0aap, phi0bppbp, phi0bbpp, phi0ccp
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ndm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_npauli); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_nleads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_npy_long(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_long)-1)) && PyErr_Occurred())) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ndm0 = __pyx_t_5;
  __pyx_v_npauli = __pyx_t_6;
  __pyx_v_nleads = __pyx_t_7;

  /* "qmeq/approach/c_lindblad.pyx":263
 *     ndm0, npauli, nleads = si.ndm0, si.npauli, si.nleads
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_lenlst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 263, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_lenlst = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 263, __pyx_L1_error)
    } else {__pyx_pybuffernd_lenlst.diminfo[0].strides = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_lenlst.diminfo[0].shape = __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_lenlst = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":264
 *     #
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_dictdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 264, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_dictdm = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 264, __pyx_L1_error)
    } else {__pyx_pybuffernd_dictdm.diminfo[0].strides = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_dictdm.diminfo[0].shape = __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_dictdm = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":265
 *     cdef np.ndarray[long_t, ndim=1] lenlst = si.lenlst
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_shiftlst0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 265, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_shiftlst0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 265, __pyx_L1_error)
    } else {__pyx_pybuffernd_shiftlst0.diminfo[0].strides = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_shiftlst0.diminfo[0].shape = __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_10 = 0;
  __pyx_v_shiftlst0 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":266
 *     cdef np.ndarray[long_t, ndim=1] dictdm = si.dictdm
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_mapdm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 266, __pyx_L1_error)
  __pyx_t_11 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_11, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_long_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mapdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 266, __pyx_L1_error)
    } else {__pyx_pybuffernd_mapdm0.diminfo[0].strides = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mapdm0.diminfo[0].shape = __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_11 = 0;
  __pyx_v_mapdm0 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":267
 *     cdef np.ndarray[long_t, ndim=1] shiftlst0 = si.shiftlst0
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_booldm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_booldm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 267, __pyx_L1_error)
    } else {__pyx_pybuffernd_booldm0.diminfo[0].strides = __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_booldm0.diminfo[0].shape = __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_12 = 0;
  __pyx_v_booldm0 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":268
 *     cdef np.ndarray[long_t, ndim=1] mapdm0 = si.mapdm0
 *     cdef np.ndarray[bool_t, ndim=1] booldm0 = si.booldm0
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_conjdm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 268, __pyx_L1_error)
  __pyx_t_13 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer, (PyObject*)__pyx_t_13, &__Pyx_TypeInfo_nn___pyx_t_4qmeq_8approach_10c_lindblad_bool_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_conjdm0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 268, __pyx_L1_error)
    } else {__pyx_pybuffernd_conjdm0.diminfo[0].strides = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_conjdm0.diminfo[0].shape = __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_13 = 0;
  __pyx_v_conjdm0 = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "qmeq/approach/c_lindblad.pyx":270
 *     cdef np.ndarray[bool_t, ndim=1] conjdm0 = si.conjdm0
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=1] i_dphi0_dt = np.zeros(ndm0, dtype=complexnp)
 *     #
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_ndm0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_15) < 0) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_14, __pyx_t_1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_15) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_15, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 270, __pyx_L1_error)
  __pyx_t_16 = ((PyArrayObject *)__pyx_t_15);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer, (PyObject*)__pyx_t_16, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_phi0 = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 270, __pyx_L1_error)
    } else {__pyx_pybuffernd_phi0.diminfo[0].strides = __pyx_pybuffernd_phi0.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_phi0.diminfo[0].shape = __pyx_pybuffernd_phi0.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_16 = 0;
  __pyx_v_phi0 = ((PyArrayObject *)__pyx_t_15);
  __pyx_t_15 = 0;

  /* "qmeq/approach/c_lindblad.pyx":271
 *     #
 *     cdef np.ndarray[complex_t, ndim=1] phi0 = np.zeros(ndm0, dtype=complexnp)
 *     cdef np.ndarray[complex_t, ndim=1] i_dphi0_dt = np.zeros(ndm0, dtype=complexnp)             # <<<<<<<<<<<<<<
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]
 */
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __pyx_t_15 = __Pyx_PyInt_From_npy_int64(__pyx_v_ndm0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = PyTuple_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_15);
  __pyx_t_15 = 0;
  __pyx_t_15 = PyDict_New(); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_complexnp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_14, __pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo___pyx_t_double_complex, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_i_dphi0_dt = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 271, __pyx_L1_error)
    } else {__pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides = __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape = __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_17 = 0;
  __pyx_v_i_dphi0_dt = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":273
 *     cdef np.ndarray[complex_t, ndim=1] i_dphi0_dt = np.zeros(ndm0, dtype=complexnp)
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]             # <<<<<<<<<<<<<<
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
 *     norm = 0
 */
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), 0, __pyx_v_npauli, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetSlice(((PyObject *)__pyx_v_phi0), __pyx_t_2, 0, __pyx_v_npauli, NULL, NULL, NULL, 1, 1, 1) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":274
 *     #
 *     phi0[0:npauli] = phi0p[0:npauli]
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]             # <<<<<<<<<<<<<<
 *     norm = 0
 *     for charge in range(si.ncharge):
 */
  __pyx_t_2 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), __pyx_v_npauli, __pyx_v_ndm0, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_15 = PyComplex_FromDoubles(0.0, 1.0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_14 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_phi0p), __pyx_v_ndm0, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_15, __pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyNumber_Add(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetSlice(((PyObject *)__pyx_v_phi0), __pyx_t_14, __pyx_v_npauli, __pyx_v_ndm0, NULL, NULL, NULL, 1, 1, 1) < 0) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

  /* "qmeq/approach/c_lindblad.pyx":275
 *     phi0[0:npauli] = phi0p[0:npauli]
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
 *     norm = 0             # <<<<<<<<<<<<<<
 *     for charge in range(si.ncharge):
 *         acharge = charge-1
 */
  __pyx_v_norm = __pyx_t_double_complex_from_parts(0, 0);

  /* "qmeq/approach/c_lindblad.pyx":276
 *     phi0[npauli:ndm0] = phi0p[npauli:ndm0] + 1j*phi0p[ndm0:]
 *     norm = 0
 *     for charge in range(si.ncharge):             # <<<<<<<<<<<<<<
 *         acharge = charge-1
 *         bcharge = charge
 */
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_ncharge); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_18 = __Pyx_PyInt_As_long(__pyx_t_14); if (unlikely((__pyx_t_18 == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 276, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_18; __pyx_t_7+=1) {
    __pyx_v_charge = __pyx_t_7;

    /* "qmeq/approach/c_lindblad.pyx":277
 *     norm = 0
 *     for charge in range(si.ncharge):
 *         acharge = charge-1             # <<<<<<<<<<<<<<
 *         bcharge = charge
 *         ccharge = charge+1
 */
    __pyx_v_acharge = (__pyx_v_charge - 1);

    /* "qmeq/approach/c_lindblad.pyx":278
 *     for charge in range(si.ncharge):
 *         acharge = charge-1
 *         bcharge = charge             # <<<<<<<<<<<<<<
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
 */
    __pyx_v_bcharge = __pyx_v_charge;

    /* "qmeq/approach/c_lindblad.pyx":279
 *         acharge = charge-1
 *         bcharge = charge
 *         ccharge = charge+1             # <<<<<<<<<<<<<<
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
    __pyx_v_ccharge = (__pyx_v_charge + 1);

    /* "qmeq/approach/c_lindblad.pyx":280
 *         bcharge = charge
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_combinations_with_replacement); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_15 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_charge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    __pyx_t_19 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_19 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_15, __pyx_int_2};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_15, __pyx_int_2};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    } else
    #endif
    {
      __pyx_t_20 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_20);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_15);
      PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_19, __pyx_t_15);
      __Pyx_INCREF(__pyx_int_2);
      __Pyx_GIVEREF(__pyx_int_2);
      PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_19, __pyx_int_2);
      __pyx_t_15 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_20, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
      __pyx_t_2 = __pyx_t_14; __Pyx_INCREF(__pyx_t_2); __pyx_t_21 = 0;
      __pyx_t_22 = NULL;
    } else {
      __pyx_t_21 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_22 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 280, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    for (;;) {
      if (likely(!__pyx_t_22)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_21 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_21); __Pyx_INCREF(__pyx_t_14); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 280, __pyx_L1_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_2, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        } else {
          if (__pyx_t_21 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_21); __Pyx_INCREF(__pyx_t_14); __pyx_t_21++; if (unlikely(0 < 0)) __PYX_ERR(0, 280, __pyx_L1_error)
          #else
          __pyx_t_14 = PySequence_ITEM(__pyx_t_2, __pyx_t_21); __pyx_t_21++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 280, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          #endif
        }
      } else {
        __pyx_t_14 = __pyx_t_22(__pyx_t_2);
        if (unlikely(!__pyx_t_14)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 280, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_14);
      }
      if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
        PyObject* sequence = __pyx_t_14;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 280, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_20 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_15 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_20 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_15 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_20);
        __Pyx_INCREF(__pyx_t_15);
        #else
        __pyx_t_20 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_15 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        #endif
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_1 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_23 = Py_TYPE(__pyx_t_1)->tp_iternext;
        index = 0; __pyx_t_20 = __pyx_t_23(__pyx_t_1); if (unlikely(!__pyx_t_20)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_20);
        index = 1; __pyx_t_15 = __pyx_t_23(__pyx_t_1); if (unlikely(!__pyx_t_15)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_15);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_23(__pyx_t_1), 2) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
        __pyx_t_23 = NULL;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_23 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 280, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_t_6 = __Pyx_PyInt_As_npy_int64(__pyx_t_20); if (unlikely((__pyx_t_6 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
      __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_15); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_v_b = __pyx_t_6;
      __pyx_v_bp = __pyx_t_5;

      /* "qmeq/approach/c_lindblad.pyx":281
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *             if bbp != -1:
 *                 if b == bp: norm = norm + phi0[bbp]
 */
      __pyx_t_24 = __pyx_v_bcharge;
      if (__pyx_t_24 < 0) __pyx_t_24 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
      __pyx_t_5 = __pyx_v_b;
      if (__pyx_t_5 < 0) __pyx_t_5 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_6 = __pyx_v_bp;
      if (__pyx_t_6 < 0) __pyx_t_6 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
      __pyx_t_25 = __pyx_v_bcharge;
      if (__pyx_t_25 < 0) __pyx_t_25 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
      __pyx_t_26 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_24, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
      if (__pyx_t_26 < 0) __pyx_t_26 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
      __pyx_v_bbp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

      /* "qmeq/approach/c_lindblad.pyx":282
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:             # <<<<<<<<<<<<<<
 *                 if b == bp: norm = norm + phi0[bbp]
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
      __pyx_t_27 = ((__pyx_v_bbp != -1LL) != 0);
      if (__pyx_t_27) {

        /* "qmeq/approach/c_lindblad.pyx":283
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 *                 if b == bp: norm = norm + phi0[bbp]             # <<<<<<<<<<<<<<
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 if bbp_bool:
 */
        __pyx_t_27 = ((__pyx_v_b == __pyx_v_bp) != 0);
        if (__pyx_t_27) {
          __pyx_t_28 = __pyx_v_bbp;
          if (__pyx_t_28 < 0) __pyx_t_28 += __pyx_pybuffernd_phi0.diminfo[0].shape;
          __pyx_v_norm = __Pyx_c_sum_double(__pyx_v_norm, (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_28, __pyx_pybuffernd_phi0.diminfo[0].strides)));
        }

        /* "qmeq/approach/c_lindblad.pyx":284
 *             if bbp != -1:
 *                 if b == bp: norm = norm + phi0[bbp]
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                 if bbp_bool:
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
 */
        __pyx_t_29 = __pyx_v_bcharge;
        if (__pyx_t_29 < 0) __pyx_t_29 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
        __pyx_t_30 = __pyx_v_b;
        if (__pyx_t_30 < 0) __pyx_t_30 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
        __pyx_t_31 = __pyx_v_bp;
        if (__pyx_t_31 < 0) __pyx_t_31 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
        __pyx_t_32 = __pyx_v_bcharge;
        if (__pyx_t_32 < 0) __pyx_t_32 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
        __pyx_t_33 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_29, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_30, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_31, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_32, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
        if (__pyx_t_33 < 0) __pyx_t_33 += __pyx_pybuffernd_booldm0.diminfo[0].shape;
        __pyx_v_bbp_bool = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_booldm0.rcbuffer->pybuffer.buf, __pyx_t_33, __pyx_pybuffernd_booldm0.diminfo[0].strides));

        /* "qmeq/approach/c_lindblad.pyx":285
 *                 if b == bp: norm = norm + phi0[bbp]
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 if bbp_bool:             # <<<<<<<<<<<<<<
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
 *                     #--------------------------------------------------
 */
        __pyx_t_27 = (__pyx_v_bbp_bool != 0);
        if (__pyx_t_27) {

          /* "qmeq/approach/c_lindblad.pyx":286
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 if bbp_bool:
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]             # <<<<<<<<<<<<<<
 *                     #--------------------------------------------------
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
 */
          __pyx_t_34 = __pyx_v_bbp;
          if (__pyx_t_34 < 0) __pyx_t_34 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
          __pyx_t_35 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_34, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides));
          __pyx_t_36 = __pyx_v_b;
          if (__pyx_t_36 < 0) __pyx_t_36 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_37 = __pyx_v_bp;
          if (__pyx_t_37 < 0) __pyx_t_37 += __pyx_pybuffernd_E.diminfo[0].shape;
          __pyx_t_38 = __pyx_v_bbp;
          if (__pyx_t_38 < 0) __pyx_t_38 += __pyx_pybuffernd_phi0.diminfo[0].shape;
          __pyx_t_39 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_phi0.diminfo[0].strides));
          __pyx_t_40 = __Pyx_c_sum_npy_float64(__pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_t_35), __Pyx_CIMAG(__pyx_t_35)), __Pyx_c_prod_npy_float64(__pyx_t_npy_float64_complex_from_parts(((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_36, __pyx_pybuffernd_E.diminfo[0].strides)) - (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_double_t *, __pyx_pybuffernd_E.rcbuffer->pybuffer.buf, __pyx_t_37, __pyx_pybuffernd_E.diminfo[0].strides))), 0), __pyx_t_npy_float64_complex_from_parts(__Pyx_CREAL(__pyx_t_39), __Pyx_CIMAG(__pyx_t_39))));
          __pyx_t_41 = __pyx_v_bbp;
          if (__pyx_t_41 < 0) __pyx_t_41 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
          *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_41, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __pyx_t_double_complex_from_parts(__Pyx_CREAL(__pyx_t_40), __Pyx_CIMAG(__pyx_t_40));

          /* "qmeq/approach/c_lindblad.pyx":288
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
 *                     #--------------------------------------------------
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):             # <<<<<<<<<<<<<<
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                         if aap != -1:
 */
          __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_product); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_20);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_42 = (__pyx_v_charge - 1);
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_15, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_42 = (__pyx_v_charge - 1);
          __pyx_t_43 = __Pyx_GetItemInt(__pyx_t_15, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_43);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __pyx_t_15 = NULL;
          __pyx_t_19 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
            __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
            if (likely(__pyx_t_15)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_20, function);
              __pyx_t_19 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_20)) {
            PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_t_1, __pyx_t_43};
            __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_20)) {
            PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_t_1, __pyx_t_43};
            __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_20, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
          } else
          #endif
          {
            __pyx_t_44 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_44);
            if (__pyx_t_15) {
              __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_44, 0, __pyx_t_15); __pyx_t_15 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_44, 0+__pyx_t_19, __pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_43);
            PyTuple_SET_ITEM(__pyx_t_44, 1+__pyx_t_19, __pyx_t_43);
            __pyx_t_1 = 0;
            __pyx_t_43 = 0;
            __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_44, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
          }
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
            __pyx_t_20 = __pyx_t_14; __Pyx_INCREF(__pyx_t_20); __pyx_t_45 = 0;
            __pyx_t_46 = NULL;
          } else {
            __pyx_t_45 = -1; __pyx_t_20 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_20);
            __pyx_t_46 = Py_TYPE(__pyx_t_20)->tp_iternext; if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 288, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          for (;;) {
            if (likely(!__pyx_t_46)) {
              if (likely(PyList_CheckExact(__pyx_t_20))) {
                if (__pyx_t_45 >= PyList_GET_SIZE(__pyx_t_20)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_14 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_45); __Pyx_INCREF(__pyx_t_14); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 288, __pyx_L1_error)
                #else
                __pyx_t_14 = PySequence_ITEM(__pyx_t_20, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                #endif
              } else {
                if (__pyx_t_45 >= PyTuple_GET_SIZE(__pyx_t_20)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_20, __pyx_t_45); __Pyx_INCREF(__pyx_t_14); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 288, __pyx_L1_error)
                #else
                __pyx_t_14 = PySequence_ITEM(__pyx_t_20, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 288, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                #endif
              }
            } else {
              __pyx_t_14 = __pyx_t_46(__pyx_t_20);
              if (unlikely(!__pyx_t_14)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 288, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_14);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_14))) || (PyList_CheckExact(__pyx_t_14))) {
              PyObject* sequence = __pyx_t_14;
              #if !CYTHON_COMPILING_IN_PYPY
              Py_ssize_t size = Py_SIZE(sequence);
              #else
              Py_ssize_t size = PySequence_Size(sequence);
              #endif
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 288, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_44 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_43 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_44 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_43 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_44);
              __Pyx_INCREF(__pyx_t_43);
              #else
              __pyx_t_44 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 288, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_44);
              __pyx_t_43 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 288, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_43);
              #endif
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_1 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 288, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_23 = Py_TYPE(__pyx_t_1)->tp_iternext;
              index = 0; __pyx_t_44 = __pyx_t_23(__pyx_t_1); if (unlikely(!__pyx_t_44)) goto __pyx_L14_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_44);
              index = 1; __pyx_t_43 = __pyx_t_23(__pyx_t_1); if (unlikely(!__pyx_t_43)) goto __pyx_L14_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_43);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_23(__pyx_t_1), 2) < 0) __PYX_ERR(0, 288, __pyx_L1_error)
              __pyx_t_23 = NULL;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              goto __pyx_L15_unpacking_done;
              __pyx_L14_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_23 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 288, __pyx_L1_error)
              __pyx_L15_unpacking_done:;
            }
            __pyx_t_47 = __Pyx_PyInt_As_npy_int64(__pyx_t_44); if (unlikely((__pyx_t_47 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
            __pyx_t_48 = __Pyx_PyInt_As_npy_int64(__pyx_t_43); if (unlikely((__pyx_t_48 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 288, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
            __pyx_v_a = __pyx_t_47;
            __pyx_v_ap = __pyx_t_48;

            /* "qmeq/approach/c_lindblad.pyx":289
 *                     #--------------------------------------------------
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]             # <<<<<<<<<<<<<<
 *                         if aap != -1:
 *                             fct_aap = 0
 */
            __pyx_t_49 = __pyx_v_acharge;
            if (__pyx_t_49 < 0) __pyx_t_49 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_48 = __pyx_v_a;
            if (__pyx_t_48 < 0) __pyx_t_48 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_47 = __pyx_v_ap;
            if (__pyx_t_47 < 0) __pyx_t_47 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_50 = __pyx_v_acharge;
            if (__pyx_t_50 < 0) __pyx_t_50 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_51 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_49, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_48, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_47, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_50, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_51 < 0) __pyx_t_51 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
            __pyx_v_aap = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_51, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

            /* "qmeq/approach/c_lindblad.pyx":290
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                         if aap != -1:             # <<<<<<<<<<<<<<
 *                             fct_aap = 0
 *                             for l in range(nleads):
 */
            __pyx_t_27 = ((__pyx_v_aap != -1LL) != 0);
            if (__pyx_t_27) {

              /* "qmeq/approach/c_lindblad.pyx":291
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                         if aap != -1:
 *                             fct_aap = 0             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 */
              __pyx_v_fct_aap = __pyx_t_double_complex_from_parts(0, 0);

              /* "qmeq/approach/c_lindblad.pyx":292
 *                         if aap != -1:
 *                             fct_aap = 0
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                             phi0aap = phi0[aap] if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else phi0[aap].conjugate()
 */
              __pyx_t_42 = __pyx_v_nleads;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                __pyx_v_l = __pyx_t_52;

                /* "qmeq/approach/c_lindblad.pyx":293
 *                             fct_aap = 0
 *                             for l in range(nleads):
 *                                 fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()             # <<<<<<<<<<<<<<
 *                             phi0aap = phi0[aap] if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else phi0[aap].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_aap*phi0aap
 */
                __pyx_t_53 = __pyx_v_l;
                __pyx_t_54 = __pyx_v_b;
                __pyx_t_55 = __pyx_v_a;
                if (__pyx_t_53 < 0) __pyx_t_53 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_54 < 0) __pyx_t_54 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_55 < 0) __pyx_t_55 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_56 = __pyx_v_l;
                __pyx_t_57 = __pyx_v_bp;
                __pyx_t_58 = __pyx_v_ap;
                if (__pyx_t_56 < 0) __pyx_t_56 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_57 < 0) __pyx_t_57 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_58 < 0) __pyx_t_58 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_aap = __Pyx_c_sum_double(__pyx_v_fct_aap, __Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_53, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_54, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_55, __pyx_pybuffernd_tLba.diminfo[2].strides)), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_56, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_57, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_58, __pyx_pybuffernd_tLba.diminfo[2].strides)))));
              }

              /* "qmeq/approach/c_lindblad.pyx":294
 *                             for l in range(nleads):
 *                                 fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                             phi0aap = phi0[aap] if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else phi0[aap].conjugate()             # <<<<<<<<<<<<<<
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_aap*phi0aap
 *                     #--------------------------------------------------
 */
              __pyx_t_59 = __pyx_v_acharge;
              if (__pyx_t_59 < 0) __pyx_t_59 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
              __pyx_t_60 = __pyx_v_a;
              if (__pyx_t_60 < 0) __pyx_t_60 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_61 = __pyx_v_ap;
              if (__pyx_t_61 < 0) __pyx_t_61 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_62 = __pyx_v_acharge;
              if (__pyx_t_62 < 0) __pyx_t_62 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
              __pyx_t_63 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_59, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_60, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_61, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_62, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
              if (__pyx_t_63 < 0) __pyx_t_63 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
              if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_63, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
                __pyx_t_64 = __pyx_v_aap;
                if (__pyx_t_64 < 0) __pyx_t_64 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_64, __pyx_pybuffernd_phi0.diminfo[0].strides));
              } else {
                __pyx_t_65 = __pyx_v_aap;
                if (__pyx_t_65 < 0) __pyx_t_65 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = __Pyx_c_conj_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_65, __pyx_pybuffernd_phi0.diminfo[0].strides)));
              }
              __pyx_v_phi0aap = __pyx_t_39;

              /* "qmeq/approach/c_lindblad.pyx":295
 *                                 fct_aap += tLba[l, b, a]*tLba[l, bp, ap].conjugate()
 *                             phi0aap = phi0[aap] if conjdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]] else phi0[aap].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_aap*phi0aap             # <<<<<<<<<<<<<<
 *                     #--------------------------------------------------
 *                     for bpp in si.statesdm[charge]:
 */
              __pyx_t_66 = __pyx_v_bbp;
              if (__pyx_t_66 < 0) __pyx_t_66 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              __pyx_t_67 = __pyx_v_bbp;
              if (__pyx_t_67 < 0) __pyx_t_67 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_67, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __Pyx_c_sum_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_66, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_fct_aap), __pyx_v_phi0aap));

              /* "qmeq/approach/c_lindblad.pyx":290
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                         if aap != -1:             # <<<<<<<<<<<<<<
 *                             fct_aap = 0
 *                             for l in range(nleads):
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":288
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
 *                     #--------------------------------------------------
 *                     for a, ap in itertools.product(si.statesdm[charge-1], si.statesdm[charge-1]):             # <<<<<<<<<<<<<<
 *                         aap = mapdm0[lenlst[acharge]*dictdm[a] + dictdm[ap] + shiftlst0[acharge]]
 *                         if aap != -1:
 */
          }
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "qmeq/approach/c_lindblad.pyx":297
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_aap*phi0aap
 *                     #--------------------------------------------------
 *                     for bpp in si.statesdm[charge]:             # <<<<<<<<<<<<<<
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                         if bppbp != -1:
 */
          __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 297, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_20);
          __pyx_t_14 = __Pyx_GetItemInt(__pyx_t_20, __pyx_v_charge, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 297, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          if (likely(PyList_CheckExact(__pyx_t_14)) || PyTuple_CheckExact(__pyx_t_14)) {
            __pyx_t_20 = __pyx_t_14; __Pyx_INCREF(__pyx_t_20); __pyx_t_45 = 0;
            __pyx_t_46 = NULL;
          } else {
            __pyx_t_45 = -1; __pyx_t_20 = PyObject_GetIter(__pyx_t_14); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 297, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_20);
            __pyx_t_46 = Py_TYPE(__pyx_t_20)->tp_iternext; if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 297, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          for (;;) {
            if (likely(!__pyx_t_46)) {
              if (likely(PyList_CheckExact(__pyx_t_20))) {
                if (__pyx_t_45 >= PyList_GET_SIZE(__pyx_t_20)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_14 = PyList_GET_ITEM(__pyx_t_20, __pyx_t_45); __Pyx_INCREF(__pyx_t_14); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 297, __pyx_L1_error)
                #else
                __pyx_t_14 = PySequence_ITEM(__pyx_t_20, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 297, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                #endif
              } else {
                if (__pyx_t_45 >= PyTuple_GET_SIZE(__pyx_t_20)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_14 = PyTuple_GET_ITEM(__pyx_t_20, __pyx_t_45); __Pyx_INCREF(__pyx_t_14); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 297, __pyx_L1_error)
                #else
                __pyx_t_14 = PySequence_ITEM(__pyx_t_20, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 297, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                #endif
              }
            } else {
              __pyx_t_14 = __pyx_t_46(__pyx_t_20);
              if (unlikely(!__pyx_t_14)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 297, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_14);
            }
            __pyx_t_68 = __Pyx_PyInt_As_npy_int64(__pyx_t_14); if (unlikely((__pyx_t_68 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 297, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_v_bpp = __pyx_t_68;

            /* "qmeq/approach/c_lindblad.pyx":298
 *                     #--------------------------------------------------
 *                     for bpp in si.statesdm[charge]:
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                         if bppbp != -1:
 *                             fct_bppbp = 0
 */
            __pyx_t_69 = __pyx_v_bcharge;
            if (__pyx_t_69 < 0) __pyx_t_69 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_68 = __pyx_v_bpp;
            if (__pyx_t_68 < 0) __pyx_t_68 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_70 = __pyx_v_bp;
            if (__pyx_t_70 < 0) __pyx_t_70 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_71 = __pyx_v_bcharge;
            if (__pyx_t_71 < 0) __pyx_t_71 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_72 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_69, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_68, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_70, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_71, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_72 < 0) __pyx_t_72 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
            __pyx_v_bppbp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_72, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

            /* "qmeq/approach/c_lindblad.pyx":299
 *                     for bpp in si.statesdm[charge]:
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                         if bppbp != -1:             # <<<<<<<<<<<<<<
 *                             fct_bppbp = 0
 *                             for a in si.statesdm[charge-1]:
 */
            __pyx_t_27 = ((__pyx_v_bppbp != -1LL) != 0);
            if (__pyx_t_27) {

              /* "qmeq/approach/c_lindblad.pyx":300
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                         if bppbp != -1:
 *                             fct_bppbp = 0             # <<<<<<<<<<<<<<
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):
 */
              __pyx_v_fct_bppbp = __pyx_t_double_complex_from_parts(0, 0);

              /* "qmeq/approach/c_lindblad.pyx":301
 *                         if bppbp != -1:
 *                             fct_bppbp = 0
 *                             for a in si.statesdm[charge-1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 301, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_42 = (__pyx_v_charge - 1);
              __pyx_t_43 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 301, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_43);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (likely(PyList_CheckExact(__pyx_t_43)) || PyTuple_CheckExact(__pyx_t_43)) {
                __pyx_t_14 = __pyx_t_43; __Pyx_INCREF(__pyx_t_14); __pyx_t_73 = 0;
                __pyx_t_74 = NULL;
              } else {
                __pyx_t_73 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_43); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 301, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_74 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 301, __pyx_L1_error)
              }
              __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
              for (;;) {
                if (likely(!__pyx_t_74)) {
                  if (likely(PyList_CheckExact(__pyx_t_14))) {
                    if (__pyx_t_73 >= PyList_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 301, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 301, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  } else {
                    if (__pyx_t_73 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 301, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 301, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  }
                } else {
                  __pyx_t_43 = __pyx_t_74(__pyx_t_14);
                  if (unlikely(!__pyx_t_43)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 301, __pyx_L1_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_43);
                }
                __pyx_t_75 = __Pyx_PyInt_As_npy_int64(__pyx_t_43); if (unlikely((__pyx_t_75 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
                __pyx_v_a = __pyx_t_75;

                /* "qmeq/approach/c_lindblad.pyx":302
 *                             fct_bppbp = 0
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                             for c in si.statesdm[charge+1]:
 */
                __pyx_t_42 = __pyx_v_nleads;
                for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                  __pyx_v_l = __pyx_t_52;

                  /* "qmeq/approach/c_lindblad.pyx":303
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]             # <<<<<<<<<<<<<<
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):
 */
                  __pyx_t_76 = __pyx_v_l;
                  __pyx_t_75 = __pyx_v_a;
                  __pyx_t_77 = __pyx_v_b;
                  if (__pyx_t_76 < 0) __pyx_t_76 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_75 < 0) __pyx_t_75 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_77 < 0) __pyx_t_77 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_t_78 = __pyx_v_l;
                  __pyx_t_79 = __pyx_v_a;
                  __pyx_t_80 = __pyx_v_bpp;
                  if (__pyx_t_78 < 0) __pyx_t_78 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_79 < 0) __pyx_t_79 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_80 < 0) __pyx_t_80 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_v_fct_bppbp = __Pyx_c_sum_double(__pyx_v_fct_bppbp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_76, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_75, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_77, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_78, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_79, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_80, __pyx_pybuffernd_tLba.diminfo[2].strides))));
                }

                /* "qmeq/approach/c_lindblad.pyx":301
 *                         if bppbp != -1:
 *                             fct_bppbp = 0
 *                             for a in si.statesdm[charge-1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 */
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "qmeq/approach/c_lindblad.pyx":304
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                             for c in si.statesdm[charge+1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 304, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_42 = (__pyx_v_charge + 1);
              __pyx_t_43 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 304, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_43);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (likely(PyList_CheckExact(__pyx_t_43)) || PyTuple_CheckExact(__pyx_t_43)) {
                __pyx_t_14 = __pyx_t_43; __Pyx_INCREF(__pyx_t_14); __pyx_t_73 = 0;
                __pyx_t_74 = NULL;
              } else {
                __pyx_t_73 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_43); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 304, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_74 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 304, __pyx_L1_error)
              }
              __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
              for (;;) {
                if (likely(!__pyx_t_74)) {
                  if (likely(PyList_CheckExact(__pyx_t_14))) {
                    if (__pyx_t_73 >= PyList_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 304, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 304, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  } else {
                    if (__pyx_t_73 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 304, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 304, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  }
                } else {
                  __pyx_t_43 = __pyx_t_74(__pyx_t_14);
                  if (unlikely(!__pyx_t_43)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 304, __pyx_L1_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_43);
                }
                __pyx_t_81 = __Pyx_PyInt_As_npy_int64(__pyx_t_43); if (unlikely((__pyx_t_81 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 304, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
                __pyx_v_c = __pyx_t_81;

                /* "qmeq/approach/c_lindblad.pyx":305
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                             phi0bppbp = phi0[bppbp] if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else phi0[bppbp].conjugate()
 */
                __pyx_t_42 = __pyx_v_nleads;
                for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                  __pyx_v_l = __pyx_t_52;

                  /* "qmeq/approach/c_lindblad.pyx":306
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]             # <<<<<<<<<<<<<<
 *                             phi0bppbp = phi0[bppbp] if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else phi0[bppbp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bppbp*phi0bppbp
 */
                  __pyx_t_82 = __pyx_v_l;
                  __pyx_t_81 = __pyx_v_c;
                  __pyx_t_83 = __pyx_v_b;
                  if (__pyx_t_82 < 0) __pyx_t_82 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_81 < 0) __pyx_t_81 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_83 < 0) __pyx_t_83 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_t_84 = __pyx_v_l;
                  __pyx_t_85 = __pyx_v_c;
                  __pyx_t_86 = __pyx_v_bpp;
                  if (__pyx_t_84 < 0) __pyx_t_84 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_85 < 0) __pyx_t_85 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_86 < 0) __pyx_t_86 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_v_fct_bppbp = __Pyx_c_sum_double(__pyx_v_fct_bppbp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_82, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_81, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_83, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_84, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_85, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_86, __pyx_pybuffernd_tLba.diminfo[2].strides))));
                }

                /* "qmeq/approach/c_lindblad.pyx":304
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, a, b].conjugate()*tLba[l, a, bpp]
 *                             for c in si.statesdm[charge+1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 */
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "qmeq/approach/c_lindblad.pyx":307
 *                                 for l in range(nleads):
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                             phi0bppbp = phi0[bppbp] if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else phi0[bppbp].conjugate()             # <<<<<<<<<<<<<<
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bppbp*phi0bppbp
 *                         #--------------------------------------------------
 */
              __pyx_t_87 = __pyx_v_bcharge;
              if (__pyx_t_87 < 0) __pyx_t_87 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
              __pyx_t_88 = __pyx_v_bpp;
              if (__pyx_t_88 < 0) __pyx_t_88 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_89 = __pyx_v_bp;
              if (__pyx_t_89 < 0) __pyx_t_89 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_90 = __pyx_v_bcharge;
              if (__pyx_t_90 < 0) __pyx_t_90 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
              __pyx_t_91 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_87, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_88, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_89, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_90, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
              if (__pyx_t_91 < 0) __pyx_t_91 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
              if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_91, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
                __pyx_t_92 = __pyx_v_bppbp;
                if (__pyx_t_92 < 0) __pyx_t_92 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_92, __pyx_pybuffernd_phi0.diminfo[0].strides));
              } else {
                __pyx_t_93 = __pyx_v_bppbp;
                if (__pyx_t_93 < 0) __pyx_t_93 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = __Pyx_c_conj_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_93, __pyx_pybuffernd_phi0.diminfo[0].strides)));
              }
              __pyx_v_phi0bppbp = __pyx_t_39;

              /* "qmeq/approach/c_lindblad.pyx":308
 *                                     fct_bppbp += -0.5*tLba[l, c, b].conjugate()*tLba[l, c, bpp]
 *                             phi0bppbp = phi0[bppbp] if conjdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]] else phi0[bppbp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bppbp*phi0bppbp             # <<<<<<<<<<<<<<
 *                         #--------------------------------------------------
 *                         bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 */
              __pyx_t_94 = __pyx_v_bbp;
              if (__pyx_t_94 < 0) __pyx_t_94 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              __pyx_t_95 = __pyx_v_bbp;
              if (__pyx_t_95 < 0) __pyx_t_95 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_95, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __Pyx_c_sum_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_94, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_fct_bppbp), __pyx_v_phi0bppbp));

              /* "qmeq/approach/c_lindblad.pyx":299
 *                     for bpp in si.statesdm[charge]:
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                         if bppbp != -1:             # <<<<<<<<<<<<<<
 *                             fct_bppbp = 0
 *                             for a in si.statesdm[charge-1]:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":310
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bppbp*phi0bppbp
 *                         #--------------------------------------------------
 *                         bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]             # <<<<<<<<<<<<<<
 *                         if bbpp != -1:
 *                             fct_bbpp = 0
 */
            __pyx_t_96 = __pyx_v_bcharge;
            if (__pyx_t_96 < 0) __pyx_t_96 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_97 = __pyx_v_b;
            if (__pyx_t_97 < 0) __pyx_t_97 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_98 = __pyx_v_bpp;
            if (__pyx_t_98 < 0) __pyx_t_98 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_99 = __pyx_v_bcharge;
            if (__pyx_t_99 < 0) __pyx_t_99 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_100 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_96, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_97, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_98, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_99, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_100 < 0) __pyx_t_100 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
            __pyx_v_bbpp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_100, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

            /* "qmeq/approach/c_lindblad.pyx":311
 *                         #--------------------------------------------------
 *                         bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                         if bbpp != -1:             # <<<<<<<<<<<<<<
 *                             fct_bbpp = 0
 *                             for a in si.statesdm[charge-1]:
 */
            __pyx_t_27 = ((__pyx_v_bbpp != -1LL) != 0);
            if (__pyx_t_27) {

              /* "qmeq/approach/c_lindblad.pyx":312
 *                         bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                         if bbpp != -1:
 *                             fct_bbpp = 0             # <<<<<<<<<<<<<<
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):
 */
              __pyx_v_fct_bbpp = __pyx_t_double_complex_from_parts(0, 0);

              /* "qmeq/approach/c_lindblad.pyx":313
 *                         if bbpp != -1:
 *                             fct_bbpp = 0
 *                             for a in si.statesdm[charge-1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 313, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_42 = (__pyx_v_charge - 1);
              __pyx_t_43 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 313, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_43);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (likely(PyList_CheckExact(__pyx_t_43)) || PyTuple_CheckExact(__pyx_t_43)) {
                __pyx_t_14 = __pyx_t_43; __Pyx_INCREF(__pyx_t_14); __pyx_t_73 = 0;
                __pyx_t_74 = NULL;
              } else {
                __pyx_t_73 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_43); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 313, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_74 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 313, __pyx_L1_error)
              }
              __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
              for (;;) {
                if (likely(!__pyx_t_74)) {
                  if (likely(PyList_CheckExact(__pyx_t_14))) {
                    if (__pyx_t_73 >= PyList_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 313, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 313, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  } else {
                    if (__pyx_t_73 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 313, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 313, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  }
                } else {
                  __pyx_t_43 = __pyx_t_74(__pyx_t_14);
                  if (unlikely(!__pyx_t_43)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 313, __pyx_L1_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_43);
                }
                __pyx_t_101 = __Pyx_PyInt_As_npy_int64(__pyx_t_43); if (unlikely((__pyx_t_101 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
                __pyx_v_a = __pyx_t_101;

                /* "qmeq/approach/c_lindblad.pyx":314
 *                             fct_bbpp = 0
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                             for c in si.statesdm[charge+1]:
 */
                __pyx_t_42 = __pyx_v_nleads;
                for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                  __pyx_v_l = __pyx_t_52;

                  /* "qmeq/approach/c_lindblad.pyx":315
 *                             for a in si.statesdm[charge-1]:
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]             # <<<<<<<<<<<<<<
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):
 */
                  __pyx_t_102 = __pyx_v_l;
                  __pyx_t_101 = __pyx_v_a;
                  __pyx_t_103 = __pyx_v_bpp;
                  if (__pyx_t_102 < 0) __pyx_t_102 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_101 < 0) __pyx_t_101 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_103 < 0) __pyx_t_103 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_t_104 = __pyx_v_l;
                  __pyx_t_105 = __pyx_v_a;
                  __pyx_t_106 = __pyx_v_bp;
                  if (__pyx_t_104 < 0) __pyx_t_104 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_105 < 0) __pyx_t_105 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_106 < 0) __pyx_t_106 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_v_fct_bbpp = __Pyx_c_sum_double(__pyx_v_fct_bbpp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_102, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_101, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_103, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_104, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_105, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_106, __pyx_pybuffernd_tLba.diminfo[2].strides))));
                }

                /* "qmeq/approach/c_lindblad.pyx":313
 *                         if bbpp != -1:
 *                             fct_bbpp = 0
 *                             for a in si.statesdm[charge-1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 */
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "qmeq/approach/c_lindblad.pyx":316
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                             for c in si.statesdm[charge+1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 */
              __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 316, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_42 = (__pyx_v_charge + 1);
              __pyx_t_43 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 316, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_43);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              if (likely(PyList_CheckExact(__pyx_t_43)) || PyTuple_CheckExact(__pyx_t_43)) {
                __pyx_t_14 = __pyx_t_43; __Pyx_INCREF(__pyx_t_14); __pyx_t_73 = 0;
                __pyx_t_74 = NULL;
              } else {
                __pyx_t_73 = -1; __pyx_t_14 = PyObject_GetIter(__pyx_t_43); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 316, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_74 = Py_TYPE(__pyx_t_14)->tp_iternext; if (unlikely(!__pyx_t_74)) __PYX_ERR(0, 316, __pyx_L1_error)
              }
              __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
              for (;;) {
                if (likely(!__pyx_t_74)) {
                  if (likely(PyList_CheckExact(__pyx_t_14))) {
                    if (__pyx_t_73 >= PyList_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyList_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 316, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 316, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  } else {
                    if (__pyx_t_73 >= PyTuple_GET_SIZE(__pyx_t_14)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_43 = PyTuple_GET_ITEM(__pyx_t_14, __pyx_t_73); __Pyx_INCREF(__pyx_t_43); __pyx_t_73++; if (unlikely(0 < 0)) __PYX_ERR(0, 316, __pyx_L1_error)
                    #else
                    __pyx_t_43 = PySequence_ITEM(__pyx_t_14, __pyx_t_73); __pyx_t_73++; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 316, __pyx_L1_error)
                    __Pyx_GOTREF(__pyx_t_43);
                    #endif
                  }
                } else {
                  __pyx_t_43 = __pyx_t_74(__pyx_t_14);
                  if (unlikely(!__pyx_t_43)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(0, 316, __pyx_L1_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_43);
                }
                __pyx_t_107 = __Pyx_PyInt_As_npy_int64(__pyx_t_43); if (unlikely((__pyx_t_107 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L1_error)
                __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
                __pyx_v_c = __pyx_t_107;

                /* "qmeq/approach/c_lindblad.pyx":317
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                             phi0bbpp = phi0[bbpp] if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else phi0[bbpp].conjugate()
 */
                __pyx_t_42 = __pyx_v_nleads;
                for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                  __pyx_v_l = __pyx_t_52;

                  /* "qmeq/approach/c_lindblad.pyx":318
 *                             for c in si.statesdm[charge+1]:
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]             # <<<<<<<<<<<<<<
 *                             phi0bbpp = phi0[bbpp] if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else phi0[bbpp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bbpp*phi0bbpp
 */
                  __pyx_t_108 = __pyx_v_l;
                  __pyx_t_107 = __pyx_v_c;
                  __pyx_t_109 = __pyx_v_bpp;
                  if (__pyx_t_108 < 0) __pyx_t_108 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_107 < 0) __pyx_t_107 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_109 < 0) __pyx_t_109 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_t_110 = __pyx_v_l;
                  __pyx_t_111 = __pyx_v_c;
                  __pyx_t_112 = __pyx_v_bp;
                  if (__pyx_t_110 < 0) __pyx_t_110 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                  if (__pyx_t_111 < 0) __pyx_t_111 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                  if (__pyx_t_112 < 0) __pyx_t_112 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                  __pyx_v_fct_bbpp = __Pyx_c_sum_double(__pyx_v_fct_bbpp, __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(-0.5, 0), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_108, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_107, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_109, __pyx_pybuffernd_tLba.diminfo[2].strides)))), (*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_110, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_111, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_112, __pyx_pybuffernd_tLba.diminfo[2].strides))));
                }

                /* "qmeq/approach/c_lindblad.pyx":316
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, a, bpp].conjugate()*tLba[l, a, bp]
 *                             for c in si.statesdm[charge+1]:             # <<<<<<<<<<<<<<
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 */
              }
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "qmeq/approach/c_lindblad.pyx":319
 *                                 for l in range(nleads):
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                             phi0bbpp = phi0[bbpp] if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else phi0[bbpp].conjugate()             # <<<<<<<<<<<<<<
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bbpp*phi0bbpp
 *                     #--------------------------------------------------
 */
              __pyx_t_113 = __pyx_v_bcharge;
              if (__pyx_t_113 < 0) __pyx_t_113 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
              __pyx_t_114 = __pyx_v_b;
              if (__pyx_t_114 < 0) __pyx_t_114 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_115 = __pyx_v_bpp;
              if (__pyx_t_115 < 0) __pyx_t_115 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_116 = __pyx_v_bcharge;
              if (__pyx_t_116 < 0) __pyx_t_116 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
              __pyx_t_117 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_113, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_114, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_115, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_116, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
              if (__pyx_t_117 < 0) __pyx_t_117 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
              if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_117, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
                __pyx_t_118 = __pyx_v_bbpp;
                if (__pyx_t_118 < 0) __pyx_t_118 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_118, __pyx_pybuffernd_phi0.diminfo[0].strides));
              } else {
                __pyx_t_119 = __pyx_v_bbpp;
                if (__pyx_t_119 < 0) __pyx_t_119 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = __Pyx_c_conj_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_119, __pyx_pybuffernd_phi0.diminfo[0].strides)));
              }
              __pyx_v_phi0bbpp = __pyx_t_39;

              /* "qmeq/approach/c_lindblad.pyx":320
 *                                     fct_bbpp += -0.5*tLba[l, c, bpp].conjugate()*tLba[l, c, bp]
 *                             phi0bbpp = phi0[bbpp] if conjdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]] else phi0[bbpp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bbpp*phi0bbpp             # <<<<<<<<<<<<<<
 *                     #--------------------------------------------------
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
 */
              __pyx_t_120 = __pyx_v_bbp;
              if (__pyx_t_120 < 0) __pyx_t_120 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              __pyx_t_121 = __pyx_v_bbp;
              if (__pyx_t_121 < 0) __pyx_t_121 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_121, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __Pyx_c_sum_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_120, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_fct_bbpp), __pyx_v_phi0bbpp));

              /* "qmeq/approach/c_lindblad.pyx":311
 *                         #--------------------------------------------------
 *                         bbpp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bpp] + shiftlst0[bcharge]]
 *                         if bbpp != -1:             # <<<<<<<<<<<<<<
 *                             fct_bbpp = 0
 *                             for a in si.statesdm[charge-1]:
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":297
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_aap*phi0aap
 *                     #--------------------------------------------------
 *                     for bpp in si.statesdm[charge]:             # <<<<<<<<<<<<<<
 *                         bppbp = mapdm0[lenlst[bcharge]*dictdm[bpp] + dictdm[bp] + shiftlst0[bcharge]]
 *                         if bppbp != -1:
 */
          }
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

          /* "qmeq/approach/c_lindblad.pyx":322
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bbpp*phi0bbpp
 *                     #--------------------------------------------------
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):             # <<<<<<<<<<<<<<
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                         if ccp != -1:
 */
          __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_43 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_product); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_43);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_42 = (__pyx_v_charge + 1);
          __pyx_t_44 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_44);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_si, __pyx_n_s_statesdm); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_42 = (__pyx_v_charge + 1);
          __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_14, __pyx_t_42, __pyx_t_4qmeq_8approach_10c_lindblad_int_t, 1, __Pyx_PyInt_From_npy_long, 0, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = NULL;
          __pyx_t_19 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_43))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_43);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_43);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_43, function);
              __pyx_t_19 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_43)) {
            PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_44, __pyx_t_1};
            __pyx_t_20 = __Pyx_PyFunction_FastCall(__pyx_t_43, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_43)) {
            PyObject *__pyx_temp[3] = {__pyx_t_14, __pyx_t_44, __pyx_t_1};
            __pyx_t_20 = __Pyx_PyCFunction_FastCall(__pyx_t_43, __pyx_temp+1-__pyx_t_19, 2+__pyx_t_19); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else
          #endif
          {
            __pyx_t_15 = PyTuple_New(2+__pyx_t_19); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_15);
            if (__pyx_t_14) {
              __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
            }
            __Pyx_GIVEREF(__pyx_t_44);
            PyTuple_SET_ITEM(__pyx_t_15, 0+__pyx_t_19, __pyx_t_44);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_15, 1+__pyx_t_19, __pyx_t_1);
            __pyx_t_44 = 0;
            __pyx_t_1 = 0;
            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_43, __pyx_t_15, NULL); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          }
          __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
          if (likely(PyList_CheckExact(__pyx_t_20)) || PyTuple_CheckExact(__pyx_t_20)) {
            __pyx_t_43 = __pyx_t_20; __Pyx_INCREF(__pyx_t_43); __pyx_t_45 = 0;
            __pyx_t_46 = NULL;
          } else {
            __pyx_t_45 = -1; __pyx_t_43 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_43);
            __pyx_t_46 = Py_TYPE(__pyx_t_43)->tp_iternext; if (unlikely(!__pyx_t_46)) __PYX_ERR(0, 322, __pyx_L1_error)
          }
          __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          for (;;) {
            if (likely(!__pyx_t_46)) {
              if (likely(PyList_CheckExact(__pyx_t_43))) {
                if (__pyx_t_45 >= PyList_GET_SIZE(__pyx_t_43)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_20 = PyList_GET_ITEM(__pyx_t_43, __pyx_t_45); __Pyx_INCREF(__pyx_t_20); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
                #else
                __pyx_t_20 = PySequence_ITEM(__pyx_t_43, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 322, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_20);
                #endif
              } else {
                if (__pyx_t_45 >= PyTuple_GET_SIZE(__pyx_t_43)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_20 = PyTuple_GET_ITEM(__pyx_t_43, __pyx_t_45); __Pyx_INCREF(__pyx_t_20); __pyx_t_45++; if (unlikely(0 < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
                #else
                __pyx_t_20 = PySequence_ITEM(__pyx_t_43, __pyx_t_45); __pyx_t_45++; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 322, __pyx_L1_error)
                __Pyx_GOTREF(__pyx_t_20);
                #endif
              }
            } else {
              __pyx_t_20 = __pyx_t_46(__pyx_t_43);
              if (unlikely(!__pyx_t_20)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 322, __pyx_L1_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_20);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_20))) || (PyList_CheckExact(__pyx_t_20))) {
              PyObject* sequence = __pyx_t_20;
              #if !CYTHON_COMPILING_IN_PYPY
              Py_ssize_t size = Py_SIZE(sequence);
              #else
              Py_ssize_t size = PySequence_Size(sequence);
              #endif
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 322, __pyx_L1_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_15 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_15 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_1);
              #else
              __pyx_t_15 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 322, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_1);
              #endif
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_44 = PyObject_GetIter(__pyx_t_20); if (unlikely(!__pyx_t_44)) __PYX_ERR(0, 322, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_44);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
              __pyx_t_23 = Py_TYPE(__pyx_t_44)->tp_iternext;
              index = 0; __pyx_t_15 = __pyx_t_23(__pyx_t_44); if (unlikely(!__pyx_t_15)) goto __pyx_L41_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_15);
              index = 1; __pyx_t_1 = __pyx_t_23(__pyx_t_44); if (unlikely(!__pyx_t_1)) goto __pyx_L41_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_1);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_23(__pyx_t_44), 2) < 0) __PYX_ERR(0, 322, __pyx_L1_error)
              __pyx_t_23 = NULL;
              __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
              goto __pyx_L42_unpacking_done;
              __pyx_L41_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_44); __pyx_t_44 = 0;
              __pyx_t_23 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 322, __pyx_L1_error)
              __pyx_L42_unpacking_done:;
            }
            __pyx_t_122 = __Pyx_PyInt_As_npy_int64(__pyx_t_15); if (unlikely((__pyx_t_122 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __pyx_t_123 = __Pyx_PyInt_As_npy_int64(__pyx_t_1); if (unlikely((__pyx_t_123 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 322, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_v_c = __pyx_t_122;
            __pyx_v_cp = __pyx_t_123;

            /* "qmeq/approach/c_lindblad.pyx":323
 *                     #--------------------------------------------------
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]             # <<<<<<<<<<<<<<
 *                         if ccp != -1:
 *                             fct_ccp = 0
 */
            __pyx_t_124 = __pyx_v_ccharge;
            if (__pyx_t_124 < 0) __pyx_t_124 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
            __pyx_t_123 = __pyx_v_c;
            if (__pyx_t_123 < 0) __pyx_t_123 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_122 = __pyx_v_cp;
            if (__pyx_t_122 < 0) __pyx_t_122 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
            __pyx_t_125 = __pyx_v_ccharge;
            if (__pyx_t_125 < 0) __pyx_t_125 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
            __pyx_t_126 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_124, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_123, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_122, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_125, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
            if (__pyx_t_126 < 0) __pyx_t_126 += __pyx_pybuffernd_mapdm0.diminfo[0].shape;
            __pyx_v_ccp = (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_mapdm0.rcbuffer->pybuffer.buf, __pyx_t_126, __pyx_pybuffernd_mapdm0.diminfo[0].strides));

            /* "qmeq/approach/c_lindblad.pyx":324
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                         if ccp != -1:             # <<<<<<<<<<<<<<
 *                             fct_ccp = 0
 *                             for l in range(nleads):
 */
            __pyx_t_27 = ((__pyx_v_ccp != -1LL) != 0);
            if (__pyx_t_27) {

              /* "qmeq/approach/c_lindblad.pyx":325
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                         if ccp != -1:
 *                             fct_ccp = 0             # <<<<<<<<<<<<<<
 *                             for l in range(nleads):
 *                                 fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 */
              __pyx_v_fct_ccp = __pyx_t_double_complex_from_parts(0, 0);

              /* "qmeq/approach/c_lindblad.pyx":326
 *                         if ccp != -1:
 *                             fct_ccp = 0
 *                             for l in range(nleads):             # <<<<<<<<<<<<<<
 *                                 fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                             phi0ccp = phi0[ccp] if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else phi0[ccp].conjugate()
 */
              __pyx_t_42 = __pyx_v_nleads;
              for (__pyx_t_52 = 0; __pyx_t_52 < __pyx_t_42; __pyx_t_52+=1) {
                __pyx_v_l = __pyx_t_52;

                /* "qmeq/approach/c_lindblad.pyx":327
 *                             fct_ccp = 0
 *                             for l in range(nleads):
 *                                 fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()             # <<<<<<<<<<<<<<
 *                             phi0ccp = phi0[ccp] if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else phi0[ccp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_ccp*phi0ccp
 */
                __pyx_t_127 = __pyx_v_l;
                __pyx_t_128 = __pyx_v_b;
                __pyx_t_129 = __pyx_v_c;
                if (__pyx_t_127 < 0) __pyx_t_127 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_128 < 0) __pyx_t_128 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_129 < 0) __pyx_t_129 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_t_130 = __pyx_v_l;
                __pyx_t_131 = __pyx_v_bp;
                __pyx_t_132 = __pyx_v_cp;
                if (__pyx_t_130 < 0) __pyx_t_130 += __pyx_pybuffernd_tLba.diminfo[0].shape;
                if (__pyx_t_131 < 0) __pyx_t_131 += __pyx_pybuffernd_tLba.diminfo[1].shape;
                if (__pyx_t_132 < 0) __pyx_t_132 += __pyx_pybuffernd_tLba.diminfo[2].shape;
                __pyx_v_fct_ccp = __Pyx_c_sum_double(__pyx_v_fct_ccp, __Pyx_c_prod_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_127, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_128, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_129, __pyx_pybuffernd_tLba.diminfo[2].strides)), __Pyx_c_conj_double((*__Pyx_BufPtrStrided3d(__pyx_t_double_complex *, __pyx_pybuffernd_tLba.rcbuffer->pybuffer.buf, __pyx_t_130, __pyx_pybuffernd_tLba.diminfo[0].strides, __pyx_t_131, __pyx_pybuffernd_tLba.diminfo[1].strides, __pyx_t_132, __pyx_pybuffernd_tLba.diminfo[2].strides)))));
              }

              /* "qmeq/approach/c_lindblad.pyx":328
 *                             for l in range(nleads):
 *                                 fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                             phi0ccp = phi0[ccp] if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else phi0[ccp].conjugate()             # <<<<<<<<<<<<<<
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_ccp*phi0ccp
 *                     #--------------------------------------------------
 */
              __pyx_t_133 = __pyx_v_ccharge;
              if (__pyx_t_133 < 0) __pyx_t_133 += __pyx_pybuffernd_lenlst.diminfo[0].shape;
              __pyx_t_134 = __pyx_v_c;
              if (__pyx_t_134 < 0) __pyx_t_134 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_135 = __pyx_v_cp;
              if (__pyx_t_135 < 0) __pyx_t_135 += __pyx_pybuffernd_dictdm.diminfo[0].shape;
              __pyx_t_136 = __pyx_v_ccharge;
              if (__pyx_t_136 < 0) __pyx_t_136 += __pyx_pybuffernd_shiftlst0.diminfo[0].shape;
              __pyx_t_137 = ((((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_lenlst.rcbuffer->pybuffer.buf, __pyx_t_133, __pyx_pybuffernd_lenlst.diminfo[0].strides)) * (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_134, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_dictdm.rcbuffer->pybuffer.buf, __pyx_t_135, __pyx_pybuffernd_dictdm.diminfo[0].strides))) + (*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_long_t *, __pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer.buf, __pyx_t_136, __pyx_pybuffernd_shiftlst0.diminfo[0].strides)));
              if (__pyx_t_137 < 0) __pyx_t_137 += __pyx_pybuffernd_conjdm0.diminfo[0].shape;
              if (((*__Pyx_BufPtrStrided1d(__pyx_t_4qmeq_8approach_10c_lindblad_bool_t *, __pyx_pybuffernd_conjdm0.rcbuffer->pybuffer.buf, __pyx_t_137, __pyx_pybuffernd_conjdm0.diminfo[0].strides)) != 0)) {
                __pyx_t_138 = __pyx_v_ccp;
                if (__pyx_t_138 < 0) __pyx_t_138 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = (*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_138, __pyx_pybuffernd_phi0.diminfo[0].strides));
              } else {
                __pyx_t_139 = __pyx_v_ccp;
                if (__pyx_t_139 < 0) __pyx_t_139 += __pyx_pybuffernd_phi0.diminfo[0].shape;
                __pyx_t_39 = __Pyx_c_conj_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_phi0.rcbuffer->pybuffer.buf, __pyx_t_139, __pyx_pybuffernd_phi0.diminfo[0].strides)));
              }
              __pyx_v_phi0ccp = __pyx_t_39;

              /* "qmeq/approach/c_lindblad.pyx":329
 *                                 fct_ccp += tLba[l, b, c]*tLba[l, bp, cp].conjugate()
 *                             phi0ccp = phi0[ccp] if conjdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]] else phi0[ccp].conjugate()
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_ccp*phi0ccp             # <<<<<<<<<<<<<<
 *                     #--------------------------------------------------
 *     i_dphi0_dt[norm_row] = 1j*(norm-1)
 */
              __pyx_t_140 = __pyx_v_bbp;
              if (__pyx_t_140 < 0) __pyx_t_140 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              __pyx_t_141 = __pyx_v_bbp;
              if (__pyx_t_141 < 0) __pyx_t_141 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
              *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_141, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __Pyx_c_sum_double((*__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_140, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_fct_ccp), __pyx_v_phi0ccp));

              /* "qmeq/approach/c_lindblad.pyx":324
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                         if ccp != -1:             # <<<<<<<<<<<<<<
 *                             fct_ccp = 0
 *                             for l in range(nleads):
 */
            }

            /* "qmeq/approach/c_lindblad.pyx":322
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_bbpp*phi0bbpp
 *                     #--------------------------------------------------
 *                     for c, cp in itertools.product(si.statesdm[charge+1], si.statesdm[charge+1]):             # <<<<<<<<<<<<<<
 *                         ccp = mapdm0[lenlst[ccharge]*dictdm[c] + dictdm[cp] + shiftlst0[ccharge]]
 *                         if ccp != -1:
 */
          }
          __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

          /* "qmeq/approach/c_lindblad.pyx":285
 *                 if b == bp: norm = norm + phi0[bbp]
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *                 if bbp_bool:             # <<<<<<<<<<<<<<
 *                     i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + (E[b]-E[bp])*phi0[bbp]
 *                     #--------------------------------------------------
 */
        }

        /* "qmeq/approach/c_lindblad.pyx":282
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:             # <<<<<<<<<<<<<<
 *                 if b == bp: norm = norm + phi0[bbp]
 *                 bbp_bool = booldm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 */
      }

      /* "qmeq/approach/c_lindblad.pyx":280
 *         bcharge = charge
 *         ccharge = charge+1
 *         for b, bp in itertools.combinations_with_replacement(si.statesdm[charge], 2):             # <<<<<<<<<<<<<<
 *             bbp = mapdm0[lenlst[bcharge]*dictdm[b] + dictdm[bp] + shiftlst0[bcharge]]
 *             if bbp != -1:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "qmeq/approach/c_lindblad.pyx":331
 *                             i_dphi0_dt[bbp] = i_dphi0_dt[bbp] + 1.0j*fct_ccp*phi0ccp
 *                     #--------------------------------------------------
 *     i_dphi0_dt[norm_row] = 1j*(norm-1)             # <<<<<<<<<<<<<<
 *     return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[npauli:ndm0].real))
 * 
 */
  __pyx_t_142 = __pyx_v_norm_row;
  if (__pyx_t_142 < 0) __pyx_t_142 += __pyx_pybuffernd_i_dphi0_dt.diminfo[0].shape;
  *__Pyx_BufPtrStrided1d(__pyx_t_double_complex *, __pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer.buf, __pyx_t_142, __pyx_pybuffernd_i_dphi0_dt.diminfo[0].strides) = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __Pyx_c_diff_double(__pyx_v_norm, __pyx_t_double_complex_from_parts(1, 0)));

  /* "qmeq/approach/c_lindblad.pyx":332
 *                     #--------------------------------------------------
 *     i_dphi0_dt[norm_row] = 1j*(norm-1)
 *     return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[npauli:ndm0].real))             # <<<<<<<<<<<<<<
 * 
 * class Approach_Lindblad(Approach):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_43 = __Pyx_GetModuleGlobalName(__pyx_n_s_np); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_t_43, __pyx_n_s_concatenate); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
  __pyx_t_43 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_i_dphi0_dt), __pyx_n_s_imag); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_43);
  __pyx_t_1 = __Pyx_PyObject_GetSlice(((PyObject *)__pyx_v_i_dphi0_dt), __pyx_v_npauli, __pyx_v_ndm0, NULL, NULL, NULL, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_real); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_43);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_43);
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_15);
  __pyx_t_43 = 0;
  __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_20))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_20);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_20);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_20, function);
    }
  }
  if (!__pyx_t_15) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_20, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_20)) {
      PyObject *__pyx_temp[2] = {__pyx_t_15, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_20, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_20)) {
      PyObject *__pyx_temp[2] = {__pyx_t_15, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_20, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_43 = PyTuple_New(1+1); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_43);
      __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_43, 0, __pyx_t_15); __pyx_t_15 = NULL;
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_43, 0+1, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_20, __pyx_t_43, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "qmeq/approach/c_lindblad.pyx":244
 * 
 * @cython.boundscheck(False)
 * def c_generate_vec_lindblad(np.ndarray[double_t, ndim=1] phi0p, sys):             # <<<<<<<<<<<<<<
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_XDECREF(__pyx_t_44);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("qmeq.approach.c_lindblad.c_generate_vec_lindblad", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_E.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_booldm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_conjdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_dictdm.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_i_dphi0_dt.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_lenlst.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mapdm0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_phi0p.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_shiftlst0.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_tLba.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_E);
  __Pyx_XDECREF((PyObject *)__pyx_v_tLba);
  __Pyx_XDECREF(__pyx_v_si);
  __Pyx_XDECREF((PyObject *)__pyx_v_lenlst);
  __Pyx_XDECREF((PyObject *)__pyx_v_dictdm);
  __Pyx_XDECREF((PyObject *)__pyx_v_shiftlst0);
  __Pyx_XDECREF((PyObject *)__pyx_v_mapdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_booldm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_conjdm0);
  __Pyx_XDECREF((PyObject *)__pyx_v_phi0);
  __Pyx_XDECREF((PyObject *)__pyx_v_i_dphi0_dt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":203
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = ((__pyx_v_info == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":206
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":207
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":209
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":212
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
    goto __pyx_L4;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":214
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  /*else*/ {
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":217
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 218, __pyx_L1_error)

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":221
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 222, __pyx_L1_error)

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":224
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":225
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (__pyx_v_copy_shape != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":229
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":230
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":231
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":232
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":233
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L11;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":235
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":236
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L11:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":237
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":238
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":239
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":242
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":243
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":246
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_2 = ((!(__pyx_v_copy_shape != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":250
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
    goto __pyx_L14;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":253
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L14:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":256
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L20_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_L20_next_or:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L19_bool_binop_done:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_1) {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 259, __pyx_L1_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":260
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":261
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      case NPY_UBYTE:
      __pyx_v_f = ((char *)"B");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":262
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      case NPY_SHORT:
      __pyx_v_f = ((char *)"h");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":263
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      case NPY_USHORT:
      __pyx_v_f = ((char *)"H");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":264
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      case NPY_INT:
      __pyx_v_f = ((char *)"i");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":265
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      case NPY_UINT:
      __pyx_v_f = ((char *)"I");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      case NPY_LONG:
      __pyx_v_f = ((char *)"l");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":267
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      case NPY_ULONG:
      __pyx_v_f = ((char *)"L");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      case NPY_LONGLONG:
      __pyx_v_f = ((char *)"q");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":269
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      case NPY_ULONGLONG:
      __pyx_v_f = ((char *)"Q");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      case NPY_FLOAT:
      __pyx_v_f = ((char *)"f");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":271
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      case NPY_DOUBLE:
      __pyx_v_f = ((char *)"d");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      case NPY_LONGDOUBLE:
      __pyx_v_f = ((char *)"g");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":273
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      case NPY_CFLOAT:
      __pyx_v_f = ((char *)"Zf");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      case NPY_CDOUBLE:
      __pyx_v_f = ((char *)"Zd");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":275
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      case NPY_CLONGDOUBLE:
      __pyx_v_f = ((char *)"Zg");
      break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      case NPY_OBJECT:
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 278, __pyx_L1_error)
      break;
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":279
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":282
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)malloc(0xFF));

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_7 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(1, 285, __pyx_L1_error)
    __pyx_v_f = __pyx_t_7;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":288
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":294
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":771
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":774
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":791
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 794, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 794, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 795, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 796, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 796, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(1, 796, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (__pyx_t_6) {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 799, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 799, __pyx_L1_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_6) {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 803, __pyx_L1_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 821, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (__pyx_t_6) {

        /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 823, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 823, __pyx_L1_error)

        /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(1, 844, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(1, 849, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":969
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
    goto __pyx_L3;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":971
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
  /*else*/ {
    Py_INCREF(__pyx_v_base);

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":972
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":973
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":974
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arr->base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":978
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":980
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":987
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 987, __pyx_L3_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":988
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 988, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 989, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 989, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":993
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 993, __pyx_L3_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":994
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 994, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 995, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 995, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":999
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 999, __pyx_L3_error)

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1000
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1000, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1001, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1001, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "c_lindblad",
    __pyx_k_Module_containing_cython_functio, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_Approach, __pyx_k_Approach, sizeof(__pyx_k_Approach), 0, 0, 1, 1},
  {&__pyx_n_s_Approach_Lindblad, __pyx_k_Approach_Lindblad, sizeof(__pyx_k_Approach_Lindblad), 0, 0, 1, 1},
  {&__pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1, __pyx_k_C_Users_Gediminas_Desktop_qmeq_1, sizeof(__pyx_k_C_Users_Gediminas_Desktop_qmeq_1), 0, 0, 1, 0},
  {&__pyx_n_s_E, __pyx_k_E, sizeof(__pyx_k_E), 0, 0, 1, 1},
  {&__pyx_n_s_Ea, __pyx_k_Ea, sizeof(__pyx_k_Ea), 0, 0, 1, 1},
  {&__pyx_n_s_Eba, __pyx_k_Eba, sizeof(__pyx_k_Eba), 0, 0, 1, 1},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_Lindblad, __pyx_k_Lindblad, sizeof(__pyx_k_Lindblad), 0, 0, 1, 1},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_Tba, __pyx_k_Tba, sizeof(__pyx_k_Tba), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_aap, __pyx_k_aap, sizeof(__pyx_k_aap), 0, 0, 1, 1},
  {&__pyx_n_s_aap_sgn, __pyx_k_aap_sgn, sizeof(__pyx_k_aap_sgn), 0, 0, 1, 1},
  {&__pyx_n_s_aapi, __pyx_k_aapi, sizeof(__pyx_k_aapi), 0, 0, 1, 1},
  {&__pyx_n_s_acharge, __pyx_k_acharge, sizeof(__pyx_k_acharge), 0, 0, 1, 1},
  {&__pyx_n_s_ap, __pyx_k_ap, sizeof(__pyx_k_ap), 0, 0, 1, 1},
  {&__pyx_n_s_aprclass, __pyx_k_aprclass, sizeof(__pyx_k_aprclass), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_bb, __pyx_k_bb, sizeof(__pyx_k_bb), 0, 0, 1, 1},
  {&__pyx_n_s_bbp, __pyx_k_bbp, sizeof(__pyx_k_bbp), 0, 0, 1, 1},
  {&__pyx_n_s_bbp_bool, __pyx_k_bbp_bool, sizeof(__pyx_k_bbp_bool), 0, 0, 1, 1},
  {&__pyx_n_s_bbp_conj, __pyx_k_bbp_conj, sizeof(__pyx_k_bbp_conj), 0, 0, 1, 1},
  {&__pyx_n_s_bbpi, __pyx_k_bbpi, sizeof(__pyx_k_bbpi), 0, 0, 1, 1},
  {&__pyx_n_s_bbpi_bool, __pyx_k_bbpi_bool, sizeof(__pyx_k_bbpi_bool), 0, 0, 1, 1},
  {&__pyx_n_s_bbpp, __pyx_k_bbpp, sizeof(__pyx_k_bbpp), 0, 0, 1, 1},
  {&__pyx_n_s_bbpp_sgn, __pyx_k_bbpp_sgn, sizeof(__pyx_k_bbpp_sgn), 0, 0, 1, 1},
  {&__pyx_n_s_bbppi, __pyx_k_bbppi, sizeof(__pyx_k_bbppi), 0, 0, 1, 1},
  {&__pyx_n_s_bcharge, __pyx_k_bcharge, sizeof(__pyx_k_bcharge), 0, 0, 1, 1},
  {&__pyx_n_s_booldm0, __pyx_k_booldm0, sizeof(__pyx_k_booldm0), 0, 0, 1, 1},
  {&__pyx_n_s_bp, __pyx_k_bp, sizeof(__pyx_k_bp), 0, 0, 1, 1},
  {&__pyx_n_s_bpp, __pyx_k_bpp, sizeof(__pyx_k_bpp), 0, 0, 1, 1},
  {&__pyx_n_s_bppbp, __pyx_k_bppbp, sizeof(__pyx_k_bppbp), 0, 0, 1, 1},
  {&__pyx_n_s_bppbp_sgn, __pyx_k_bppbp_sgn, sizeof(__pyx_k_bppbp_sgn), 0, 0, 1, 1},
  {&__pyx_n_s_bppbpi, __pyx_k_bppbpi, sizeof(__pyx_k_bppbpi), 0, 0, 1, 1},
  {&__pyx_n_s_bvec, __pyx_k_bvec, sizeof(__pyx_k_bvec), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_s_c_generate_current_lindblad, __pyx_k_c_generate_current_lindblad, sizeof(__pyx_k_c_generate_current_lindblad), 0, 0, 1, 1},
  {&__pyx_n_s_c_generate_kern_lindblad, __pyx_k_c_generate_kern_lindblad, sizeof(__pyx_k_c_generate_kern_lindblad), 0, 0, 1, 1},
  {&__pyx_n_s_c_generate_tLba, __pyx_k_c_generate_tLba, sizeof(__pyx_k_c_generate_tLba), 0, 0, 1, 1},
  {&__pyx_n_s_c_generate_vec_lindblad, __pyx_k_c_generate_vec_lindblad, sizeof(__pyx_k_c_generate_vec_lindblad), 0, 0, 1, 1},
  {&__pyx_n_s_ccharge, __pyx_k_ccharge, sizeof(__pyx_k_ccharge), 0, 0, 1, 1},
  {&__pyx_n_s_ccp, __pyx_k_ccp, sizeof(__pyx_k_ccp), 0, 0, 1, 1},
  {&__pyx_n_s_ccp_sgn, __pyx_k_ccp_sgn, sizeof(__pyx_k_ccp_sgn), 0, 0, 1, 1},
  {&__pyx_n_s_ccpi, __pyx_k_ccpi, sizeof(__pyx_k_ccpi), 0, 0, 1, 1},
  {&__pyx_n_s_charge, __pyx_k_charge, sizeof(__pyx_k_charge), 0, 0, 1, 1},
  {&__pyx_n_s_combinations_with_replacement, __pyx_k_combinations_with_replacement, sizeof(__pyx_k_combinations_with_replacement), 0, 0, 1, 1},
  {&__pyx_n_s_complexnp, __pyx_k_complexnp, sizeof(__pyx_k_complexnp), 0, 0, 1, 1},
  {&__pyx_n_s_concatenate, __pyx_k_concatenate, sizeof(__pyx_k_concatenate), 0, 0, 1, 1},
  {&__pyx_n_s_conjdm0, __pyx_k_conjdm0, sizeof(__pyx_k_conjdm0), 0, 0, 1, 1},
  {&__pyx_n_s_cp, __pyx_k_cp, sizeof(__pyx_k_cp), 0, 0, 1, 1},
  {&__pyx_n_s_current, __pyx_k_current, sizeof(__pyx_k_current), 0, 0, 1, 1},
  {&__pyx_n_s_dictdm, __pyx_k_dictdm, sizeof(__pyx_k_dictdm), 0, 0, 1, 1},
  {&__pyx_n_s_dlst, __pyx_k_dlst, sizeof(__pyx_k_dlst), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_doublenp, __pyx_k_doublenp, sizeof(__pyx_k_doublenp), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_energy_current, __pyx_k_energy_current, sizeof(__pyx_k_energy_current), 0, 0, 1, 1},
  {&__pyx_n_s_fct_aap, __pyx_k_fct_aap, sizeof(__pyx_k_fct_aap), 0, 0, 1, 1},
  {&__pyx_n_s_fct_bbpp, __pyx_k_fct_bbpp, sizeof(__pyx_k_fct_bbpp), 0, 0, 1, 1},
  {&__pyx_n_s_fct_bppbp, __pyx_k_fct_bppbp, sizeof(__pyx_k_fct_bppbp), 0, 0, 1, 1},
  {&__pyx_n_s_fct_ccp, __pyx_k_fct_ccp, sizeof(__pyx_k_fct_ccp), 0, 0, 1, 1},
  {&__pyx_n_s_fcta, __pyx_k_fcta, sizeof(__pyx_k_fcta), 0, 0, 1, 1},
  {&__pyx_n_s_fctc, __pyx_k_fctc, sizeof(__pyx_k_fctc), 0, 0, 1, 1},
  {&__pyx_n_s_funcp, __pyx_k_funcp, sizeof(__pyx_k_funcp), 0, 0, 1, 1},
  {&__pyx_n_s_generate_current, __pyx_k_generate_current, sizeof(__pyx_k_generate_current), 0, 0, 1, 1},
  {&__pyx_n_s_generate_fct, __pyx_k_generate_fct, sizeof(__pyx_k_generate_fct), 0, 0, 1, 1},
  {&__pyx_n_s_generate_kern, __pyx_k_generate_kern, sizeof(__pyx_k_generate_kern), 0, 0, 1, 1},
  {&__pyx_n_s_generate_vec, __pyx_k_generate_vec, sizeof(__pyx_k_generate_vec), 0, 0, 1, 1},
  {&__pyx_n_s_heat_current, __pyx_k_heat_current, sizeof(__pyx_k_heat_current), 0, 0, 1, 1},
  {&__pyx_n_s_i_dphi0_dt, __pyx_k_i_dphi0_dt, sizeof(__pyx_k_i_dphi0_dt), 0, 0, 1, 1},
  {&__pyx_n_s_imag, __pyx_k_imag, sizeof(__pyx_k_imag), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_itype, __pyx_k_itype, sizeof(__pyx_k_itype), 0, 0, 1, 1},
  {&__pyx_n_s_kern, __pyx_k_kern, sizeof(__pyx_k_kern), 0, 0, 1, 1},
  {&__pyx_n_s_kerntype, __pyx_k_kerntype, sizeof(__pyx_k_kerntype), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_last_row, __pyx_k_last_row, sizeof(__pyx_k_last_row), 0, 0, 1, 1},
  {&__pyx_n_s_leads, __pyx_k_leads, sizeof(__pyx_k_leads), 0, 0, 1, 1},
  {&__pyx_n_s_lenlst, __pyx_k_lenlst, sizeof(__pyx_k_lenlst), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mapdm0, __pyx_k_mapdm0, sizeof(__pyx_k_mapdm0), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_mulst, __pyx_k_mulst, sizeof(__pyx_k_mulst), 0, 0, 1, 1},
  {&__pyx_n_s_mytypes, __pyx_k_mytypes, sizeof(__pyx_k_mytypes), 0, 0, 1, 1},
  {&__pyx_n_s_ncharge, __pyx_k_ncharge, sizeof(__pyx_k_ncharge), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndm0, __pyx_k_ndm0, sizeof(__pyx_k_ndm0), 0, 0, 1, 1},
  {&__pyx_n_s_ndm0r, __pyx_k_ndm0r, sizeof(__pyx_k_ndm0r), 0, 0, 1, 1},
  {&__pyx_n_s_ndm1, __pyx_k_ndm1, sizeof(__pyx_k_ndm1), 0, 0, 1, 1},
  {&__pyx_n_s_nleads, __pyx_k_nleads, sizeof(__pyx_k_nleads), 0, 0, 1, 1},
  {&__pyx_n_s_nmany, __pyx_k_nmany, sizeof(__pyx_k_nmany), 0, 0, 1, 1},
  {&__pyx_n_s_norm, __pyx_k_norm, sizeof(__pyx_k_norm), 0, 0, 1, 1},
  {&__pyx_n_s_norm_row, __pyx_k_norm_row, sizeof(__pyx_k_norm_row), 0, 0, 1, 1},
  {&__pyx_n_s_norm_rowp, __pyx_k_norm_rowp, sizeof(__pyx_k_norm_rowp), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_npauli, __pyx_k_npauli, sizeof(__pyx_k_npauli), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_phi0, __pyx_k_phi0, sizeof(__pyx_k_phi0), 0, 0, 1, 1},
  {&__pyx_n_s_phi0aap, __pyx_k_phi0aap, sizeof(__pyx_k_phi0aap), 0, 0, 1, 1},
  {&__pyx_n_s_phi0bbp, __pyx_k_phi0bbp, sizeof(__pyx_k_phi0bbp), 0, 0, 1, 1},
  {&__pyx_n_s_phi0bbpp, __pyx_k_phi0bbpp, sizeof(__pyx_k_phi0bbpp), 0, 0, 1, 1},
  {&__pyx_n_s_phi0bppbp, __pyx_k_phi0bppbp, sizeof(__pyx_k_phi0bppbp), 0, 0, 1, 1},
  {&__pyx_n_s_phi0ccp, __pyx_k_phi0ccp, sizeof(__pyx_k_phi0ccp), 0, 0, 1, 1},
  {&__pyx_n_s_phi0p, __pyx_k_phi0p, sizeof(__pyx_k_phi0p), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_product, __pyx_k_product, sizeof(__pyx_k_product), 0, 0, 1, 1},
  {&__pyx_n_s_qd, __pyx_k_qd, sizeof(__pyx_k_qd), 0, 0, 1, 1},
  {&__pyx_n_s_qmeq_approach_c_lindblad, __pyx_k_qmeq_approach_c_lindblad, sizeof(__pyx_k_qmeq_approach_c_lindblad), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_real, __pyx_k_real, sizeof(__pyx_k_real), 0, 0, 1, 1},
  {&__pyx_n_s_rez, __pyx_k_rez, sizeof(__pyx_k_rez), 0, 0, 1, 1},
  {&__pyx_n_s_shiftlst0, __pyx_k_shiftlst0, sizeof(__pyx_k_shiftlst0), 0, 0, 1, 1},
  {&__pyx_n_s_si, __pyx_k_si, sizeof(__pyx_k_si), 0, 0, 1, 1},
  {&__pyx_n_s_statesdm, __pyx_k_statesdm, sizeof(__pyx_k_statesdm), 0, 0, 1, 1},
  {&__pyx_n_s_symq, __pyx_k_symq, sizeof(__pyx_k_symq), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_tLba, __pyx_k_tLba, sizeof(__pyx_k_tLba), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tlst, __pyx_k_tlst, sizeof(__pyx_k_tlst), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 218, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(1, 799, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 989, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "qmeq/approach/c_lindblad.pyx":48
 *     cdef int_t nleads = si.nleads
 *     #
 *     cdef np.ndarray[double_t, ndim=1] rez = np.zeros(2, dtype=doublenp)             # <<<<<<<<<<<<<<
 *     #
 *     cdef np.ndarray[complex_t, ndim=3] tLba = np.zeros((nleads, si.nmany, si.nmany), dtype=complexnp)
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "qmeq/approach/c_lindblad.pyx":32
 * @cython.cdivision(True)
 * @cython.boundscheck(False)
 * def c_generate_tLba(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_tuple__11 = PyTuple_Pack(18, __pyx_n_s_sys, __pyx_n_s_Tba, __pyx_n_s_E, __pyx_n_s_si, __pyx_n_s_mulst, __pyx_n_s_tlst, __pyx_n_s_dlst, __pyx_n_s_itype, __pyx_n_s_b, __pyx_n_s_a, __pyx_n_s_bcharge, __pyx_n_s_acharge, __pyx_n_s_charge, __pyx_n_s_l, __pyx_n_s_Eba, __pyx_n_s_nleads, __pyx_n_s_rez, __pyx_n_s_tLba); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 18, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1, __pyx_n_s_c_generate_tLba, 32, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 32, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":66
 * 
 * @cython.boundscheck(False)
 * def c_generate_kern_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 */
  __pyx_tuple__13 = PyTuple_Pack(52, __pyx_n_s_sys, __pyx_n_s_E, __pyx_n_s_tLba, __pyx_n_s_si, __pyx_n_s_symq, __pyx_n_s_norm_rowp, __pyx_n_s_bbp_bool, __pyx_n_s_bbpi_bool, __pyx_n_s_charge, __pyx_n_s_acharge, __pyx_n_s_bcharge, __pyx_n_s_ccharge, __pyx_n_s_l, __pyx_n_s_nleads, __pyx_n_s_aap_sgn, __pyx_n_s_bppbp_sgn, __pyx_n_s_bbpp_sgn, __pyx_n_s_ccp_sgn, __pyx_n_s_b, __pyx_n_s_bp, __pyx_n_s_bbp, __pyx_n_s_bbpi, __pyx_n_s_bb, __pyx_n_s_a, __pyx_n_s_ap, __pyx_n_s_aap, __pyx_n_s_aapi, __pyx_n_s_bpp, __pyx_n_s_bppbp, __pyx_n_s_bppbpi, __pyx_n_s_bbpp, __pyx_n_s_bbppi, __pyx_n_s_c, __pyx_n_s_cp, __pyx_n_s_ccp, __pyx_n_s_ccpi, __pyx_n_s_norm_row, __pyx_n_s_last_row, __pyx_n_s_ndm0, __pyx_n_s_npauli, __pyx_n_s_fct_aap, __pyx_n_s_fct_bppbp, __pyx_n_s_fct_bbpp, __pyx_n_s_fct_ccp, __pyx_n_s_lenlst, __pyx_n_s_dictdm, __pyx_n_s_shiftlst0, __pyx_n_s_mapdm0, __pyx_n_s_booldm0, __pyx_n_s_conjdm0, __pyx_n_s_kern, __pyx_n_s_bvec); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 52, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1, __pyx_n_s_c_generate_kern_lindblad, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 66, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":193
 * 
 * @cython.boundscheck(False)
 * def c_generate_current_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_tuple__15 = PyTuple_Pack(31, __pyx_n_s_sys, __pyx_n_s_phi0p, __pyx_n_s_E, __pyx_n_s_tLba, __pyx_n_s_si, __pyx_n_s_bbp_conj, __pyx_n_s_acharge, __pyx_n_s_bcharge, __pyx_n_s_ccharge, __pyx_n_s_charge, __pyx_n_s_l, __pyx_n_s_nleads, __pyx_n_s_c, __pyx_n_s_a, __pyx_n_s_b, __pyx_n_s_bp, __pyx_n_s_bbp, __pyx_n_s_ndm0, __pyx_n_s_ndm1, __pyx_n_s_npauli, __pyx_n_s_fcta, __pyx_n_s_fctc, __pyx_n_s_phi0bbp, __pyx_n_s_lenlst, __pyx_n_s_dictdm, __pyx_n_s_shiftlst0, __pyx_n_s_mapdm0, __pyx_n_s_conjdm0, __pyx_n_s_current, __pyx_n_s_energy_current, __pyx_n_s_phi0); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 31, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1, __pyx_n_s_c_generate_current_lindblad, 193, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 193, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":244
 * 
 * @cython.boundscheck(False)
 * def c_generate_vec_lindblad(np.ndarray[double_t, ndim=1] phi0p, sys):             # <<<<<<<<<<<<<<
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_tuple__17 = PyTuple_Pack(49, __pyx_n_s_phi0p, __pyx_n_s_sys, __pyx_n_s_E, __pyx_n_s_tLba, __pyx_n_s_si, __pyx_n_s_norm_row, __pyx_n_s_bbp_bool, __pyx_n_s_charge, __pyx_n_s_acharge, __pyx_n_s_bcharge, __pyx_n_s_ccharge, __pyx_n_s_l, __pyx_n_s_nleads, __pyx_n_s_aap_sgn, __pyx_n_s_bppbp_sgn, __pyx_n_s_bbpp_sgn, __pyx_n_s_ccp_sgn, __pyx_n_s_b, __pyx_n_s_bp, __pyx_n_s_bbp, __pyx_n_s_bb, __pyx_n_s_a, __pyx_n_s_ap, __pyx_n_s_aap, __pyx_n_s_bpp, __pyx_n_s_bppbp, __pyx_n_s_bbpp, __pyx_n_s_c, __pyx_n_s_cp, __pyx_n_s_ccp, __pyx_n_s_ndm0, __pyx_n_s_npauli, __pyx_n_s_fct_aap, __pyx_n_s_fct_bppbp, __pyx_n_s_fct_bbpp, __pyx_n_s_fct_ccp, __pyx_n_s_norm, __pyx_n_s_phi0aap, __pyx_n_s_phi0bppbp, __pyx_n_s_phi0bbpp, __pyx_n_s_phi0ccp, __pyx_n_s_lenlst, __pyx_n_s_dictdm, __pyx_n_s_shiftlst0, __pyx_n_s_mapdm0, __pyx_n_s_booldm0, __pyx_n_s_conjdm0, __pyx_n_s_phi0, __pyx_n_s_i_dphi0_dt); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(2, 0, 49, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_Gediminas_Desktop_qmeq_1, __pyx_n_s_c_generate_vec_lindblad, 244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initc_lindblad(void); /*proto*/
PyMODINIT_FUNC initc_lindblad(void)
#else
PyMODINIT_FUNC PyInit_c_lindblad(void); /*proto*/
PyMODINIT_FUNC PyInit_c_lindblad(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_c_lindblad(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("c_lindblad", __pyx_methods, __pyx_k_Module_containing_cython_functio, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_qmeq__approach__c_lindblad) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "qmeq.approach.c_lindblad")) {
      if (unlikely(PyDict_SetItemString(modules, "qmeq.approach.c_lindblad", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) __PYX_ERR(1, 155, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) __PYX_ERR(1, 168, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) __PYX_ERR(1, 172, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) __PYX_ERR(1, 181, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) __PYX_ERR(1, 861, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("qmeq.specfuncc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "func_pauli", (void (**)(void))&__pyx_f_4qmeq_9specfuncc_func_pauli, "__pyx_t_4qmeq_9specfuncc_int_t (__pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_double_t, __pyx_t_4qmeq_9specfuncc_int_t, PyArrayObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "qmeq/approach/c_lindblad.pyx":7
 * from __future__ import division
 * from __future__ import print_function
 * import numpy as np             # <<<<<<<<<<<<<<
 * import itertools
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":8
 * from __future__ import print_function
 * import numpy as np
 * import itertools             # <<<<<<<<<<<<<<
 * 
 * from ..mytypes import doublenp
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_itertools, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":10
 * import itertools
 * 
 * from ..mytypes import doublenp             # <<<<<<<<<<<<<<
 * from ..mytypes import complexnp
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_doublenp);
  __Pyx_GIVEREF(__pyx_n_s_doublenp);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_doublenp);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_mytypes, __pyx_t_2, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_doublenp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_doublenp, __pyx_t_2) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":11
 * 
 * from ..mytypes import doublenp
 * from ..mytypes import complexnp             # <<<<<<<<<<<<<<
 * 
 * from ..specfuncc cimport func_pauli
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_complexnp);
  __Pyx_GIVEREF(__pyx_n_s_complexnp);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_complexnp);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_mytypes, __pyx_t_3, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_complexnp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_complexnp, __pyx_t_3) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":14
 * 
 * from ..specfuncc cimport func_pauli
 * from ..aprclass import Approach             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Approach);
  __Pyx_GIVEREF(__pyx_n_s_Approach);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Approach);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_aprclass, __pyx_t_2, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Approach); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Approach, __pyx_t_2) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":32
 * @cython.cdivision(True)
 * @cython.boundscheck(False)
 * def c_generate_tLba(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[complex_t, ndim=3] Tba = sys.leads.Tba
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_4qmeq_8approach_10c_lindblad_1c_generate_tLba, NULL, __pyx_n_s_qmeq_approach_c_lindblad); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_generate_tLba, __pyx_t_3) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":66
 * 
 * @cython.boundscheck(False)
 * def c_generate_kern_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 *     cdef np.ndarray[complex_t, ndim=3] tLba = sys.tLba
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_4qmeq_8approach_10c_lindblad_3c_generate_kern_lindblad, NULL, __pyx_n_s_qmeq_approach_c_lindblad); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_generate_kern_lindblad, __pyx_t_3) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":193
 * 
 * @cython.boundscheck(False)
 * def c_generate_current_lindblad(sys):             # <<<<<<<<<<<<<<
 *     cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_4qmeq_8approach_10c_lindblad_5c_generate_current_lindblad, NULL, __pyx_n_s_qmeq_approach_c_lindblad); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_generate_current_lindblad, __pyx_t_3) < 0) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":244
 * 
 * @cython.boundscheck(False)
 * def c_generate_vec_lindblad(np.ndarray[double_t, ndim=1] phi0p, sys):             # <<<<<<<<<<<<<<
 *     #cdef np.ndarray[double_t, ndim=1] phi0p = sys.phi0
 *     cdef np.ndarray[double_t, ndim=1] E = sys.qd.Ea
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_4qmeq_8approach_10c_lindblad_7c_generate_vec_lindblad, NULL, __pyx_n_s_qmeq_approach_c_lindblad); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_generate_vec_lindblad, __pyx_t_3) < 0) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "qmeq/approach/c_lindblad.pyx":334
 *     return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[npauli:ndm0].real))
 * 
 * class Approach_Lindblad(Approach):             # <<<<<<<<<<<<<<
 * 
 *     kerntype = 'Lindblad'
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_Approach); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_2, __pyx_n_s_Approach_Lindblad, __pyx_n_s_Approach_Lindblad, (PyObject *) NULL, __pyx_n_s_qmeq_approach_c_lindblad, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "qmeq/approach/c_lindblad.pyx":336
 * class Approach_Lindblad(Approach):
 * 
 *     kerntype = 'Lindblad'             # <<<<<<<<<<<<<<
 *     generate_fct = c_generate_tLba
 *     generate_kern = c_generate_kern_lindblad
 */
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_kerntype, __pyx_n_s_Lindblad) < 0) __PYX_ERR(0, 336, __pyx_L1_error)

  /* "qmeq/approach/c_lindblad.pyx":337
 * 
 *     kerntype = 'Lindblad'
 *     generate_fct = c_generate_tLba             # <<<<<<<<<<<<<<
 *     generate_kern = c_generate_kern_lindblad
 *     generate_current = c_generate_current_lindblad
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_c_generate_tLba); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_generate_fct, __pyx_t_5) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "qmeq/approach/c_lindblad.pyx":338
 *     kerntype = 'Lindblad'
 *     generate_fct = c_generate_tLba
 *     generate_kern = c_generate_kern_lindblad             # <<<<<<<<<<<<<<
 *     generate_current = c_generate_current_lindblad
 *     generate_vec = c_generate_vec_lindblad
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_c_generate_kern_lindblad); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_generate_kern, __pyx_t_5) < 0) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "qmeq/approach/c_lindblad.pyx":339
 *     generate_fct = c_generate_tLba
 *     generate_kern = c_generate_kern_lindblad
 *     generate_current = c_generate_current_lindblad             # <<<<<<<<<<<<<<
 *     generate_vec = c_generate_vec_lindblad
 * #---------------------------------------------------------------------------------------------------
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_c_generate_current_lindblad); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_generate_current, __pyx_t_5) < 0) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "qmeq/approach/c_lindblad.pyx":340
 *     generate_kern = c_generate_kern_lindblad
 *     generate_current = c_generate_current_lindblad
 *     generate_vec = c_generate_vec_lindblad             # <<<<<<<<<<<<<<
 * #---------------------------------------------------------------------------------------------------
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_c_generate_vec_lindblad); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyObject_SetItem(__pyx_t_4, __pyx_n_s_generate_vec, __pyx_t_5) < 0) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "qmeq/approach/c_lindblad.pyx":334
 *     return np.concatenate((i_dphi0_dt.imag, i_dphi0_dt[npauli:ndm0].real))
 * 
 * class Approach_Lindblad(Approach):             # <<<<<<<<<<<<<<
 * 
 *     kerntype = 'Lindblad'
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_Approach_Lindblad, __pyx_t_2, __pyx_t_4, NULL, 0, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Approach_Lindblad, __pyx_t_5) < 0) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "qmeq/approach/c_lindblad.pyx":1
 * """Module containing cython functions, which generate first order Lindblad kernel.             # <<<<<<<<<<<<<<
 *    For docstrings see documentation of module lindblad."""
 * 
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "../../../../Python/lib/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init qmeq.approach.c_lindblad", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init qmeq.approach.c_lindblad");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFormatCheck */
static CYTHON_INLINE int __Pyx_IsLittleEndian(void) {
  unsigned int n = 1;
  return *(unsigned char*)(&n) != 0;
}
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t < '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static CYTHON_INLINE PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_IsLittleEndian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
      case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}
static CYTHON_INLINE void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static CYTHON_INLINE int __Pyx_GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  if (obj == Py_None || obj == NULL) {
    __Pyx_ZeroBuffer(buf);
    return 0;
  }
  buf->buf = NULL;
  if (__Pyx_GetBuffer(obj, buf, flags) == -1) goto fail;
  if (buf->ndim != nd) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if ((unsigned)buf->itemsize != dtype->size) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_ZeroBuffer(buf);
  return -1;
}
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (info->buf == NULL) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}

/* GetModuleGlobalName */
  static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* GetItemInt */
    static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyFunctionFastCall */
    #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyCFunctionFastCall */
    #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
    static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* SetItemInt */
    static CYTHON_INLINE int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely((n >= 0) & (n < PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o))) {
#else
    if (is_list || PySequence_Check(o)) {
#endif
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* SliceObject */
      static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SliceObject */
      static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
      static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
      static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* PyObjectCallMethO */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
      #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseNoneIterError */
          static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* SaveResetException */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
          #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* Import */
            static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
            static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
            static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
            static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CodeObjectCache */
            static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
            #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
        if (PyObject_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) { __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view); return; }
    Py_DECREF(obj);
    view->obj = NULL;
}
#endif


            /* CIntFromPyVerify */
            #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_long(npy_long value) {
    const npy_long neg_one = (npy_long) -1, const_zero = (npy_long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_long),
                                     little, !is_unsigned);
    }
}

/* Declarations */
            #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
            #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value) {
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value) {
    const npy_uint8 neg_one = (npy_uint8) -1, const_zero = (npy_uint8) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint8) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint8),
                                     little, !is_unsigned);
    }
}

/* Declarations */
            #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_npy_float64_complex __pyx_t_npy_float64_complex_from_parts(npy_float64 x, npy_float64 y) {
      return ::std::complex< npy_float64 >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_npy_float64_complex __pyx_t_npy_float64_complex_from_parts(npy_float64 x, npy_float64 y) {
      return x + y*(__pyx_t_npy_float64_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_npy_float64_complex __pyx_t_npy_float64_complex_from_parts(npy_float64 x, npy_float64 y) {
      __pyx_t_npy_float64_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
            #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_sum_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
        __pyx_t_npy_float64_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_diff_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
        __pyx_t_npy_float64_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_prod_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
        __pyx_t_npy_float64_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_quot_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
        if (b.imag == 0) {
            return __pyx_t_npy_float64_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_npy_float64_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                npy_float64 r = b.imag / b.real;
                npy_float64 s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_npy_float64_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            npy_float64 r = b.real / b.imag;
            npy_float64 s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_npy_float64_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_quot_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
        if (b.imag == 0) {
            return __pyx_t_npy_float64_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            npy_float64 denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_npy_float64_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_neg_npy_float64(__pyx_t_npy_float64_complex a) {
        __pyx_t_npy_float64_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_npy_float64(__pyx_t_npy_float64_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_conj_npy_float64(__pyx_t_npy_float64_complex a) {
        __pyx_t_npy_float64_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE npy_float64 __Pyx_c_abs_npy_float64(__pyx_t_npy_float64_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_npy_float64_complex __Pyx_c_pow_npy_float64(__pyx_t_npy_float64_complex a, __pyx_t_npy_float64_complex b) {
            __pyx_t_npy_float64_complex z;
            npy_float64 r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    npy_float64 denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_npy_float64(a, a);
                        return __Pyx_c_prod_npy_float64(a, a);
                    case 3:
                        z = __Pyx_c_prod_npy_float64(a, a);
                        return __Pyx_c_prod_npy_float64(z, a);
                    case 4:
                        z = __Pyx_c_prod_npy_float64(a, a);
                        return __Pyx_c_prod_npy_float64(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_npy_float64(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
            #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
            #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) -1, const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
            static CYTHON_INLINE npy_long __Pyx_PyInt_As_npy_long(PyObject *x) {
    const npy_long neg_one = (npy_long) -1, const_zero = (npy_long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_long, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) >= 2 * PyLong_SHIFT) {
                            return (npy_long) (((((npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) >= 3 * PyLong_SHIFT) {
                            return (npy_long) (((((((npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) >= 4 * PyLong_SHIFT) {
                            return (npy_long) (((((((((npy_long)digits[3]) << PyLong_SHIFT) | (npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_long) (((npy_long)-1)*(((((npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_long) ((((((npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_long) (((npy_long)-1)*(((((((npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_long) ((((((((npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_long) (((npy_long)-1)*(((((((((npy_long)digits[3]) << PyLong_SHIFT) | (npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_long) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_long) ((((((((((npy_long)digits[3]) << PyLong_SHIFT) | (npy_long)digits[2]) << PyLong_SHIFT) | (npy_long)digits[1]) << PyLong_SHIFT) | (npy_long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_long) -1;
        }
    } else {
        npy_long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_long) -1;
        val = __Pyx_PyInt_As_npy_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_long");
    return (npy_long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_long");
    return (npy_long) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *x) {
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 2 * PyLong_SHIFT) {
                            return (npy_int64) (((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 3 * PyLong_SHIFT) {
                            return (npy_int64) (((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 4 * PyLong_SHIFT) {
                            return (npy_int64) (((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) ((((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int64) -1;
        }
    } else {
        npy_int64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int64) -1;
        val = __Pyx_PyInt_As_npy_int64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int64");
    return (npy_int64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int64");
    return (npy_int64) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
            #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
            #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FunctionImport */
            #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
